#include <Framework/Framework.h>
#include <Framework/Sprite.h>
#include <Framework/LuaScript.h>
#include <Framework/TextureLibrary.h>
#include <cJSON.h>
#include "Animation.h"

#include <string>
#include <map>

namespace {
	std::map<std::string, AnimationPrefab*>		g_animPrefabs;
	std::map<std::string, AnimationSetPrefab*>	g_animSetPrefabs;
}

const char* const enumToString(const AnimationPlayMode& val) {
	switch (val) {
	case AnimationPlayMode::Loop: return "Loop";
	case AnimationPlayMode::OneShot: return "OneShot";
	default: SDL_assert(!"wat");
	}

	return nullptr;
}

const char* const enumToString(const AnimationType& val) {
	switch (val) {
		case AnimationType::Spawn: return "Spawn";
		case AnimationType::Idle: return "Idle";
		case AnimationType::IdleNoWeapon: return "IdleNoWeapon";
		case AnimationType::Move: return "Move";
		case AnimationType::MoveNoWeapon: return "MoveNoWeapon";
		case AnimationType::Attack: return "Attack";
		case AnimationType::Die: return "Die";
		case AnimationType::Explosion: return "Explosion";
		default: SDL_assert(!"wat");
	}

	return nullptr;
}


AnimationPrefab* loadAnimationPrefab(const char* const path) {
	const std::string configFile = std::string(path) + "/animation.json";

	char *txt = jamReadFile(configFile.c_str(), nullptr);
	cJSON* doc = cJSON_Parse(txt);

	if (!doc)
	{
		SDL_LogError(SDL_LOG_CATEGORY_APPLICATION, "Failed to load animation prefab: %s", cJSON_GetErrorPtr());
		SDL_assert(doc);
	}

	const int numFrames = cJSON_GetObjectItem(doc, "frames")->valueint;
	AnimationPrefab *const animPrefab = new AnimationPrefab;

	for (int i = 1; i <= numFrames; ++i) {
		AnimationFrame frame;
		{
			char file[1024] = { 0 };
			sprintf(file, "%s/%d.png", path, i);
			frame.layers[int(AnimationFrameLayer::Color)] = loadSprite(file, 0xff00ff, true, false);
		}
		
		{
			char file[1024] = { 0 };
			sprintf(file, "%s/%d_pivot.png", path, i);

			if (FILE *f = fopen(file, "r"))
			{
				frame.layers[int(AnimationFrameLayer::Pivot)] = loadSprite(file, 0xff00ff, true, false);
				fclose(f);
			}
		}

		animPrefab->frames.push_back(frame);
	}

	animPrefab->playMode = stringToEnum(cJSON_GetObjectItem(doc, "play-mode")->valuestring, AnimationPlayMode::OneShot);

	if (cJSON_GetObjectItem(doc, "fps")) {
		animPrefab->framesPerSecond = float(cJSON_GetObjectItem(doc, "fps")->valuedouble);
	}

	return animPrefab;
}

AnimationSetPrefab* loadAnimationSet(const char* const path) {
	const std::string configFile = jamTranslatePath(std::string(path) + "/animationSet.json");

	char *txt = jamReadFile(configFile.c_str(), nullptr);
	cJSON* doc = cJSON_Parse(txt);

	if (!doc)
	{
		SDL_LogError(SDL_LOG_CATEGORY_APPLICATION, "Failed to load animation set prefab: %s", cJSON_GetErrorPtr());
		SDL_assert(doc);
	}

	cJSON* const animationList = cJSON_GetObjectItem(doc, "animations");
	SDL_assert(animationList != nullptr);

	AnimationSetPrefab *const animSetPrefab = new AnimationSetPrefab;

	const int numAnimations = cJSON_GetArraySize(animationList);
	for (int animIt = 0; animIt < numAnimations; ++animIt) {
		cJSON *const animNameObj = cJSON_GetArrayItem(animationList, animIt);
		const char* const animName = animNameObj->valuestring;
		const AnimationType animType = stringToEnum(animName, AnimationType::NumItems);

		if (animType != AnimationType::NumItems) {
			const std::string name = (std::string(path) + "/") + animName;
			AnimationPrefab *const animPrefab = loadAnimationPrefab(jamTranslatePath(name.c_str()));
			g_animPrefabs[name] = animPrefab;
			animSetPrefab->animations[int(animType)] = animPrefab;
		}
	}

	return animSetPrefab;
}

void loadAnimationLibrary(const char* const filename) {
	char *txt = jamReadFile(filename, nullptr);
	cJSON* doc = cJSON_Parse(txt);

	if (!doc)
	{
		SDL_LogError(SDL_LOG_CATEGORY_APPLICATION, "Failed to load animation prefab: %s", cJSON_GetErrorPtr());
		SDL_assert(doc);
	}


	cJSON* const animationSets = cJSON_GetObjectItem(doc, "animationSets");
	SDL_assert(animationSets != nullptr);

	const int numAnimationSets = cJSON_GetArraySize(animationSets);
	for (int setIt = 0; setIt < numAnimationSets; ++setIt) {
		cJSON *const animNameObj = cJSON_GetArrayItem(animationSets, setIt);
		const std::string animSetName = animNameObj->valuestring;
		g_animSetPrefabs[animSetName] = loadAnimationSet(animSetName.c_str());
	}

	/// todo: aliasses
	
	/*LuaScript cfg;
	if (cfg.load(filename)) {
		printf("Loading animation library from '%s'\n", configFile.c_str());

		{
			auto animationSetList = cfg.getGlobal("animationSets");
			for (int i = 1, m = animationSetList.length(); i <= m; ++i) {
				const std::string animSetName = animationSetList[i];
				g_animSetPrefabs[animSetName] = loadAnimationSet(animSetName.c_str());
			}
		}

		{
			auto aliasList = cfg.getGlobal("aliases");
			for (int i = 1, m = aliasList.length(); i <= m; ++i) {
				auto aliasObj = aliasList[i];
				SDL_assert(aliasObj.isTable());
				SDL_assert(aliasObj.length() >= 2);

				const std::string src = aliasObj[1];
				for (int j = 2, lm = aliasObj.length(); j <= lm; ++j) {
					const std::string alias = aliasObj[j];
					g_animSetPrefabs[alias] = g_animSetPrefabs[src];
				}				
			}
		}
	} else {
		SDL_LogError(SDL_LOG_CATEGORY_APPLICATION, "Failed to load animation db.");
		SDL_assert(!"onoz");
	}*/
}

const AnimationSetPrefab* getAnimationSetPrefab(const char* const name) {
	return g_animSetPrefabs[name];
}

AnimationSet* g_animationSetList = nullptr;

AnimationSet* createAnimationSet(const AnimationSetPrefab *const prefab) {
	AnimationSet *const res = new AnimationSet;
	res->prefab = prefab;

	if (nullptr == g_animationSetList) {
		g_animationSetList = res;
	}
	else {
		res->intrusiveNext = g_animationSetList;
		g_animationSetList->intrusivePrev = res;
		g_animationSetList = res;
	}

	return res;
}

void updateAnimationSet(AnimationSet& animSet, const float dt) {
	while (animSet.playQueue.size() > 0) {
		const AnimationType currentType = animSet.playQueue.front();
		const AnimationPrefab* anim = animSet.prefab->animations[int(currentType)];
		if (!anim) {
			break;;
		}

		u32 curFrame = u32(animSet.currentTime * anim->framesPerSecond);

		if (curFrame >= u32(anim->frames.size())) {
			if (anim->playMode == AnimationPlayMode::Loop) {
				curFrame %= u32(anim->frames.size());
			}
			else {
				animSet.playQueue.pop();
				animSet.currentTime = 0.0f;
				continue;
			}
		}

		animSet.currentAnimationFrame = curFrame;
		animSet.currentAnimationType = currentType;
		animSet.currentTime += dt;
		break;
	}
}

void updateAnimations(float dt) {
	for (AnimationSet *it = g_animationSetList; it; it = it->intrusiveNext) {
		updateAnimationSet(*it, dt);
	}
}

