#pragma once
#include <array>
#include <queue>
#include <vector>

struct Sprite;


enum class AnimationType {
	Spawn,
	Idle,
	IdleNoWeapon,
	Move,
	MoveNoWeapon,
	Attack,
	Die,
	Explosion,
	NumItems,
};

enum class AnimationFrameLayer {
	Color,
	Pivot,
	Shadow,
	NumItems
};

enum class AnimationPlayMode {
	Loop,
	OneShot,
	NumItems
};

const char* const enumToString(const AnimationPlayMode& val);

struct AnimationFrame {
	std::array<Sprite*, int(AnimationFrameLayer::NumItems)> layers;
};

struct AnimationPrefab {
	std::vector<AnimationFrame>	frames;
	AnimationPlayMode			playMode = AnimationPlayMode::Loop;
	float						framesPerSecond = 5.0f;
};

struct AnimationSetPrefab {
	std::array<AnimationPrefab*, int(AnimationType::NumItems)> animations = {};
};

struct AnimationSet {
	const AnimationSetPrefab*	prefab = nullptr;
	std::queue<AnimationType>	playQueue;
	float						currentTime = 0.0f;

	// These are read-only
	AnimationType	currentAnimationType = AnimationType::NumItems;
	int				currentAnimationFrame = 0;

	AnimationSet *intrusiveNext = nullptr;
	AnimationSet *intrusivePrev = nullptr;


	void playAnimation(AnimationType animType) {
		while (!playQueue.empty()) {
			playQueue.pop();
		}

		playQueue.push(animType);
	}

	void enqueueAnimation(AnimationType animType) {
		playQueue.push(animType);
	}

	bool isActive() const {
		return currentAnimationType != AnimationType::NumItems;
	}

	const AnimationFrame& getCurrentFrame() const
	{
		return prefab->animations[int(currentAnimationType)]->frames[currentAnimationFrame];
	}

};

void						loadAnimationLibrary(const char* const filename);
const AnimationSetPrefab*	getAnimationSetPrefab(const char* const name);
AnimationSet*				createAnimationSet(const AnimationSetPrefab *const prefab);
void						updateAnimations(float dt);
