#include <Framework/Framework.h>
#include <Framework/PhysicsDebugDraw.h>
#include <Framework/PhysicsSystem.h>
#include <Framework/PrimitiveBatch.h>
#include <Framework/SoundLibrary.h>
#include <Framework/Sprite.h>
#include <Framework/TextureLibrary.h>

#include "Animation.h"
#include "BoatyMusic.h"
#include "CameraShake.h"
#include "Common.h"
#include "GridRenderer.h"
#include "ParticleSystem.h"
#include "Debris.h"
#include "Raft.h"
#include "RaftGen.h"
#include "RaftPartLibrary.h"
#include "RaftRenderer.h"
#include "RaftUI.h"
#include "WaterGrid.h"
#include "WaterWorld.h"
#include "WeaponSystem.h"

#include <functional>
#include <set>
#include <imgui.h>
#include <string>
#include <time.h>

#define BUTTON_A 0
#define BUTTON_B 1
#define BUTTON_X 2
#define BUTTON_Y 3
#define BUTTON_LEFT_BUMPER 4
#define BUTTON_RIGHT_BUMPER 5
#define BUTTON_SELECT 6
#define BUTTON_START 7
#define BUTTON_LEFT_STICK 8
#define BUTTON_RIGHT_STICK 9

enum class InputKey {
	Left,
	Right,
	Up,
	Down,
	Fire1,
	Fire2,
	Action,
	NumItems,
};

class BoatyGame : public IGame
{
public:

	WaterWorld waterWorld;

	vec2 cameraPosition;
	vec2 viewTranslation;
	SDL_Joystick * gameController = nullptr;
	Rafterizer rafterizer;
	DebrisSystem debrisSystem;

	static IGame* gameFactory() { return new BoatyGame; }

	WaterGrid waterGrid;
	WaterSimulator waterSimulator;
	WaterRenderer waterRenderer;
	BoatyMusic boatyMusic;
	GridRenderer gridRenderer;
	RaftPartLibrary raftPartLibrary = RaftPartLibrary("parts.json");
	RaftGen raftGen = RaftGen(&raftPartLibrary);
    
    Sprite* introImg = nullptr;
	Sprite* waterImg = nullptr;
	Sprite* peck0Img = nullptr;
	Sprite* peck1Img = nullptr;

	CameraShake shake;
	PlayerInputState playerInputState;
	//WeaponSystem weaponSystem;
	RaftUI raftUI;

	Debris* debrisTest1 = nullptr;
	Sprite* debrisSprite = nullptr;

	Flock flock1;

	float m_zoom = 1.0f;

	BoatyGame()
	{
		gridRenderer.startPoint = vec2(gridRenderer.gridSize.x / 2, -gridRenderer.gridSize.y/2);

		g_textureLibrary = new TextureLibrary();
		g_textureLibrary->defaultColorKey = 0;
		g_textureLibrary->defaultClip = false;
		g_textureLibrary->defaultRepeat = false;
		g_textureLibrary->load("textures.json");

		loadAnimationLibrary("animationLibrary.json");
        
		waterImg = loadSprite(g_textureLibrary->getFilename("water"), 0, false, true);
        introImg = loadSprite(g_textureLibrary->getFilename("introImg"), 0, false, false);
		peck0Img = loadSprite(g_textureLibrary->getFilename("peck0"), 0, false, false);
		peck1Img = loadSprite(g_textureLibrary->getFilename("peck1"), 0, false, false);

		debrisSprite = loadSprite(g_textureLibrary->getFilename("debris"), 0, false, false);
		debrisSystem.setSprite(debrisSprite);
		debrisSystem.spawnRandomDebris(200, 50, 1.0f);

		// Player raft
		{
			bool isPlayer = true;
			vec2 position = vec2(0.0f);
			//waterWorld.addRaft(raftGen.makeTestRaft(position, isPlayer), isPlayer);

			Raft* r = new Raft;
			r->addPart(ivec2(0, 0), g_raftPartLibrary->get("door"));
			r->addPayload(0, ivec2(0, 0), g_raftPartLibrary->get("survivor"));
			r->addPayload(0, ivec2(0, 1), g_raftPartLibrary->get("tomato"));
			r->createPhysics(position, isPlayer);
			waterWorld.addRaft(r, isPlayer);
		}

		if (SDL_NumJoysticks() > 0)
		{
			//Load joystick
			gameController = SDL_JoystickOpen(0);
			if (gameController)
			{
				printf("Found a controller\n");
			}
		}

		//debrisSystem.spawnDebrisAtPoint(vec2(10, 0));

		boatyMusic.load();

		// TODO: the grid should increase in size with resolution
        const int waterGridSize = 45;

		waterSimulator.init(waterGrid, waterGridSize);
		g_waterWorld->m_waterSimulator = &waterSimulator;
		g_waterWorld->m_waterGrid = &waterGrid;
		g_waterWorld->m_raftGen = &raftGen;


		setupIntroText();
	}

	void setupIntroText()
	{
		if (1)
		{
			RaftUI::StoryMessage item;
			item.sprite = g_textureLibrary->getSprite("StartScreen");
			raftUI.m_storyMessages.push_back(item);
		}
		if (1)
		{
			RaftUI::StoryMessage item;
			item.text = "After a radical and uneducated United States administration took power in the year 2017,\n"
			"efforts to stabilize climate change were halted, due to an ignorant belief that such a thing did not exist.\n"
			"In the distant future, the polar ice caps have completely melted, and the sea level has risen over 7,683m.";
			raftUI.m_storyMessages.push_back(item);
		}

		if (1)
		{
			RaftUI::StoryMessage item;
			item.text = "Nearly all land on Earth is submerged, except for Mexico, \n"
			"saved by the titanic wall built around the entire border.\n"
			"Hardly anyone survived \"The Great Melting\", and the last remnants of \n"
			"human civilization now live on ramshackle floating rafts, house boats, or ships.";
			raftUI.m_storyMessages.push_back(item);
		}

		if (1)
		{
			RaftUI::StoryMessage item;
			item.text = "While the exact location of Mexico is no longer known, many survivors\n"
			"are intent on walking once again on dry land, often turning on one another for the \n"
			"chance to start again in the new world.";
			raftUI.m_storyMessages.push_back(item);
		}

		if (1)
		{
			RaftUI::StoryMessage item;
			item.text = "You were captain of the prestigious \"Boaty McBoatFace\", originally a research vessel, \n"
			"but modified to survive in the harsh realities of present day post-apocalypse. Hearing a rumour that\n "
			"you had discovered the mythical location of Mexico, raiders attempted to board \"Boaty McBoatFace\", \n"
			"but the conflict ending up destroying it.";
			raftUI.m_storyMessages.push_back(item);
		}

		if (1)
		{
			RaftUI::StoryMessage item;
			item.text = "You awaken floating in the ocean, saved by a sturdy door, \n"
			"the only remaining piece of your magnificent ship. With steadfast resolve, \n"
			"and random trash floating in the ocean, you plan to rebuild what you have lost, \n"
			"and find your way to Mexico - last dry land on Earth.";
			raftUI.m_storyMessages.push_back(item);
		}
	}

	bool drawPhysicsDebug = false;
	bool inputsHeld[int(InputKey::NumItems)] = { false };

	virtual bool onEvent(const SDL_Event& e) override
	{
		mat2 physToRend = mat2(physCoordToGfx(vec2(1, 0)), physCoordToGfx(vec2(0, 1)));
		mat2 rendToPhys = glm::inverse(physToRend);
		rendToPhys[0] = glm::normalize(rendToPhys[0]);
		rendToPhys[1] = glm::normalize(rendToPhys[1]);

		vec2 moveVec = vec2(0, 0);
		bool handled = false;

		switch (e.type)
		{
			case SDL_KEYDOWN:
			case SDL_KEYUP:
			{
				InputKey input = InputKey::NumItems;
				switch (e.key.keysym.sym) {
				case SDLK_a:
				case SDLK_LEFT: 
					input = InputKey::Left; 
					break;
				case SDLK_RIGHT: 
				case SDLK_d:
					input = InputKey::Right; 
					break;
				case SDLK_UP: 
				case SDLK_w:
					input = InputKey::Up; 
					break;
				case SDLK_DOWN: 
				case SDLK_s:
					input = InputKey::Down; 
					break;
				case SDLK_x: 
					input = InputKey::Fire1; 
					break;
				case SDLK_z:
					input = InputKey::Fire2;
					break;
				case SDLK_SPACE:
					input = InputKey::Action;
					break;
				case SDLK_F1: if (e.key.state == SDL_RELEASED) drawPhysicsDebug = !drawPhysicsDebug; break;
				case SDLK_F2: if (e.key.state == SDL_RELEASED) raftUI.m_showAvailableParts = !raftUI.m_showAvailableParts; break;
				}
				if (input != InputKey::NumItems) {
					inputsHeld[int(input)] = e.key.state == SDL_PRESSED;
				}
				break;
			}
		}

		moveVec.x += inputsHeld[int(InputKey::Right)] ? 1.0f : 0.0f;
		moveVec.x -= inputsHeld[int(InputKey::Left)] ? 1.0f : 0.0f;
		moveVec.y += inputsHeld[int(InputKey::Up)] ? 1.0f : 0.0f;
		moveVec.y -= inputsHeld[int(InputKey::Down)] ? 1.0f : 0.0f;
		moveVec = rendToPhys * moveVec;
		if (length(moveVec) > 1.0f) moveVec = normalize(moveVec);

		playerInputState.fire1 = inputsHeld[int(InputKey::Fire1)];
		playerInputState.fire2 = inputsHeld[int(InputKey::Fire2)];

		int mouseX, mouseY;
		SDL_GetMouseState(&mouseX, &mouseY); 

		int windowX, windowY;
		SDL_GetWindowSize(g_window, &windowX, &windowY);

		vec2 targetPos = gfxCoordToPhys(vec2(mouseX, windowY - mouseY - 1) + viewTranslation);
		//vec2 dir = glm::normalize(targetPos - waterWorld.m_playerRaft->getPhysicsPosition());

		playerInputState.shootTarget = targetPos;
		playerInputState.moveDir = moveVec;

		if (inputsHeld[(int)InputKey::Action]) {
			playerInputState.action = true;
		}

		return handled;
	}

	void handleJoystickInput(float dt)
	{
		if (SDL_NumJoysticks() > 0)
		{
			SDL_JoystickUpdate();

			// right stick
			float faster = 2;
			{
				Sint16 joyaxis_x = SDL_JoystickGetAxis(gameController, 4);
				Sint16 joyaxis_y = SDL_JoystickGetAxis(gameController, 3);

				float speed = 0.3f;
				if (SDL_JoystickGetButton(gameController, BUTTON_RIGHT_BUMPER)) // right bumper
					speed *= faster;

				if (joyaxis_x < -8000) cameraPosition.x -= speed * dt;
				if (joyaxis_x > 8000) cameraPosition.x += speed * dt;
				if (joyaxis_y < -8000) cameraPosition.y += speed * dt;
				if (joyaxis_y > 8000) cameraPosition.y -= speed * dt;
			}

			{
				Sint16 joyaxis_x = SDL_JoystickGetAxis(gameController, 0);
				Sint16 joyaxis_y = SDL_JoystickGetAxis(gameController, 1);

				float speed = 0.1f;
				if (SDL_JoystickGetButton(gameController, BUTTON_RIGHT_BUMPER)) // right bumper
					speed *= faster;

				mat2 physToRend = mat2(physCoordToGfx(vec2(1, 0)), physCoordToGfx(vec2(0, 1)));
				mat2 rendToPhys = glm::inverse(physToRend);
				rendToPhys[0] = glm::normalize(rendToPhys[0]);
				rendToPhys[1] = glm::normalize(rendToPhys[1]);

				vec2 moveVec = vec2(0, 0);
				if (joyaxis_x < -8000) moveVec.x -= speed * dt;
				if (joyaxis_x > 8000) moveVec.x += speed * dt;
				if (joyaxis_y < -8000) moveVec.y += speed * dt;
				if (joyaxis_y > 8000) moveVec.y -= speed * dt;
				moveVec = rendToPhys * moveVec;

				//if (length(moveVec) > 1.0f) 
				//	moveVec = normalize(moveVec);

				playerInputState.moveDir = moveVec;
			}

			// sdl is weird... axis 2 is the triggers (both of them) so decode here
			Sint16 trigger = SDL_JoystickGetAxis(gameController, 2);
			Sint16 left = 0;
			Sint16 right = 0;

			if (trigger != -1)
			{
				if (trigger < 0)
					right = -trigger;
				if (trigger > 0)
					left = trigger;
			}

			float joyaxis_x = SDL_JoystickGetAxis(gameController, 4) / 32767.0f;
			float joyaxis_y = SDL_JoystickGetAxis(gameController, 3) / 32767.0f;

			float angle = atan2(joyaxis_y, joyaxis_x);
			vec2 dir = vec2(sin(angle), cos(angle));

			playerInputState.action = !!SDL_JoystickGetButton(gameController, BUTTON_A);
			playerInputState.shootTarget = dir * 123456789.0f;
			playerInputState.fire1 = left > 800;
			playerInputState.fire2 = right > 800;

			if(!raftUI.m_showAvailableParts)
				raftUI.m_showAvailableParts = !!SDL_JoystickGetButton(gameController, BUTTON_START);
			else
				raftUI.m_showAvailableParts = !SDL_JoystickGetButton(gameController, BUTTON_B);
		}
	}

	void update(float dt) override
	{

		boatyMusic.playMusic();
		handleJoystickInput(dt);

		waterSimulator.step(waterGrid);

		auto prim = g_primitiveRenderer;
		auto ctx = Gfx_FrameBegin(g_device);

		ivec2 drawableSize;
		SDL_GL_GetDrawableSize(g_window, &drawableSize.x, &drawableSize.y);

		{
			GfxRect scissor = {};
			scissor.right = drawableSize.x;
			scissor.bottom = drawableSize.y;
			Viewport viewport(0, 0, drawableSize.x, drawableSize.y);

			Gfx_SetViewport(ctx, viewport);
			Gfx_SetScissorRect(ctx, scissor);
		}

		Gfx_Clear(ctx, ColorRGBA(1.0f, 1.0f, 1.0f));


		m_zoom = 1.0f + (float)waterWorld.m_playerRaft->m_parts.size() / 20.0f;
		m_zoom = glm::min(m_zoom, 2.0f);

		aabb2 viewBounds;
		viewBounds.min = vec2(0.0f);
		viewBounds.max = vec2(drawableSize) * m_zoom;

		vec2 offset;
		if (shake.running())
		{
			offset = shake.offset;
		}

		shake.update(dt);

		viewTranslation = (vec2(viewBounds.dimensions()) * vec2(-0.5f, -0.3f) + cameraPosition) + offset;
		viewBounds.translate(viewTranslation);
		prim->begin2d(viewBounds);


		gridRenderer.drawTiles(ctx, prim);

		vec2 waterGridOffset = gfxCoordToPhys(viewTranslation/* + vec2(drawableSize.x * -0.5f, drawableSize.y * 0.5f)*/);
		waterGridOffset -= glm::fract(waterGridOffset) + vec2(2);
        waterRenderer.renderMesh(prim, waterGrid, waterImg, waterSimulator.getTimeIndex(), waterGridOffset, m_zoom);
        waterRenderer.renderControls(prim, waterGrid, waterSimulator.getTimeIndex());

		BlendStateHandle blendState_alpha = Gfx_BlendState_Create(g_device, BlendStateDescr::make_alphablend());
		Gfx_SetBlendState(ctx, blendState_alpha);

		if (!raftUI.wantPauseSimulation())
		{
			WorldUpdateInfo worldUpdateInfo;
			worldUpdateInfo.playerInputState = playerInputState;
			worldUpdateInfo.dt = dt;
			worldUpdateInfo.shake = &shake;
			waterWorld.update(viewBounds, worldUpdateInfo);
		}

		updateAnimations(dt / 1000.f);
		waterWorld.render(rafterizer, debrisSystem);
		prim->flush();

		if (drawPhysicsDebug)
		{
			g_physicsSystem->physicsDebugDraw->offset = vec2(0, 0);
			g_physicsSystem->physicsDebugDraw->prim = prim;
			g_physicsSystem->physicsWorld->DrawDebugData();
		}

		if (waterWorld.m_playerRaft)
		{
			debrisSystem.spawnDebris(waterWorld.m_playerRaft->getPhysicsPosition(), 1.0f);

			// more complex way of finding camera position because we can't 'just' use the first part
			// since it could've died
			vec2 center; 
			float count = 0;
			for (auto part : waterWorld.m_playerRaft->m_parts)
			{
				if (part.alive())
				{
					center += part.getRenderPosition();
					count++;
				}
			}
			center /= count;

			cameraPosition = glm::mix(cameraPosition, center, 0.05f);
		}

		//prim->end2d();
		//prim->begin2d(drawableSize.x, drawableSize.y);

		waterWorld.renderFlocks(viewBounds, prim, peck0Img, peck1Img);

		raftUI.selectRaft(waterWorld.m_playerRaft);
		raftUI.update(playerInputState);
		raftUI.renderWorldSpace();

		prim->end2d();

		
		prim->begin2d(drawableSize.x, drawableSize.y);
		
		raftUI.renderScreenSpace();
		prim->end2d();
        

        const bool showIntro = false;
        if (showIntro)
        {
            aabb2 viewBoundsIntro;
            viewBoundsIntro.min = vec2(0.0f);
            viewBoundsIntro.max = vec2(drawableSize);
            
            prim->begin2d(viewBoundsIntro);
            drawSprite(prim, *introImg, vec2(0,0));
            prim->end2d();
        }

		// waterSimulator.edit(waterGrid);
		// updateDebugMenu();

		playerInputState.action = false;
	}

	void updateDebugMenu()
	{
		return;
		ImGui::SetNextWindowCollapsed(true, ImGuiSetCondition_FirstUseThisSession);
		ImGui::Begin("Debug");
		ImGui::SliderFloat("Zoom", &m_zoom, 0.01f, 10.0f);
		ImGui::End();
	}
};

int main(int argc, char *argv[])
{
	g_physicsSystem = new PhysicsSystem(vec2(0, 0), 1.0f);
	g_physicsSystem->physicsDebugDraw->transform[0] = physCoordToGfx(vec2(1, 0));
	g_physicsSystem->physicsDebugDraw->transform[1] = physCoordToGfx(vec2(0, 1));

	PixelSize = 1;
	printf("Starting main\n");

	srand((uint)time(0));

	JamConfig cfg;
	cfg.name = "Boaty McDoorFace (tm)";
	cfg.dataDirectory = "";
	cfg.width = 1280;
	cfg.height = 720;
	cfg.argc = argc;
	cfg.argv = argv;
	cfg.resizable = true;
	cfg.gameFactory = BoatyGame::gameFactory;

	return jamMainLoop(cfg);
}
