#include "Debris.h"
#include "Raft.h"
#include "RaftPartLibrary.h"
#include "WaterWorld.h"

#include <random>
#include <algorithm>

#include <Framework/Framework.h>
#include <Framework/PhysicsSystem.h>

void Debris::createPhysics(vec2 position)
{
	collisionCallback = std::bind(&Debris::onCollision, this, std::placeholders::_1);

	m_body = ::createPhysicsBody(ColliderDesc::makeBox(m_size.x * 0.5f, m_size.y * 0.5f), position + (vec2)m_offset);
	//m_body->SetLinearDamping(0.0f);
	m_body->SetLinearDamping(0.93f);
	m_body->SetUserData(this);
}

bool Debris::onCollision(BoatyBase* collector)
{
	if (auto raftPart = dynamic_cast<RaftPart*>(collector)) {
		auto blueprint = g_raftPartLibrary->getRandom();

		if (raftPart->m_isPlayer) {
			g_waterWorld->collectPart(blueprint);
			m_debrisSystem->playPickupSound();
		}

		m_debrisSystem->removeDebris(this); // deletes itself!.... don't reference this instance after :) trollolol
		return true; // destroy
	} else {
		return false;
	}
}

DebrisSystem::DebrisSystem()
{
	m_pickupSound = jamLoadSoundClip("Sounds/woodCrunch.wav");
}

DebrisSystem::~DebrisSystem()
{
	for (auto debris : m_activeDebris)
	{
		delete debris;
	}

	m_activeDebris.clear();
}

void DebrisSystem::spawnRandomDebris(int gridSize, int amount, float minimumAllowableDistance)
{
	std::random_device rd;
	std::mt19937 e2(rd());
	std::uniform_real_distribution<> distX(-gridSize * 0.5f, gridSize * 0.5f);
	std::uniform_real_distribution<> distY(-gridSize * 0.5f, gridSize * 0.5f);

	std::vector<vec2> keepers;

	keepers.push_back(vec2(distX(e2), distY(e2)));

	for (int k = 2; k <= amount; ++k)
	{
		float thisX = float(distX(e2));
		float thisY = float(distY(e2));

		float minDistance = std::numeric_limits<float>::max();
		for (auto& keeper : keepers)
		{
			const float testX = thisX - keeper.x;
			const float testY = thisY - keeper.y;

			const float distance = sqrtf(testX*testX + testY*testY);
			minDistance = std::min(minDistance, distance);
		}

		if (minDistance >= minimumAllowableDistance)
		{
			keepers.push_back(vec2(thisX, thisY));
		}
	}

	for (auto& keeper : keepers)
	{
		spawnDebrisAtPoint(keeper);
	}
}


void DebrisSystem::spawnDebris(vec2 playerPos, float minimumAllowableDistance)
{
	float spawnWindowSize = 50.0f;
	float noSpawnRadius = 25.0f;
	float removeRadiusSize = 100.0f;

	std::random_device rd;
	std::mt19937 e2(rd());
	std::uniform_real_distribution<> distX(-spawnWindowSize, spawnWindowSize);
	std::uniform_real_distribution<> distY(-spawnWindowSize, spawnWindowSize);

	const int targetDebris = 50;

	for (int i = int(m_activeDebris.size()) - 1; i >= 0; --i) {
		if (length(playerPos - m_activeDebris[i]->getPhysicalPosition()) > removeRadiusSize) {
			g_physicsSystem->physicsWorld->DestroyBody(m_activeDebris[i]->m_body);
			delete m_activeDebris[i];
			//despawnRaft(m_rafts[i]);
			m_activeDebris[i] = m_activeDebris.back();
			m_activeDebris.resize(m_activeDebris.size() - 1);
		}
	}

	if (this->m_activeDebris.size() < targetDebris)
	{
		for (int retry = 0; retry < 10; ++retry) {
			vec2 pos(distX(e2), distY(e2));
			//if (pos.x > -noSpawnWindowSize && pos.x < noSpawnWindowSize) continue;
			//if (pos.y > -noSpawnWindowSize && pos.y < noSpawnWindowSize) continue;
			if (length(pos) < noSpawnRadius) continue;

			spawnDebrisAtPoint(pos + playerPos);
			break;
		}
	}
}

void DebrisSystem::playPickupSound()
{
	jamPlaySound(m_pickupSound);
}

Debris* DebrisSystem::spawnDebrisAtPoint(vec2 point)
{
	auto debris = new Debris(ivec2(1, 1), ivec2(0, 0), this);
	debris->createPhysics(point);
	m_activeDebris.push_back(debris);
	return debris;
}

void DebrisSystem::removeDebris(Debris* debris)
{
	assert(debris);

	// Does not destroy physics (should already be done)
	auto iter = std::find(m_activeDebris.begin(), m_activeDebris.end(), debris);
	if (iter != m_activeDebris.end())
	{
		m_activeDebris.erase(iter);
	}

	//delete debris;
}

const std::vector<Debris*>& DebrisSystem::activeDebris() const
{
	return m_activeDebris;
}