#include "Raft.h"
#include "RaftPartLibrary.h"
#include "WaterWorld.h"

#include <Framework/Framework.h>
#include <Framework/PhysicsSystem.h>
#include <Framework/TextureLibrary.h>
#include <Framework/Sprite.h>

RaftPart::RaftPart(RaftPartBlueprint* blueprint, ivec2 offset, Raft* parent)
	: m_blueprint(blueprint)
	, m_offset(offset)
	, m_parent(parent)
{
	m_sprite = cloneSprite(g_textureLibrary->getSprite(blueprint->texture));
}

float RaftPart::addDestruction(ivec2 origin, int radius, bool stealAlpha)
{
	auto& pixels = m_sprite->pixels;

	Sprite *baseSprite = nullptr;
	if(stealAlpha)
		baseSprite = g_textureLibrary->getSprite(m_blueprint->texture);

	for (int y = -radius; y <= radius; y++)
	{
		int ky = origin.y + y;
		if (ky >= 0 && ky < m_sprite->h)
		{
			for (int x = -radius; x <= radius; x++)
			{
				if (x*x + y*y <= radius*radius)
				{
					int kx = origin.x + x;

					if (kx >= 0 && kx < m_sprite->w)
					{
						int idx = kx + ky * m_sprite->w;
						if(stealAlpha)
							pixels[idx].a = baseSprite->pixels[idx].a;
						else
							pixels[idx].a = 0;
					}
				}
			}
		}
	}

	Gfx_Texture_Update(g_device, m_sprite->texture, pixels.data());

	return (pixelsCovered / ((float)m_sprite->w * m_sprite->h));
}

void RaftPart::createPhysics(vec2 position, bool isPlayer)
{
	if (!isPlayer)
	{
		// graham: We want a collision callback to influence the combat AI (not players)
		collisionCallback = std::bind(&RaftPart::onCollision, this, std::placeholders::_1);
	}

	m_isPlayer = isPlayer;
	unsigned short collisionMask = (isPlayer ? COLLISION_MASK_PLAYER : COLLISION_MASK_NPC);
	unsigned short category = (COLLISION_MASK_PLAYER | COLLISION_MASK_NPC); // collised with player & npc
	m_body = ::createPhysicsBody(ColliderDesc::makeBox(size().x * 0.5f, size().y * 0.5f), position + (vec2)m_offset - offsetFromPhysCenterToGrid(), collisionMask, category);
	m_body->SetLinearDamping(0.93f);
	m_body->SetUserData(this);

	m_weapons.clear();
	for (auto& pload : payload) {
		if (pload.m_blueprint->weapon == "shotgun") {
			m_weapons.push_back(RaftWeapon());
			m_weapons.back().offset = offsetFromPhysCenterToGrid() + vec2(pload.m_offset - m_offset);
		}
	}
}

bool RaftPart::onCollision(BoatyBase* other)
{
	auto otherRaftPart = dynamic_cast<RaftPart*>(other);

	if (otherRaftPart == nullptr) // Ignore non-raft parts [i.e. debris]
		return false; // Don't destroy

	if (otherRaftPart->m_isPlayer && m_isPlayer) // ignore player self collisions)
		return false; // Don't destroy

	if (!otherRaftPart->m_isPlayer && !m_isPlayer) // ignore ai self collisions)
		return false; // Don't destroy

	assert(otherRaftPart->m_isPlayer);
	assert(!m_isPlayer);
	assert(m_parent->m_ai.enabled);
	m_parent->m_ai.onCollision(otherRaftPart);

	g_waterWorld->playSmashSound();

	return false; // Don't destroy
}

ivec2 RaftPart::size() const
{
	return m_blueprint->size;
}

b2Joint* Raft::createLinkPhysics(const RaftLink& link)
{
	b2DistanceJointDef jointDef;
	jointDef.bodyA = m_parts[link.part0].m_body;
	jointDef.bodyB = m_parts[link.part1].m_body;
	jointDef.collideConnected = true;
	vec2 part0Center = vec2(m_parts[link.part0].m_offset) + vec2(m_parts[link.part0].size()) * 0.5f;
	vec2 part1Center = vec2(m_parts[link.part1].m_offset) + vec2(m_parts[link.part1].size()) * 0.5f;
	jointDef.length = length(part0Center - part1Center);
	jointDef.dampingRatio = 0.5f;
	jointDef.frequencyHz = 2.0f;
	b2Joint* joint = g_physicsSystem->physicsWorld->CreateJoint(&jointDef);
	return joint;
}

void Raft::findLinks()
{
	m_links.clear();

	// Create links between neighboring parts
	std::set<uint> otherParts;
	for (uint i0 = 0; i0 < m_parts.size(); ++i0) {
		auto& part0 = m_parts[i0];
		otherParts.clear();

		iterRectOutline(
			part0.size() + 2,
			part0.m_offset - 1,
			[&](int x, int y)
		{
			if (rafter.test(x, y, Rafter::Bit::Platform, Rafter::Bit(0)) || rafter.test(x, y, Rafter::Bit::Attachment, Rafter::Bit(0))) {
				otherParts.insert(rafter.getPlatformId(x, y));
			}
		});

		for (uint i1 : otherParts) {
			if (i0 > i1) {
				link(i0, i1);
			}
		}
	}
}

void Raft::createPhysics(vec2 position, bool isPlayer)
{
	m_isPlayer = isPlayer;

	for (RaftPart& part : m_parts) {
		part.createPhysics(position, isPlayer);
	}

	for (RaftLink& link : m_links) {
		m_joints.push_back(createLinkPhysics(link));
	}
}

void Raft::createAI()
{
	initializeRaftAI(this, m_ai);
}

void Raft::move(vec2 dir)
{
	if(!m_isPlayer)
	{
		float moveLen = length(dir);
		assert(moveLen == moveLen);
		if (moveLen > 1) {
			int k = 0;
		}
	}

	/*for (RaftPart& part : m_parts) {
		part.m_body->ApplyLinearImpulse(dir, part.m_body->GetPosition(), true);
		//part.m_body->SetLinearVelocity(dir);
	}*/

	/*int idx = getPrimaryPartIdx();
	if (idx >= 0)
	{
		m_parts[idx].m_body->ApplyLinearImpulse(dir, m_parts[idx].m_body->GetPosition(), true);
	}
	*/

	for (RaftPart& part : m_parts) {
		if (part.alive()) {
			float thrust = 0.0f;
			for (RaftPart& payl : part.payload) {
				thrust += payl.m_blueprint->thrust;
			}

			if (thrust > 0) {
				part.m_body->ApplyLinearImpulse(dir * thrust, part.m_body->GetPosition(), true);
			}
		}
		//part.m_body->SetLinearVelocity(dir);
	}
}

bool Raft::isLinked(const RaftPart &part)
{
	if (part.m_body == nullptr)
		return false;

	if (m_parts.size() > 0 && &m_parts[0] == &part) {
		return true;
	}

	for (RaftLink& link : m_links)
	{
		// need to check if both objects in the link are alive
		if (m_parts[link.part0].alive() && m_parts[link.part1].alive())
		{
			// and make sure that we're the link we're looking for
			if (m_parts[link.part0].m_body == part.m_body)
			{
				// check if the other part of the 'link' appears in the joint list (eg. if the joint is still alive)
				b2JointEdge* current = part.m_body->GetJointList();

				while (current)
				{
					if (current->other == m_parts[link.part1].m_body)
						return true;
					current = current->next;
				}
			}

			if (m_parts[link.part1].m_body == part.m_body)
			{
				b2JointEdge* current = part.m_body->GetJointList();

				while (current)
				{
					if (current->other == m_parts[link.part0].m_body)
						return true;
					current = current->next;
				}
			}
		}
	}

	return false;
}
void Raft::shoot(vec2 target)
{
	for (auto& part : m_parts) {
		if (part.alive()) {
			for (auto& w : part.m_weapons) {
				if (w.shoot()) {
					g_waterWorld->m_weaponSystem.shotgun(
						part.getPhysicalPosition() + w.offset,
						target, 0.1f, 1, m_isPlayer);
				}
			}
		}
	}
}

bool Raft::hasWeapon() const
{
	for (auto& part : m_parts) {
		if (part.alive()) {
			if (part.m_weapons.size() > 0)
				return true;
		}
	}

	return false;
}

vec2 RaftPart::getRenderPosition() const {
	return physCoordToGfx(getPhysicalPosition()) + g_waterWorld->getObjectWaveSway(getPhysicalPosition()) * m_swayStrength;
}

void Raft::update()
{
	for (RaftPart& part : m_parts)
	{
		if (part.alive()) {
			for (RaftPart& pload : part.payload) {
				part.doRegen(pload.m_blueprint->regen / 60);
			}
		}
	}

	for (RaftPart& part : m_parts)
	{
		for (auto& w : part.m_weapons)
			w.update();
	}

	// copy physical position from the physics system
	for (RaftPart& part : m_parts)
	{
		if (part.m_body)
		{
			if (part.getHealth() <= 0) {
				g_physicsSystem->physicsWorld->DestroyBody(part.m_body);
				part.m_body = nullptr;

				// TODO: nuke this part
				part.m_sprite = nullptr;
			} else {
				part.m_physicalPosition = part.m_body->GetPosition();
			}
		}
	}

	for (RaftPart& part : m_parts)
	{
		if (part.alive() && part.m_body) {
			b2MassData md;
			part.m_body->GetMassData(&md);
			float massScale = part.totalMass() / md.mass;
			md.mass *= massScale;
			md.I *= massScale;
			part.m_body->SetMassData(&md);
		}
	}

	if (m_ai.enabled)
	{
		auto playerRaft = g_waterWorld->m_playerRaft;

		// Can this raft see the player?
		bool canSeePlayer = canSeeEachOther(playerRaft, m_ai.raft);

		if (m_ai.machine.getCurrentState()->getId() == "observing")
		{
			if (canSeePlayer)
				m_ai.machine.nextState("spotted_enemy");
		}
		else if (m_ai.machine.getCurrentState()->getId() == "wait")
		{
			m_ai.remainingMoves--;
			if (m_ai.remainingMoves <= 0)
			{
				m_ai.machine.nextState(m_ai.pendingState);
				m_ai.pendingState.clear();
			}
		}
		else if (m_ai.machine.getCurrentState()->getId() == "combat")
		{
			if (canSeePlayer)
			{
				// Do I have a weapon?
				const bool hasWeapon = m_ai.raft->hasWeapon();
				if (hasWeapon)
				{
					m_ai.machine.nextState("pewpew");
				}
				else
				{
					// No weapon, no worries. RAMMING SPEED
					m_ai.machine.nextState("ram");
				}
			}
			else
			{
				m_ai.machine.nextState("lost_enemy");
			}
		}
		else if (m_ai.machine.getCurrentState()->getId() == "pewpew")
		{
			if (canSeePlayer)
			{
				m_ai.raft->shoot(playerRaft->getPhysicsPosition());
			}
			else
			{
				m_ai.machine.nextState("patrol");
			}
			
		}
		else if (m_ai.machine.getCurrentState()->getId() == "ramming")
		{
			if (canSeePlayer)
			{
				auto playerRaft = g_waterWorld->m_playerRaft;
				m_ai.lastKnownPlayerPosition = playerRaft->getPhysicsPosition();

				//float maxSpeed = 1.1;
				//vec2 wantedVelocity = maxSpeed * normalize(m_ai.lastKnownPlayerPosition - getPhysicsPosition());
				//m_ai.raft->move(wantedVelocity - normalize(vec2(getPhysicsBody()->GetLinearVelocity())));

				m_ai.raft->move(1.4 * glm::normalize(m_ai.lastKnownPlayerPosition - m_ai.raft->getPhysicsPosition()));
			}
			else
			{
				m_ai.machine.nextState("patrol");
			}
		}
		else if (m_ai.machine.getCurrentState()->getId() == "reverse")
		{
			float maxSpeed = 1;
			vec2 wantedVelocity = maxSpeed * normalize(m_ai.lastKnownPlayerPosition - getPhysicsPosition());
			//m_ai.raft->move(-(wantedVelocity - normalize(vec2(getPhysicsBody()->GetLinearVelocity()))));
			m_ai.raft->move(normalize(-(m_ai.lastKnownPlayerPosition - m_ai.raft->getPhysicsPosition()*0.5f)));
			m_ai.remainingMoves--;
			if (m_ai.remainingMoves <= 0)
			{
				m_ai.machine.nextState("done");
			}
		}
		else if (m_ai.machine.getCurrentState()->getId() == "exploring")
		{
			if (canSeePlayer)
				m_ai.machine.nextState("spotted_enemy");
			else
			{
				m_ai.wanderTarget += vec2(randomUnitFloat() * 2 - 1, randomUnitFloat() * 2 - 1) * m_ai.wanderJitter;
				m_ai.wanderTarget = normalize(m_ai.wanderTarget) * m_ai.wanderRadius;

				vec2 targetLocal = m_ai.wanderTarget + vec2(m_ai.wanderDistance, 0);
				if (getPhysicsBody())
				{
					vec2 targetWorld = getPhysicsBody()->GetWorldPoint(targetLocal);
					// todo: move targetlocal in a better direction
					vec2 dir = targetWorld - getPhysicsBody()->GetPosition();
					m_ai.raft->move(0.5 * normalize(dir));
				}
			}
		}
		else if (m_ai.machine.getCurrentState()->getId() == "patrolling")
		{
			if (canSeePlayer)
				m_ai.machine.nextState("spotted_enemy");
			else
			{
				// Keep moving towards goal! Come on little buddy, you can do it!
				m_ai.raft->move(glm::normalize(glm::normalize(m_ai.patrolFinish - m_ai.patrolStart)));
				
				m_ai.remainingMoves--;
				
				if (m_ai.remainingMoves <= 0)
				{
					if (m_ai.remainingAttempts > 0)
					{
						m_ai.remainingMoves = 10 * 60 /*hz*/; // sec
						vec2 temp = m_ai.patrolStart;
						m_ai.patrolStart = m_ai.patrolFinish;
						m_ai.patrolFinish = temp;
						m_ai.remainingAttempts--;
					}
					else
					{
						// Give up! Loser
						m_ai.machine.nextState("explore");
					}
				}
			}
		}
	}
}

void Raft::replaceWith(const Raft& templateRaft)
{
	vec2 primaryPartPosition = m_parts[0].getPhysicalPosition();
	bool isPlayer = m_parts[0].m_isPlayer;

	m_links.clear();

	// all joints on a body are destroyed once a body is destroyed
	m_joints.clear();

	for (RaftPart& it : m_parts) {
		if(it.m_body) // don't double delete
		g_physicsSystem->physicsWorld->DestroyBody(it.m_body);
	}

	m_parts.clear();

	m_parts = templateRaft.m_parts;
	m_links = templateRaft.m_links;
	rafter = templateRaft.rafter;

	for (RaftPart& it : m_parts) {
		it.m_swayStrength = 1.0f;
		it.m_parent = this;
	}

	createPhysics(primaryPartPosition, isPlayer);
}

/*
void Raft::addPartAndCreatePhysics(int parentIdx, const RaftPart& part)
{
	SDL_assert(part.m_blueprint->type == RaftPartType::Attachment || part.m_blueprint->type == RaftPartType::Platform);

	RaftLink newLink;
	newLink.part0 = parentIdx;
	newLink.part1 = (u32)m_parts.size();

	m_parts.push_back(part);
	const RaftPart& parentPart = m_parts[parentIdx];
	m_parts.back().createPhysics(parentPart.getPhysicalPosition() + (vec2)part.m_offset, parentPart.m_isPlayer);

	m_links.push_back(newLink);
	m_joints.push_back(createLinkPhysics(newLink));
}
*/

const char* enumToString(RaftPartType type)
{
	switch (type)
	{
	case RaftPartType::Platform: return "Platform";
	case RaftPartType::Attachment: return "Attachment";
	case RaftPartType::Payload: return "Payload";
	default: return "unknown";
	}
}
