#pragma once

#include <Framework/Framework.h>
#include <Framework/Math.h>
#include <Box2D/Box2D.h>
#include "Common.h"
#include "RaftAI.h"

#include <vector>
#include <string>

struct Sprite;
struct RaftPartBlueprint;

const char* enumToString(RaftPartType t);

/*struct RaftPayload {
	Sprite* m_sprite = nullptr;
	ivec2 m_offset = ivec2(0, 0);
};*/

struct Rafter {
	enum class Bit {
		Platform = 1 << int(RaftPartType::Platform),
		Payload = 1 << int(RaftPartType::Payload),
		Attachment = 1 << int(RaftPartType::Attachment),
		PlatformOrPayload = Platform | Payload,
		PlatformOrAttachment = Platform | Attachment,
	};

	static Bit bit(RaftPartType t) {
		return Bit(1 << int(t));
	}

	enum { HalfSize = 9 };
	u8 bitmap[HalfSize * 2 + 1][HalfSize * 2 + 1];
	u16 platformIdMap[HalfSize * 2 + 1][HalfSize * 2 + 1];

	Rafter() {
		clear();
	}

	void clear() {
		for (int y = 0; y < HalfSize * 2 + 1; ++y)
			for (int x = 0; x < HalfSize * 2 + 1; ++x)
				bitmap[y][x] = false;
	}

	bool test(int x, int y, Bit reqBit, Bit exclBit) const {
		if (x < -HalfSize || x > HalfSize) return true;
		if (y < -HalfSize || y > HalfSize) return true;
		u8 cur = bitmap[y + HalfSize][x + HalfSize];
		return (cur & u8(reqBit)) == u8(reqBit) && (cur & u8(exclBit)) == 0;
	}

	void set(int x, int y, Bit bit) {
		if (x < -HalfSize || x > HalfSize) return;
		if (y < -HalfSize || y > HalfSize) return;
		bitmap[y + HalfSize][x + HalfSize] |= u8(bit);
	}

	void setPlatformId(int x, int y, u16 id) {
		if (x < -HalfSize || x > HalfSize) return;
		if (y < -HalfSize || y > HalfSize) return;
		platformIdMap[y + HalfSize][x + HalfSize] = id;
	}

	u16 getPlatformId(int x, int y) const {
		if (x < -HalfSize || x > HalfSize) return 0;
		if (y < -HalfSize || y > HalfSize) return 0;
		return platformIdMap[y + HalfSize][x + HalfSize];
	}
};

struct RaftWeapon {
	float cooldown = 0.0f;
	float fireRate = 1.0f;
	vec2 offset = vec2(0, 0);

	void update() {
		cooldown -= 1.f / 60;
	}

	bool shoot() {
		if (cooldown <= 0.0f) {
			cooldown = 1.f / fireRate;
			return true;
		} else {
			return false;
		}
	}
};

struct RaftPart : BoatyBase
{
	vec2 m_physicalPosition;
	float m_swayStrength = 1.0f;

	std::vector<RaftPart> payload;
	RaftPartBlueprint* m_blueprint;
	std::vector<RaftWeapon> m_weapons;

	ivec2 size() const;
	bool alive() const { return getHealth() > 0.0f; }

	float totalMass() const {
		float res = m_blueprint->mass;
		for (auto& p : payload) {
			res += p.m_blueprint->mass;
		}
		return res;
	}

	// Defined in terms of tiles or (0.5, 1) doors.
	//ivec2 m_size = ivec2(1, 1);
	ivec2 m_offset = ivec2(0, 0);

	b2Body* m_body = nullptr;
	Sprite* m_sprite = nullptr;
	struct Raft* m_parent = nullptr;

	bool m_isPlayer = false;
	int pixelsCovered = 0;

	void createPhysics(vec2 position, bool isPlayer);
	bool onCollision(BoatyBase* collector);

	vec2 getPhysicalPosition() const {
		return m_physicalPosition;
	}

	vec2 getRenderPosition() const;

	vec2 offsetFromPhysCenterToGrid() const {
		return vec2(size()) * -0.5f + 0.5f;
	}

	int area() const {
		return size().x * size().y;
	}
	bool m_spriteIsShared = true;
	float addDestruction(ivec2 origin, int radius, bool stealAlpha);

	virtual void damageCallback(float h, HealthInfo info) override
	{
		addDestruction(info.damageLocation, glm::mix(20, 40, h), false);
	}

	virtual void regenCallback(float h, HealthInfo info) override
	{
		// undo destruction
		addDestruction(info.damageLocation, glm::mix(20, 40, h), true);
	}

	RaftPart(RaftPartBlueprint* blueprint, ivec2 offset, Raft* parent);
};

struct RaftLink {
	u32 part0;
	u32 part1;
};

struct Raft {
	std::vector<RaftPart> m_parts;
	std::vector<RaftLink> m_links;
	std::vector<b2Joint*> m_joints;
	Rafter rafter;
	RaftAI m_ai;
	bool m_isPlayer = false;

	b2Joint* createLinkPhysics(const RaftLink& link);
	void findLinks();
	void createPhysics(vec2 position, bool isPlayer);
	void createAI();

	void move(vec2 dir);
	void shoot(vec2 target);
	void link(u32 a, u32 b) { m_links.push_back(RaftLink{ a, b }); }

	bool hasWeapon() const;

	vec2 getPhysicsPosition() const {
		for (const auto& part : m_parts) {
			if (part.alive()) return part.getPhysicalPosition();
		}
		return vec2();
	}

	b2Body* getPhysicsBody() const {
		for (const auto& part : m_parts) {
			if (part.alive()) return part.m_body;
		}
		return nullptr;
	}

	bool isLinked(const RaftPart &part);

	int getPrimaryPartIdx() const
	{
		for (u32 i = 0; i < m_parts.size(); i++)
		{
			if (m_parts[i].alive() && m_parts[i].m_body)
			{
				return i;
			}
		}
		return -1;
	}

	vec3 getBoundingCircle()
	{
		vec3 boundingCircle; // xy = center, z = radius

		aabb2 boundingBox = aabbExpandStart();
		
		for (const RaftPart& part : m_parts)
		{
			if(part.alive() && isLinked(part))
				aabbExpand(boundingBox, part.getPhysicalPosition());
		}

		const float magicValue = 1.3f; // graham: scientific

		vec2 center = aabbCenter(boundingBox);
		boundingCircle.x = center.x;
		boundingCircle.y = center.y;
		boundingCircle.z = std::max(1.f, magicValue * std::max(boundingBox.width() * 0.5f, boundingBox.height() * 0.5f));

		return boundingCircle;
	}

	vec3 getVisibilityCircle()
	{
		vec3 boundingCircle = getBoundingCircle();
		boundingCircle.z *= 1.8f; // graham: scientific
		return boundingCircle;
	}

	void update();

	template <typename Fn>
	void iterAttachmentPoints(const ivec2& attachmentSize, Fn fn) const {
		uint partIdx = 0;

		auto addCandidate = [&](int x, int y) {
			if (canAddPart(ivec2(x, y), attachmentSize, Rafter::Bit(0), Rafter::Bit::PlatformOrAttachment)) {
				fn(ivec2(x, y), partIdx);
			}
		};

		for (const RaftPart& part : m_parts) {
			if (part.m_blueprint->type == RaftPartType::Platform) {
				iterRectOutline(
					part.size() + attachmentSize + 1,
					part.m_offset - attachmentSize,
					addCandidate);
			}
			++partIdx;
		}
	}

	template <typename Fn>
	void iterPayloadPoints(const ivec2& attachmentSize, Fn fn) const {
		uint partIdx = 0;

		auto addCandidate = [&](int x, int y) {
			if (canAddPart(ivec2(x, y), attachmentSize, Rafter::Bit::Platform, Rafter::Bit::Payload)) {
				fn(ivec2(x, y), partIdx);
			}
		};

		for (const RaftPart& part : m_parts) {
			if (part.m_blueprint->type == RaftPartType::Platform) {
				for (int y = 0; y < part.size().y; ++y) {
					for (int x = 0; x < part.size().x; ++x) {
						addCandidate(x + part.m_offset.x, y + part.m_offset.y);
					}
				}
			}
			++partIdx;
		}
	}

	bool canAddPart(const ivec2& offset, const ivec2& size, Rafter::Bit reqBit, Rafter::Bit exclBit) const {
		for (int y = 0; y < size.y; ++y)
			for (int x = 0; x < size.x; ++x)
				if (!rafter.test(x + offset.x, y + offset.y, reqBit, exclBit)) return false;
		return true;
	}

	//bool canAddPart(const ivec2& offset, RaftPartBlueprint* part) const;

	void rasterPart(ivec2 offset, ivec2 size, Rafter::Bit bit) {
		/*if (!canAddPart(offset, size, bit)) {
			return false;
		}*/

		for (int y = 0; y < size.y; ++y) {
			for (int x = 0; x < size.x; ++x) {
				rafter.set(x + offset.x, y + offset.y, bit);
			}
		}
	}

	void addPart(ivec2 offset, RaftPartBlueprint* part);
	void addPayload(int partIdx, ivec2 offset, RaftPartBlueprint* part);

	/*
	void addPartAndCreatePhysics(int parentIdx, const RaftPart& part);
	*/

	void replaceWith(const Raft& templateRaft);
};
