#include "RaftUI.h"
#include "Raft.h"
#include "RaftPartLibrary.h"
#include "WaterWorld.h"
#include "RaftRenderer.h"

#include <Framework/PrimitiveBatch.h>
#include <Framework/BitmapFont.h>
#include <Framework/TextureLibrary.h>
#include <Framework/Sprite.h>

RaftUI* g_raftUI = nullptr;

std::vector<RaftPartSlot> getExtensionPartSlots(const Raft& raft, RaftPartBlueprint* part) {
	std::vector<RaftPartSlot> result;
	if (part->type == RaftPartType::Payload) {
		raft.iterPayloadPoints(part->size, [&](ivec2 p, uint parentPartIdx) {
			result.push_back(RaftPartSlot { p, int(parentPartIdx), vec2(0.0f) });
		});
	} else {
		raft.iterAttachmentPoints(part->size, [&](ivec2 p, uint parentPartIdx) {
			result.push_back(RaftPartSlot{ p, int(parentPartIdx), vec2(0.0f) });
		});
	}

	for(RaftPartSlot& it: result)
	{
		it.physicalPosition = raft.m_parts[it.parentPartIndex].getPhysicalPosition() + vec2(it.localPosition);
	}
	return result;
}

void addPart(Raft& raft, int parentPartIndex, vec2 position, const char* partName)
{
	printf("Adding part '%s' to raft at position %f %f\n", partName, position.x, position.y);
}

RaftUI::RaftUI()
{
	SDL_assert(g_raftUI == nullptr);
	g_raftUI = this;
	m_fontRegular = new BitmapFontRenderer("Fonts/ampersand.fnt", "Fonts");
	m_fontLarge = new BitmapFontRenderer("Fonts/ampersandLarge.fnt", "Fonts");
}

RaftUI::~RaftUI()
{
	delete m_fontRegular;
	delete m_fontLarge;
}

void RaftUI::update(const PlayerInputState& input)
{
	if (!g_waterWorld->m_playerRaft->m_parts[0].alive() || m_showGameOver){
		m_showGameOver = true;
		return;
	}

	if (!m_storyMessages.empty() && m_cursorMoveCooldown == 0){
		if (input.action){
			m_cursorMoveCooldown = 15;
			m_storyMessages.erase(m_storyMessages.begin());
		}
	}

	updateAvailablePartsMenu();


	if (input.moveDir == vec2(0.0f)){
		m_cursorMoveCooldown = 0;
	}

	if (!m_menuItems.empty() && !m_showselectAttachmentPoint) {
		m_showAvailableParts = true;
	}

	if (m_menuItems.empty() && m_showAvailableParts) {
		m_showAvailableParts = false;
	}


	if (m_showAvailableParts && m_menuItems.size() > 0) {
		if (m_cursorMoveCooldown == 0 && input.moveDir.x != 0) {
			if (input.moveDir.x < -0.5f) {
				--m_selectedMenuItem;
				m_cursorMoveCooldown = 15;
			}
			else if (input.moveDir.x > 0.5f) {
				++m_selectedMenuItem;
				m_cursorMoveCooldown = 15;
			}
			m_selectedMenuItem = ((int)m_menuItems.size() + (u32)m_selectedMenuItem) % (int)m_menuItems.size();
		}

		m_selectedPartName = m_menuItems[m_selectedMenuItem].partName;
	}


	if (m_showselectAttachmentPoint) {

		auto findAttachmentSlot = [&](vec2 dirPhys) {
			vec2 dir = normalize(physCoordToGfx(dirPhys));
			int bestAttachmentSlot = m_selectedAttachmentSlot;
			float bestDist = FLT_MAX;
			const RaftPartSlot& currentSlot = m_availablePartSlots[m_selectedAttachmentSlot];
			vec2 currentPos = physCoordToGfx((vec2)currentSlot.localPosition);
			for (size_t i = 0; i < m_availablePartSlots.size(); ++i) {
				if (i == m_selectedAttachmentSlot) {
					continue;
				}
				const auto& partSlot = m_availablePartSlots[i];
				vec2 partPos = physCoordToGfx((vec2)partSlot.localPosition);
				vec2 partDir = normalize(partPos - currentPos);
				float dist = glm::length(partPos-currentPos);
				float dp = dot(partDir, dir);
				if (dp > 0 && dist < bestDist)
				{
					bestAttachmentSlot = (int)i;
					bestDist = dist;
				}
				static int a; a = 0;
			}
			return bestAttachmentSlot;
		};

		if (m_cursorMoveCooldown == 0 && input.moveDir != vec2(0.0f)) {
			int foundAttachmentSlot = findAttachmentSlot(input.moveDir);
			if (foundAttachmentSlot != m_selectedAttachmentSlot)
			{
				m_cursorMoveCooldown = 15;
				m_selectedAttachmentSlot = foundAttachmentSlot;
			}
		}
	}


	if (m_cursorMoveCooldown > 0) {
		--m_cursorMoveCooldown;
	}

	if (m_showselectAttachmentPoint)
	{
		if (input.action)
		{
			m_showselectAttachmentPoint = false;

			printf("ATTACH PART!");

			//m_selectedRaft->addPartAndCreatePhysics(m_availablePartSlots[m_selectedAttachmentSlot].parentPartIndex, m_tempRaft.m_parts.back());
			m_selectedRaft->replaceWith(m_tempRaft);

			g_waterWorld->m_unattachedParts[m_selectedMenuItem] = g_waterWorld->m_unattachedParts.back();;
			g_waterWorld->m_unattachedParts.resize(g_waterWorld->m_unattachedParts.size() - 1);
			//updateAvailablePartsMenu();
		}
	}

	if (m_showAvailableParts)
	{
		m_showselectAttachmentPoint = false;

		m_selectedPartName = g_waterWorld->m_unattachedParts[0]->name;

		SDL_assert(m_selectedRaft);
		auto partBlueprint = g_raftPartLibrary->get(m_selectedPartName.c_str());
		m_availablePartSlots = getExtensionPartSlots(*m_selectedRaft, partBlueprint);

		if (input.action)
		{ 
			if (m_availablePartSlots.size() > 0) {
				// TODO: confirm selection action
				m_showAvailableParts = false;
				m_showselectAttachmentPoint = true;
			} else {
				// TODO: BEEEEEEEEP, no slots available
			}
		} else if (input.fire1 && !m_lastInputState.fire1 && g_waterWorld->m_unattachedParts.size() > 0) {
			g_waterWorld->m_unattachedParts[m_selectedMenuItem] = g_waterWorld->m_unattachedParts.back();;
			g_waterWorld->m_unattachedParts.resize(g_waterWorld->m_unattachedParts.size() - 1);
			updateAvailablePartsMenu();
		}
	}

	m_selectedAttachmentSlot = glm::min<int>(m_selectedAttachmentSlot, (int)m_availablePartSlots.size() - 1);
	m_selectedAttachmentSlot = glm::max<int>(m_selectedAttachmentSlot, 0);

	if (input.action)
	{
		m_showMessage = false;
	}

	m_lastInputState = input;
}

void RaftUI::selectRaft(Raft* raft)
{
	if (m_selectedRaft == raft) {
		return;
	}

	m_selectedRaft = raft;

	updateAvailablePartsMenu();
}

void RaftUI::renderWorldSpace()
{
	if (m_showGameOver){
		return;
	}

	if (!m_storyMessages.empty()){
		return;
	}

	if (m_showselectAttachmentPoint){
		renderSelectAttachmentPointWS();
	}
}

void RaftUI::renderGameOver()
{
	auto prim = g_primitiveRenderer;

	ivec2 drawableSize;
	SDL_GL_GetDrawableSize(g_window, &drawableSize.x, &drawableSize.y);

	dimWorld(255);

	vec2 centerPos = (vec2)drawableSize / 2.0f;
	const Sprite* sprite = g_textureLibrary->getSprite("DeadScreen");
	if (sprite)
	{
		vec2 spriteSize = vec2(0.0f);
		spriteSize.x = (float)sprite->w;
		spriteSize.y = (float)sprite->h;
		drawSprite(prim, *sprite, centerPos - spriteSize*0.5f);
	}
}

void RaftUI::renderStoryMessage()
{
	auto prim = g_primitiveRenderer;

	ivec2 drawableSize;
	SDL_GL_GetDrawableSize(g_window, &drawableSize.x, &drawableSize.y);

	dimWorld(255);

	vec2 centerPos = (vec2)drawableSize / 2.0f;

	vec2 textPos = centerPos;
	vec2 spritePos = centerPos;

	StoryMessage& message = m_storyMessages[0];

	if (message.sprite)
	{
		vec2 spriteSize = vec2(0.0f);
		spriteSize.x = (float)message.sprite->w;
		spriteSize.y = (float)message.sprite->h;

		drawSprite(prim, *message.sprite, centerPos - spriteSize*0.5f);
	}


	vec2 textSize = vec2(0.0f);
	if (!message.text.empty()) {
		textSize = m_fontRegular->measureString(message.text.c_str());

		textPos -= textSize * vec2(0.5f, -0.5f);
		m_fontRegular->drawString(prim, textPos, message.text.c_str());
	}

}

void RaftUI::renderScreenSpace()
{
	if (m_showGameOver){
		renderGameOver();
		return;
	}

	if (!m_storyMessages.empty()){
		renderStoryMessage();
		return;
	}

	if (!m_selectedRaft) {
		return;
	}

	const Raft& raft = *m_selectedRaft;

	auto prim = g_primitiveRenderer;
	prim->setLineWidth(1.0f);

	ivec2 drawableSize;
	SDL_GL_GetDrawableSize(g_window, &drawableSize.x, &drawableSize.y);

	{
		char str[1024];

		sprintf(str, 
			"Boaty Mc DoorFace\n"
			"Parts: %d", 
			(int)raft.m_parts.size());

		vec2 size = m_fontRegular->measureString(str);

		vec2 textPos = vec2(10.0f, 10.0f + size.y);

		float borderSize = 5.0f;
		aabb2 textBox;
		textBox.min = vec2(0.0f);
		textBox.max = size;
		textBox.translate(vec2(0, -textBox.height()));
		textBox.min -= borderSize;
		textBox.max += borderSize;
		textBox.translate(textPos);
		prim->drawRect(textBox, ColorRGBA8::Black(64));

		m_fontRegular->drawString(prim, textPos + vec2(2.0f, -2.0f), str, ColorRGBA8::Black());
		m_fontRegular->drawString(prim, textPos + vec2(0.0f), str, ColorRGBA8::White());
	}

	if (m_showAvailableParts) {
		// renderAvailableParts();
		renderKeepOrTrash();
	}

	if (m_showMessage) {
		renderMessage();
	}

	if (m_showselectAttachmentPoint){
		renderSelectAttachmentPoint();
	}
}

void RaftUI::renderAvailableParts()
{
	vec2 pos = vec2(200.0f, 400.0f);

	auto prim = g_primitiveRenderer;
	prim->setLineWidth(1.0f);

	ivec2 drawableSize;
	SDL_GL_GetDrawableSize(g_window, &drawableSize.x, &drawableSize.y);

	dimWorld();

	char str[1024];

	int itemCounter = 0;
	for (const MenuItem& it : m_menuItems)
	{
		const std::string& name = it.text;
		const Sprite* sprite = it.sprite;

		sprintf(str, "%s", name.c_str());
		vec2 textSize = m_fontRegular->measureString(str);
		vec2 spriteSize = vec2(0.0f);

		float elemWidth = textSize.x;

		if (sprite) {
			spriteSize.x = (float)sprite->w;
			spriteSize.y = (float)sprite->h;
		}

		elemWidth = glm::max(elemWidth, spriteSize.x);

		pos += vec2(elemWidth * 0.5f, 0.0f);

		vec2 textOffset = vec2(-textSize.x/2, 0.0f);
		vec2 spriteOffset = vec2(0.0f);

		if (sprite){
			vec2 spriteOffset = vec2(-spriteSize.x / 2, 0.0f);
			drawSprite(prim, *sprite, pos+spriteOffset);
		}

		m_fontRegular->drawString(prim, pos+textOffset+vec2(2, -2), str, ColorRGBA8(0, 0, 0));
		m_fontRegular->drawString(prim, pos+textOffset, str, ColorRGBA8(255, 200, 50));

		if (itemCounter == m_selectedMenuItem)
		{
			const Sprite* arrowSprite = g_textureLibrary->getSprite("arrowUp");
			drawSprite(prim, *arrowSprite, pos + vec2(-0.5f*arrowSprite->w, -(float)arrowSprite->h - textSize.y));
		}

		pos += vec2(elemWidth * 0.5f + 10.0f, 0.0f);

		++itemCounter;
	}

}

void RaftUI::renderMessage()
{
	auto prim = g_primitiveRenderer;

	ivec2 drawableSize;
	SDL_GL_GetDrawableSize(g_window, &drawableSize.x, &drawableSize.y);

	dimWorld();

	vec2 centerPos = (vec2)drawableSize / 2.0f;

	vec2 textPos = centerPos;
	vec2 captionPos = centerPos;
	vec2 spritePos = centerPos;

	vec2 captionSize = vec2(0.0f);
	if (!m_message.caption.empty()) {
		captionSize = m_fontLarge->measureString(m_message.caption.c_str());
	}

	vec2 textSize = vec2(0.0f);
	if (!m_message.text.empty()) {
		textSize = m_fontRegular->measureString(m_message.text.c_str());
	}

	captionPos -= captionSize * vec2(0.5f, -0.5f);
	textPos -= textSize * vec2(0.5f, -0.5f);

	textPos.y -= captionSize.y;

	m_fontLarge->drawString(prim, captionPos + vec2(3.0f, -3.0f), m_message.caption.c_str(), ColorRGBA8::Black());
	m_fontLarge->drawString(prim, captionPos, m_message.caption.c_str(), m_message.captionColor);

	m_fontRegular->drawString(prim, textPos, m_message.text.c_str());
}

void RaftUI::renderKeepOrTrash()
{
	if (m_menuItems.empty()) {
		return;
	}

	auto prim = g_primitiveRenderer;

	ivec2 drawableSize;
	SDL_GL_GetDrawableSize(g_window, &drawableSize.x, &drawableSize.y);

	dimWorld();

	vec2 centerPos = (vec2)drawableSize / 2.0f;

	vec2 messagePos = centerPos;
	vec2 captionPos = centerPos;
	vec2 spritePos = centerPos;
	vec2 spriteSize = vec2(0.0f);

	char captionText[1024];
	sprintf(captionText, "Item name");

	char messageText[1024];

	if (m_availablePartSlots.empty()) {
		sprintf(messageText, "Sadly, you don't have space for it... Press [X] to trash.");
	}
	else {
		sprintf(messageText, "Keep it [SPACE] or trash it [X]?");
	}

	{
		const MenuItem& it = m_menuItems[0];

		const std::string& name = it.text;
		const Sprite* sprite = it.sprite;

		sprintf(captionText, "You found an awesome %s!", name.c_str());

		if (sprite) {
			spriteSize.x = (float)sprite->w;
			spriteSize.y = (float)sprite->h;
		}

		if (sprite) {
			vec2 spriteOffset = vec2(-spriteSize.x / 2, 0.0f);
			drawSprite(prim, *sprite, centerPos - spriteSize*0.5f);
		}
	}

	vec2 captionSize = vec2(0.0f);
	captionSize = m_fontLarge->measureString(captionText);

	vec2 textSize = vec2(0.0f);
	textSize = m_fontRegular->measureString(messageText);

	captionPos -= captionSize * vec2(0.5f, -1.0f);
	captionPos.y += spriteSize.y;

	messagePos -= textSize * vec2(0.5f, -0.5f);
	messagePos.y -= spriteSize.y * 0.5f;

	messagePos.y -= captionSize.y;

	m_fontLarge->drawString(prim, captionPos + vec2(3.0f, -3.0f), captionText, ColorRGBA8::Black());
	m_fontLarge->drawString(prim, captionPos, captionText, ColorRGBA8(255, 200, 50));

	m_fontRegular->drawString(prim, messagePos, messageText);
}

bool RaftUI::wantPauseSimulation() const
{
	return m_showAvailableParts || m_showMessage || m_showselectAttachmentPoint ||
		!m_storyMessages.empty() || m_showGameOver;
}

void RaftUI::updateAvailablePartsMenu()
{
	m_menuItems.clear();
	//for (auto it : g_raftPartLibrary->library)
	for (auto it : g_waterWorld->m_unattachedParts)
	{
		MenuItem item;
		item.sprite = g_textureLibrary->getSprite(it->texture);
		item.text = it->name;
		item.size = vec2(item.sprite->w, item.sprite->h);
		item.partName = it->name;
		m_menuItems.push_back(item);
	}

	//m_selectedMenuItem = 0;
	m_selectedMenuItem = glm::clamp(int(m_selectedMenuItem), 0, int(g_waterWorld->m_unattachedParts.size() - 1));
}

void RaftUI::dimWorld(u8 alpha)
{
	auto prim = g_primitiveRenderer;

	ivec2 drawableSize;
	SDL_GL_GetDrawableSize(g_window, &drawableSize.x, &drawableSize.y);

	aabb2 windowRect;
	windowRect.min = vec2(0.0f);
	windowRect.max = (vec2)drawableSize;
	prim->drawRect(windowRect, ColorRGBA8::Black(alpha));
}

void RaftUI::renderSelectAttachmentPoint()
{
	auto prim = g_primitiveRenderer;

	ivec2 drawableSize;
	SDL_GL_GetDrawableSize(g_window, &drawableSize.x, &drawableSize.y);

	dimWorld(255);

	vec2 centerPos = (vec2)drawableSize / 2.0f;

	vec2 textPos = vec2(centerPos.x, 100);;
	vec2 captionPos = vec2(centerPos.x, drawableSize.y - 100);

	const char* captionStr = "Attach new part!";
	char textStr[1024];
	sprintf(textStr, "Put your new %s somewhere on your raft!", m_selectedPartName.c_str());

	vec2 captionSize = vec2(0.0f);
	captionSize = m_fontLarge->measureString(captionStr);

	vec2 textSize = vec2(0.0f);
	textSize = m_fontRegular->measureString(textStr);

	captionPos -= captionSize * vec2(0.5f, -0.5f);
	textPos -= textSize * vec2(0.5f, -0.5f);

	m_fontLarge->drawString(prim, captionPos + vec2(3.0f, -3.0f), captionStr, ColorRGBA8::Black());
	m_fontLarge->drawString(prim, captionPos, captionStr, ColorRGBA8(255, 200, 50));

	m_fontRegular->drawString(prim, textPos + vec2(3.0f, -3.0f), textStr, ColorRGBA8::Black());
	m_fontRegular->drawString(prim, textPos, textStr);

	// draw the damn raft

	Rafterizer rafterizer;
	m_tempRaft = *m_selectedRaft;

	RaftPartBlueprint* partBlueprint = g_raftPartLibrary->get(m_selectedPartName.c_str());
	if (partBlueprint->type == RaftPartType::Payload) {
		const auto& parentPart = m_selectedRaft->m_parts[m_availablePartSlots[m_selectedAttachmentSlot].parentPartIndex];
		m_tempRaft.addPayload(
			m_availablePartSlots[m_selectedAttachmentSlot].parentPartIndex,
			m_availablePartSlots[m_selectedAttachmentSlot].localPosition, partBlueprint);
	}
	else {
		m_tempRaft.addPart(m_availablePartSlots[m_selectedAttachmentSlot].localPosition, partBlueprint);
		//m_tempRaft.link(m_availablePartSlots[m_selectedAttachmentSlot].parentPartIndex, (u32)m_tempRaft.m_parts.size() - 1);
	}

	m_tempRaft.findLinks();

	vec2 raftOffset = gfxCoordToPhys(centerPos);
	for (RaftPart& part : m_tempRaft.m_parts)
	{
		part.m_physicalPosition = (vec2)part.m_offset + raftOffset - part.offsetFromPhysCenterToGrid();
		part.m_swayStrength = 0.0f;
	}

	rafterizer.add(m_tempRaft);

	rafterizer.m_enableHealthBars = false;
	rafterizer.render();

	const Sprite* partSprite = g_textureLibrary->getSprite(partBlueprint->texture);
	const Sprite* placeholderCube = g_textureLibrary->getSprite("placeholderCube");

	int counter = 0;
	for (auto& partSlot : m_availablePartSlots)
	{
		vec2 p = physCoordToGfx((vec2)partSlot.localPosition + raftOffset);

		prim->setLineWidth(2.0f);
		prim->drawCross(p, 8.0f, ColorRGBA8(0, 255, 0));

		if (counter == m_selectedAttachmentSlot){
			// prim->drawDiagonalCross(p, 15.0f, ColorRGBA8(255, 0, 0));

			const Sprite* arrowSprite = g_textureLibrary->getSprite("arrowUp");
			drawSprite(prim, *arrowSprite,
				p + vec2(-0.5f*arrowSprite->w, -(float)arrowSprite->h)
				+ vec2(0.0f, -20.0f)
			);
		}
		else
		{
// 			drawSprite(prim, *placeholderCube,
// 				p + vec2(-0.5f*placeholderCube->w, -(float)placeholderCube->h)
// 			);
		}

		++counter;
	}
}

void RaftUI::renderSelectAttachmentPointWS()
{
	ivec2 drawableSize;
	SDL_GL_GetDrawableSize(g_window, &drawableSize.x, &drawableSize.y);

	if (!m_selectedRaft) {
		return;
	}
}

