#pragma once

#include <set>
#include <functional>
#include <iostream>
#include <iomanip>
#include <assert.h>

template <typename _STYPE, typename _ETYPE>
class StateMachine
{
private:
	struct StateCompare;

public:
	typedef StateMachine<_STYPE, _ETYPE>* StateMachinePtr;

	class State
	{
		struct EventCompare;
		friend StateMachine;
		typedef State* StatePtr;

	public:
		class Event
		{
			friend State;
			typedef Event* EventPtr;

		public:
			_ETYPE getId() const;
			StatePtr getState() const;
			
			bool hasAction() const;
			void bindAction(std::function<void(EventPtr, StatePtr, StatePtr)> action);

		private:
			Event(_ETYPE eventId, StatePtr targetState);
			~Event();

		private:
			_ETYPE m_id;
			StatePtr m_state;
			std::function<void(EventPtr, StatePtr, StatePtr)> m_action;
		};

		typedef Event* EventPtr;

	private:
		struct EventCompare
		{
			bool operator()(const EventPtr lhs, const EventPtr rhs) const
			{
				return rhs->m_id < lhs->m_id;
			}
		};

		typedef std::set<EventPtr, EventCompare> EventSet;

	public:
		_STYPE getId() const;
		bool isAccepted() const;
		bool isTerminal() const;

		EventPtr addEvent(_ETYPE eventId);
		EventPtr addEvent(_ETYPE eventId, _STYPE nextState, bool accepted = false);
		EventPtr getEvent(_ETYPE eventId) const;
		
		bool hasEntryAction() const;
		bool hasExitAction() const;
		
		StatePtr bindEntryAction(std::function<void(StatePtr)> entryAction);
		StatePtr bindExitAction(std::function<void(StatePtr)> exitAction);
		
		const EventSet& getEvents() const;

		const StateMachine* getParent() const { return m_parent; }

	private:
		State(StateMachine* parent, _STYPE stateId, bool accepted);
		~State();

		void doAction(EventPtr event);
		
		typename EventSet::const_iterator findEvent(_ETYPE eventId) const;
		typename EventSet::iterator findEvent(_ETYPE eventId);

	private:
		StateMachine* m_parent;
		
		_STYPE m_id;
		
		bool m_accepted;
		EventSet m_events;
		
		std::function<void(StatePtr)> m_entryAction;
		std::function<void(StatePtr)> m_exitAction;
	};

	typedef State* StatePtr;

private:
	struct StateCompare
	{
		bool operator()(const StatePtr lhs, const StatePtr rhs) const
		{
			return rhs->m_id < lhs->m_id;
		}
	};

	typedef std::set<StatePtr, StateCompare> StateSet;

public:
	StateMachine();
	~StateMachine();

	void setUserData(void* userData) { m_userData = userData; }
	void* getUserData() const { return m_userData; }

	StatePtr addState(_STYPE stateId, bool accepted = false);
	StatePtr getState(_STYPE stateId) const;
	StatePtr getCurrentState() const;
	StatePtr getInitialState() const;
	StatePtr nextState(_ETYPE eventId);
	StatePtr reset();

	const StateSet& getStates() const;

	void dumpStates();

private:
	typename StateSet::const_iterator findState(_STYPE stateId) const;
	typename StateSet::iterator findState(_STYPE stateId);

private:
	StateSet m_states;
	StatePtr m_initial;
	StatePtr m_current;

	void* m_userData = nullptr;
};

template <typename _STYPE, typename _ETYPE>
_ETYPE StateMachine<_STYPE, _ETYPE>::State::Event::getId() const
{
	return m_id;
}

template <typename _STYPE, typename _ETYPE>
typename  StateMachine<_STYPE, _ETYPE>::State::StatePtr
StateMachine<_STYPE, _ETYPE>::State::Event::getState() const
{
	return m_state;
}

template <typename _STYPE, typename _ETYPE>
bool StateMachine<_STYPE, _ETYPE>::State::Event::hasAction() const
{
	return bool(m_action);
}

template <typename _STYPE, typename _ETYPE>
void StateMachine<_STYPE, _ETYPE>::State::Event::bindAction(
	std::function<void(EventPtr, StatePtr, StatePtr)> action)
{
	m_action = action;
}

template <typename _STYPE, typename _ETYPE>
StateMachine<_STYPE, _ETYPE>::State::Event::Event(_ETYPE eventId, StatePtr targetState)
: m_id(eventId)
, m_state(targetState)
{
}

template <typename _STYPE, typename _ETYPE>
StateMachine<_STYPE, _ETYPE>::State::Event::~Event()
{
}

template <typename _STYPE, typename _ETYPE>
_STYPE StateMachine<_STYPE, _ETYPE>::State::getId() const
{
	return m_id;
}

template <typename _STYPE, typename _ETYPE>
bool StateMachine<_STYPE, _ETYPE>::State::isAccepted() const
{
	return m_accepted;
}

template <typename _STYPE, typename _ETYPE>
bool StateMachine<_STYPE, _ETYPE>::State::isTerminal() const
{
	return m_events.size() == 0;
}

template <typename _STYPE, typename _ETYPE>
typename StateMachine<_STYPE, _ETYPE>::State::EventPtr
StateMachine<_STYPE, _ETYPE>::State::addEvent(_ETYPE eventId)
{
	typename EventSet::const_iterator it = findEvent(eventId);
	if (it == m_events.end())
	{
		EventPtr event = new Event(eventId, this);
		m_events.insert(event);
		return event;
	}
	else
	{
		assert(false);
		return nullptr;
	}
}

template <typename _STYPE, typename _ETYPE>
typename StateMachine<_STYPE, _ETYPE>::State::EventPtr
StateMachine<_STYPE, _ETYPE>::State::addEvent(_ETYPE eventId, _STYPE nextState, bool accepted)
{
	typename EventSet::const_iterator it = findEvent(eventId);

	StatePtr state = m_parent->addState(nextState, accepted);
	if (it == m_events.end())
	{
		EventPtr event = new Event(eventId, state);
		m_events.insert(event);
		return event;
	}
	else
	{
		assert(false);
		return nullptr;
	}
}

template <typename _STYPE, typename _ETYPE>
typename StateMachine<_STYPE, _ETYPE>::State::EventPtr
StateMachine<_STYPE, _ETYPE>::State::getEvent(_ETYPE eventId) const
{
	typename EventSet::const_iterator it = findEvent(eventId);
	if (it != m_events.end())
		return *it;

	return nullptr;
}

template <typename _STYPE, typename _ETYPE>
void StateMachine<_STYPE, _ETYPE>::State::doAction(EventPtr event)
{
	event->m_action(event, this, event->getState());
}

template <typename _STYPE, typename _ETYPE>
StateMachine<_STYPE, _ETYPE>::State::State(StateMachine* parent, _STYPE stateId, bool accepted)
: m_parent(parent)
, m_id(stateId)
, m_accepted(accepted)
, m_events()
{
}

template <typename _STYPE, typename _ETYPE>
StateMachine<_STYPE, _ETYPE>::State::~State()
{
	for (auto event : m_events)
		delete event;
}

template <typename _STYPE, typename _ETYPE>
typename StateMachine<_STYPE, _ETYPE>::State::EventSet::const_iterator
StateMachine<_STYPE, _ETYPE>::State::findEvent(_ETYPE eventId) const
{
	Event noop(eventId, nullptr);
	return m_events.find(&noop);
}

template <typename _STYPE, typename _ETYPE>
typename StateMachine<_STYPE, _ETYPE>::State::EventSet::iterator
StateMachine<_STYPE, _ETYPE>::State::findEvent(_ETYPE eventId)
{
	Event noop(eventId, nullptr);
	return m_events.find(&noop);
}

template <typename _STYPE, typename _ETYPE>
bool StateMachine<_STYPE, _ETYPE>::State::hasEntryAction() const
{
	return bool(m_entryAction);
}

template <typename _STYPE, typename _ETYPE>
bool StateMachine<_STYPE, _ETYPE>::State::hasExitAction() const
{
	return bool(m_exitAction);
}

template <typename _STYPE, typename _ETYPE>
typename StateMachine<_STYPE, _ETYPE>::State::StatePtr
StateMachine<_STYPE, _ETYPE>::State::bindEntryAction(std::function<void(StatePtr)> entryAction)
{
	m_entryAction = entryAction;
	return this;
}

template <typename _STYPE, typename _ETYPE>
typename StateMachine<_STYPE, _ETYPE>::State::StatePtr
StateMachine<_STYPE, _ETYPE>::State::bindExitAction(std::function<void(StatePtr)> exitAction)
{
	m_exitAction = exitAction;
	return this;
}

template <typename _STYPE, typename _ETYPE>
const typename StateMachine<_STYPE, _ETYPE>::State::EventSet&
StateMachine<_STYPE, _ETYPE>::State::getEvents() const
{
	return m_events;
}

template <typename _STYPE, typename _ETYPE>
StateMachine<_STYPE, _ETYPE>::StateMachine()
: m_states()
, m_initial(nullptr)
, m_current(nullptr)
{
}

template <typename _STYPE, typename _ETYPE>
StateMachine<_STYPE, _ETYPE>::~StateMachine()
{
	for (auto state : m_states)
		delete state;
}

template <typename _STYPE, typename _ETYPE>
typename StateMachine<_STYPE, _ETYPE>::StatePtr
StateMachine<_STYPE, _ETYPE>::addState(_STYPE stateId, bool accepted)
{
	typename StateSet::const_iterator it = findState(stateId);
	if (it == m_states.end())
	{
		StatePtr state = new State(this, stateId, accepted);

		if (m_states.size() == 0)
		{
			m_initial = state;
			m_current = m_initial;
		}

		m_states.insert(state);
		return state;
	}
	else
		return *it;
}

template <typename _STYPE, typename _ETYPE>
typename StateMachine<_STYPE, _ETYPE>::StatePtr
StateMachine<_STYPE, _ETYPE>::getState(_STYPE stateId) const
{
	typename StateSet::const_iterator it = findState(stateId);
	if (it != m_states.end())
		return *it;

	return nullptr;
}

template <typename _STYPE, typename _ETYPE>
typename StateMachine<_STYPE, _ETYPE>::StatePtr
StateMachine<_STYPE, _ETYPE>::getCurrentState() const
{
	return m_current;
}

template <typename _STYPE, typename _ETYPE>
typename StateMachine<_STYPE, _ETYPE>::StatePtr
StateMachine<_STYPE, _ETYPE>::getInitialState() const
{
	return m_initial;
}

template <typename _STYPE, typename _ETYPE>
typename StateMachine<_STYPE, _ETYPE>::StatePtr
StateMachine<_STYPE, _ETYPE>::nextState(_ETYPE eventId)
{
	typename State::EventPtr event = m_current->getEvent(eventId);
	assert(event != nullptr);

	StatePtr nextState = event->getState();
	
	if (event->hasAction())
		m_current->doAction(event);

	if (m_current != nextState)
	{
		if (m_current->hasExitAction())
			m_current->m_exitAction(nextState);

		if (nextState->hasEntryAction())
			nextState->m_entryAction(m_current);

		m_current = nextState;
	}

	return m_current;
}

template <typename _STYPE, typename _ETYPE>
typename StateMachine<_STYPE, _ETYPE>::StatePtr
StateMachine<_STYPE, _ETYPE>::reset()
{
	m_current = m_initial;
	return m_current;
}

template <typename _STYPE, typename _ETYPE>
const typename StateMachine<_STYPE, _ETYPE>::StateSet&
StateMachine<_STYPE, _ETYPE>::getStates() const
{
	return m_states;
}

template <typename _STYPE, typename _ETYPE>
void StateMachine<_STYPE, _ETYPE>::dumpStates()
{
	std::cout << std::left << std::setfill(' ')
		<< std::setw(16) << "State"
		//<< std::setw(16) << "Accepted"
		//<< std::setw(16) << "Terminal"
		<< std::setw(16) << "Event"
		//<< std::setw(16) << "Action"
		<< std::setw(16) << "NextState" << std::endl;

	for (auto state : m_states)
	{
		if (state->getEvents().size() == 0)
		{
			std::cout << std::left << std::setfill(' ')
				<< std::setw(16) << state->getId()
				//<< std::setw(16) << std::boolalpha << state->isAccepted()
				//<< std::setw(16) << std::boolalpha << state->isTerminal()
				<< std::setw(16) << "---"
				//<< std::setw(16) << "---"
				<< std::setw(16) << "---" << std::endl;
		}

		for (auto event : state->getEvents())
		{
			std::cout << std::left << std::setfill(' ')
				<< std::setw(16) << state->getId()
				//<< std::setw(16) << std::boolalpha << state->isAccepted()
				//<< std::setw(16) << std::boolalpha << state->isTerminal()
				<< std::setw(16) << event->getId()
				//<< std::setw(16) << std::boolalpha << event->hasAction()
				<< std::setw(16) << ((event->getState() == nullptr) ? "---" : event->getState()->getId()) << std::endl;
		}
	}

	std::cout << std::setfill('-') << std::setw(80) << "" << std::endl;
}

template <typename _STYPE, typename _ETYPE>
typename StateMachine<_STYPE, _ETYPE>::StateSet::const_iterator
StateMachine<_STYPE, _ETYPE>::findState(_STYPE stateId) const
{
	State noop(nullptr, stateId, false);
	return m_states.find(&noop);
}

template <typename _STYPE, typename _ETYPE>
typename StateMachine<_STYPE, _ETYPE>::StateSet::iterator
StateMachine<_STYPE, _ETYPE>::findState(_STYPE stateId)
{
	State noop(nullptr, stateId, false);
	return m_states.find(&noop);
}
