#include "WaterGrid.h"

#include "Common.h"
#include <algorithm>
#include <imgui.h>

WaterSimulator::WaterSimulator()
{

}

WaterSimulator::~WaterSimulator()
{

}

void WaterSimulator::init(WaterGrid& grid, int gridSize)
{
    const int xSize = gridSize;
    const int ySize = gridSize;
    
    grid.vertices.resize((xSize + 1) * (ySize + 1));
    for (int i = 0, y = 0; y <= ySize; ++y)
    {
        for (int x = 0; x <= xSize; ++x, ++i)
        {
            grid.vertices[i] = vec2(x, y) + vec2(xSize * -0.5f, ySize * 0.0f);
        }
    }
    
    grid.triangles.resize(xSize * ySize * 6);
    for (int ti = 0, vi = 0, y = 0; y < ySize; ++y, ++vi)
    {
        for (int x = 0; x < xSize; ++x, ti += 6, ++vi)
        {
            grid.triangles[ti] = vi;
            grid.triangles[ti + 3] = grid.triangles[ti + 2] = vi + 1;
            grid.triangles[ti + 4] = grid.triangles[ti + 1] = vi + xSize + 1;
            grid.triangles[ti + 5] = vi + xSize + 2;
        }
    }
}

vec2 calculateWaveControlCurve(float x)
{
	vec2 result;
#if 0 // https://medium.com/@gordonnl/wind-waker-graphics-analysis-a0b575a31127#.lxnnpwkis
	result.x = (sinf(x) + sinf(2.2f*x + 5.52f) + sinf(2.9f*x + 0.93f) + sinf(4.6f*x + 8.94f)) / 4.0f;
	result.y = (cosf(x) + cosf(2.2f*x + 5.52f) + cosf(2.9f*x + 0.93f) + cosf(4.6f*x + 8.94f)) / 4.0f;
#else
	result.x = sinf(x);
	result.y = cosf(x);
#endif
	return result;
}

float WaterSimulator::evalHeightAtPoint(WaterGrid& grid, vec2 ij0)
{
	vec2 derp = calculateWaveControlCurve(ij0.x*grid.particleXDelay + ij0.y*grid.particleYDelay + m_timeIndex*grid.particleSpeed) * grid.particleOrbitDiameter;
	return derp.y;
}

void WaterSimulator::step(WaterGrid& grid)
{
    m_timeIndex++;
}

void WaterSimulator::edit(WaterGrid& grid)
{
	ImGui::SetNextWindowCollapsed(true, ImGuiSetCondition_FirstUseThisSession);
	ImGui::Begin("Water Simulation");
		ImGui::SliderFloat("Particle Speed", &grid.particleSpeed, -0.1f, 0.1f);
		ImGui::SliderFloat("Particle X Delay", &grid.particleXDelay, -1.0f, 1.0f);
		ImGui::SliderFloat("Particle Y Delay", &grid.particleYDelay, -1.0f, 1.0f);
		ImGui::SliderFloat("Particle Radius", &grid.particleRadius, 0.0f, 20.0f);
		ImGui::SliderFloat("Orbit Diameter", &grid.particleOrbitDiameter, 0.0f, 50.0f);
		//ImGui::SliderInt("Orbit Distance", &grid.distanceBetweenOrbits, 4, 100);
		ImGui::SliderFloat("UV Tiling", &grid.uvTiling, 0.05f, 1.4f);
	ImGui::End();
}

WaterRenderer::WaterRenderer()
{

}

WaterRenderer::~WaterRenderer()
{

}

void WaterRenderer::renderControls(PrimitiveBatch* prim, WaterGrid& grid, int timeIndex)
{
	for (unsigned int i = 0; i < grid.matrixSize; i++)
	{
		for (unsigned int j = 0; j < grid.matrixSize; j++)
		{
            vec2 shiftedPos =physCoordToGfx(vec2(i, j));
            
            glm::vec2 particleVtx;

            const float particlePos = i*grid.particleYDelay + j*grid.particleXDelay + timeIndex*grid.particleSpeed;

            particleVtx.x = cosf(particlePos)*grid.particleOrbitDiameter + shiftedPos.x;
            particleVtx.y = sinf(particlePos)*grid.particleOrbitDiameter + shiftedPos.y;
			prim->drawCross(particleVtx, grid.particleRadius, grid.particleColor);
		}
	}
}


vec2 calculateSurface(vec2 ij, float time)
{
	float x = (sinf(ij.x * 1.0f + time * 1.0f) + sinf(ij.x * 2.3f + time * 1.5f) + sinf(ij.x * 3.3f + time * 0.4f)) / 3.0f;
    float y = (sinf(ij.y * 1.0f + time * 1.0f) + sinf(ij.y * 2.3f + time * 1.5f) + sinf(ij.y * 3.3f + time * 0.4f)) / 3.0f;
	return vec2(x, y);
}

void WaterRenderer::renderMesh(PrimitiveBatch* prim, WaterGrid& grid, Sprite* image, int timeIndex, vec2 gridOffset, float zoom)
{
    prim->setTexture(image->texture);
    prim->setSampler(PrimitiveBatch::SamplerState_Linear);
    
    int triangleBase = 0;
    int trianglesLeft = (int)grid.triangles.size() / 3;
    while (trianglesLeft > 0)
    {
        int triangleCount = std::min(trianglesLeft, (int)prim->getMaxBatchVertices() / 3);
        PrimitiveBatch::BatchVertex* verts = prim->drawVertices(PrimitiveType::TriangleList, triangleCount*3);
        for (int triangle = 0; triangle < triangleCount; ++triangle)
        {
            const int triangleIndex = (triangle + triangleBase) * 3;
            
            const int i0 = grid.triangles[triangleIndex + 0];
            const int i1 = grid.triangles[triangleIndex + 1];
            const int i2 = grid.triangles[triangleIndex + 2];

            vec2 ij0 = (grid.vertices[i0] * zoom + gridOffset);
			vec2 ij1 = (grid.vertices[i1] * zoom + gridOffset);
			vec2 ij2 = (grid.vertices[i2] * zoom + gridOffset);
                        
            vec2 p0 = physCoordToGfx(ij0);
            vec2 p1 = physCoordToGfx(ij1);
            vec2 p2 = physCoordToGfx(ij2);
            
            const vec2 particlePos0 = calculateWaveControlCurve(ij0.x*grid.particleXDelay + ij0.y*grid.particleYDelay + timeIndex*grid.particleSpeed);
			const vec2 particlePos1 = calculateWaveControlCurve(ij1.x*grid.particleXDelay + ij1.y*grid.particleYDelay + timeIndex*grid.particleSpeed);
			const vec2 particlePos2 = calculateWaveControlCurve(ij2.x*grid.particleXDelay + ij2.y*grid.particleYDelay + timeIndex*grid.particleSpeed);
			
			auto& v0 = verts[(triangle * 3) + 0];
			auto& v1 = verts[(triangle * 3) + 1];
			auto& v2 = verts[(triangle * 3) + 2];

            v0.pos.x = particlePos0.x * grid.particleOrbitDiameter + p0.x;
            v0.pos.y = particlePos0.y * grid.particleOrbitDiameter + p0.y;
            v0.pos.z = 0.0f;
            
            v1.pos.x = particlePos1.x * grid.particleOrbitDiameter + p1.x;
            v1.pos.y = particlePos1.y * grid.particleOrbitDiameter + p1.y;
            v1.pos.z = 0.0f;
            
            v2.pos.x = particlePos2.x * grid.particleOrbitDiameter + p2.x;
            v2.pos.y = particlePos2.y * grid.particleOrbitDiameter + p2.y;
            v2.pos.z = 0.0f;
            
            v0.tex = ij0 * grid.uvTiling;
            v1.tex = ij1 * grid.uvTiling;
            v2.tex = ij2 * grid.uvTiling;
            
            // 6,66,115
            ColorRGBA8 oceanColor1 = { 118,182,196 };
            ColorRGBA8 oceanColor2 = { 6,66,115 }; // gradient for debugging
            
			v0.col0 = ColorRGBA::White();// oceanColor2;
            v1.col0 = ColorRGBA::White();// oceanColor2;
            v2.col0 = ColorRGBA::White();// oceanColor2;
        }

        triangleBase += triangleCount;
        trianglesLeft -= triangleCount;
        
        prim->flush();
    }
}
