#include "WaterWorld.h"
#include "RaftRenderer.h"
#include "Animation.h"
#include "WaterGrid.h"
#include "RaftGen.h"
#include "RaftPartLibrary.h"

#include <Framework/TextureLibrary.h>
#include <Framework/Sprite.h>
#include <Framework/PhysicsSystem.h>
#include <Framework/PrimitiveBatch.h>

#include "CameraShake.h"

WaterWorld* g_waterWorld = nullptr;

WaterWorld::WaterWorld()
{
	SDL_assert(g_waterWorld == nullptr);
	g_waterWorld = this;

	m_smashSound = jamLoadSoundClip("Sounds/woodCrunch.wav");
	m_smashSound.doneCallback = std::bind(&WaterWorld::doneSmashSound, this);
}

void despawnRaft(Raft* raft)
{
	for (RaftPart& it : raft->m_parts) {
		if (it.m_body) // don't double delete
			g_physicsSystem->physicsWorld->DestroyBody(it.m_body);
	
	}

	delete raft;
}

void WaterWorld::spawnRafts()
{
	const float enemyDistanceMin = 32;
	const float enemyDistanceMax = 48;
	const float enemyDespawnDistance = 80;

	const vec2 playerPos = m_playerRaft->getPhysicsPosition();

	for (int i = int(m_rafts.size()) - 1; i >= 0; --i) {
		if (length(playerPos - m_rafts[i]->getPhysicsPosition()) > enemyDespawnDistance) {
			despawnRaft(m_rafts[i]);
			m_rafts[i] = m_rafts.back();
			m_rafts.resize(m_rafts.size() - 1);
		}
	}

#define U01 (float(rand()) / RAND_MAX)

	if (m_rafts.size() < 7) {
		vec2 bestPos = vec2(0, 0);
		float bestDist = -1;

		// Find a random position around the player, pick the one farthest from all the rafts
		for (int retry = 0; retry < 10; ++retry) {
			const float enemyOffsetAngle = U01 * float(M_PI * 2);
			const float enemyDistance = glm::mix(enemyDistanceMin, enemyDistanceMax, U01);
			vec2 position = vec2(cos(enemyOffsetAngle) * enemyDistance, sin(enemyOffsetAngle) * enemyDistance) + playerPos;

			float distSum = 0.0f;
			for (Raft* r : m_rafts) {
				vec2 off = r->getPhysicsPosition() - position;
				if (!r->m_isPlayer) distSum += dot(off, off);
			}

			if (distSum > bestDist) {
				bestDist = distSum;
				bestPos = position;
			}
		}

		bool isPlayer = false;
		addRaft(m_raftGen->makeTestRaft(bestPos, isPlayer), isPlayer);
	}

#undef U01
}

bool circlesOverlap(vec3 a, vec3 b)
{
	vec2 d = vec2(a.x - b.x, a.y - b.y);
	float dist = sqrtf(d.x*d.x + d.y*d.y);
	return (dist < a.z + b.z);
}

void WaterWorld::update(aabb2 viewBounds, const WorldUpdateInfo& info)
{
	// Search for flocks to remove
	for (auto iter = m_flocks.begin(); iter != m_flocks.end(); ++iter)
	{
		aabb2 bounds = iter->getBounds();
		if (!intersectBoxBox(bounds, viewBounds))
		{
			//iter = m_flocks.erase(iter);
		}
	}

	const int toSpawn = flockCount - m_flocks.size();

	// more complex way of finding camera position because we can't 'just' use the first part
	// since it could've died
	vec2 center;
	float count = 0;
	for (auto part : g_waterWorld->m_playerRaft->m_parts)
	{
		if (part.alive())
		{
			center += part.getRenderPosition();
			count++;
		}
	}
	center /= count;

	vec2 cameraPosition = glm::mix(cameraPosition, center, 0.05f);


	for (int i = 0; i < toSpawn; ++i)
	{
		m_flocks.push_back(Flock());
		auto& flock = m_flocks.back();
		//vec2 loc = physCoordToGfx(cameraPosition);
		vec2 loc = physCoordToGfx(g_waterWorld->m_playerRaft->getPhysicsPosition());// cameraPosition;
		loc.x += ((randomUnitFloat() - 0.5f) * 1280) * 9.3 - 1280;
		loc.y += ((randomUnitFloat() - 0.5f) * 720) * 9.3 - 720;
		//loc = physCoordToGfx(loc);
		//loc = gfxCoordToPhys(loc);
		for (int j = 0; j < 1; ++j)
		{
			flock.add(Peck(loc + vec2(j)));
		}
	}

	// Update remaining flocks
	for (Flock& flock : m_flocks)
	{
		m_peckFX.step(flock);
	}
	
	spawnRafts();

	for (Raft* raft : m_rafts)
	{
		raft->update();
	}

	if(info.playerInputState.fire1)
	{
		vec2 p;
		if (m_playerRaft)
			p = m_playerRaft->getPhysicsPosition();

		m_playerRaft->shoot(info.playerInputState.shootTarget);

		//m_weaponSystem.shotgun(p, info.playerInputState.shootTarget - p, 0.2f);
	}

	if(info.playerInputState.fire2)
	{
		vec2 p;
		if (m_playerRaft)
			p = m_playerRaft->getPhysicsPosition();

		m_weaponSystem.grenade(p);

		if (!info.shake->running())
		{
			info.shake->start(100.0f, 100.0f);
		}
	}


	if (m_playerRaft)
	{
		m_playerRaft->move(info.playerInputState.moveDir);
	}

	m_weaponSystem.update(info.dt);


	g_physicsSystem->handleCollision([&](b2Contact* contact, std::set<b2Body*>& destroy)
	{
		b2Body* a = contact->GetFixtureA()->GetBody();
		b2Body* b = contact->GetFixtureB()->GetBody();

		auto ba = (BoatyBase*)a->GetUserData();
		auto bb = (BoatyBase*)b->GetUserData();
		assert(ba);
		assert(bb);

		Debris* debrisA = dynamic_cast<Debris*>(ba);
		Debris* debrisB = dynamic_cast<Debris*>(bb);
		const bool debrisContact = (debrisA != nullptr || debrisB != nullptr);

		if (debrisA && debrisB) {
			return;
		}

		ColorRGBA8 color[] = {
			{ 185, 122, 87 },
			{ 204, 157, 132 },
			{ 105, 65, 44 },
			{ 137, 85, 56 },
			{ 129, 108, 65 },
			{ 163, 136, 82 },
		};

		b2Manifold* manifold = contact->GetManifold();

		bool aDestroy = (manifold->pointCount > 0 && ba->collisionCallback) ? ba->collisionCallback(bb) : false;
		bool bDestroy = (manifold->pointCount > 0 && bb->collisionCallback) ? bb->collisionCallback(ba) : false;

		RaftPart* ra = dynamic_cast<RaftPart*>(ba);
		bool aplayer = ra && ra->m_isPlayer;
		bool anpc = ra && !ra->m_isPlayer;

		RaftPart* rb = dynamic_cast<RaftPart*>(bb);
		bool bplayer = rb && rb->m_isPlayer;
		bool bnpc = rb && !rb->m_isPlayer;

		// Both raft parts belong to the same parent raft.
		const bool selfRaftPartCollision = (ra && rb && ra->m_parent == rb->m_parent);

		Bullet* abullet = dynamic_cast<Bullet*>(ba);
		Bullet* bbullet = dynamic_cast<Bullet*>(bb);


		if ((abullet && rb) || (bbullet && ra))
		{
			Bullet *bullet = abullet ? abullet : bbullet;
			RaftPart *raftPart = ra ? ra : rb;

			raftPart->doDamage(bullet->info.damage, ivec2(rand() % raftPart->m_sprite->w, rand() % raftPart->m_sprite->h));
			bullet->doDamage(-100.0f, ivec2()); // just kill it

			vec2 c = bullet->body->GetPosition();
			m_weaponSystem.spawnExplosion(c);

			// Disarm it for further collisions.
			bullet->info.damage = 0;
		}

		if (aDestroy)
			destroy.insert(a);

		if (bDestroy)
			destroy.insert(b);

		if (ra && rb)
		{
			if (ra->m_parent != rb->m_parent)
			{
				float maxImpulse = 0.0f;

				for (int i = 0; i < manifold->pointCount; i++) 
				{
					const b2ManifoldPoint &point = manifold->points[i];
					if (point.normalImpulse > maxImpulse)
					{
						maxImpulse = point.normalImpulse;
					}
				}

				if (maxImpulse > 1.0f) {
					float m0 = ra->totalMass();
					float m1 = rb->totalMass();
					float d0 = m1 / m0;
					float d1 = m0 / m1;
					ra->doDamage(0.1f * d0, ivec2(rand() % ra->m_sprite->w, rand() % ra->m_sprite->h));
					rb->doDamage(0.1f * d1, ivec2(rand() % rb->m_sprite->w, rand() % rb->m_sprite->h));
				}
			}
		}

		for (int i = 0; i < manifold->pointCount; i++)
		{
			const b2ManifoldPoint &point = manifold->points[i];

			if (point.normalImpulse > 2 && !debrisContact && !selfRaftPartCollision && ((ra && ra->m_isPlayer) || (rb && rb->m_isPlayer)))
			{
				if (!info.shake->running())
				{
					info.shake->start(point.normalImpulse * 30.0f, 100.0f);
				}

				for (int k = 0; k < 20; k++)
				{
					Particle& p = particles.addParticle();
					b2Vec2 pos = (k & 1 ? a : b)->GetPosition();
					vec2 rpos = physCoordToGfx(vec2(pos.x, pos.y));
					p.position = vec3(rpos.x, 0, rpos.y);
					p.color = color[k % 5];
				}
			}
		}
	});

	particles.update(info.dt);

	g_physicsSystem->physicsWorld->Step(1.0f / 60, 20, 20);
}

void WaterWorld::render(Rafterizer& rafterizer, DebrisSystem& debrisSystem)
{
	mat2 boundingXform;
	boundingXform[0] = physCoordToGfx(vec2(1, 0));
	boundingXform[1] = physCoordToGfx(vec2(0, 1));

	for (auto raft : m_rafts)
	{
		vec3 boundingCircle = raft->getBoundingCircle();
		vec2 boundingCenter = vec2(boundingCircle.x, boundingCircle.y);
		//g_primitiveRenderer->drawCircle(boundingXform, boundingCenter, boundingCircle.z, ColorRGBA::Orange());

		vec3 visibilityCircle = raft->getVisibilityCircle();
		vec2 visibilityCenter = vec2(visibilityCircle.x, visibilityCircle.y);
		g_primitiveRenderer->drawCircle(boundingXform, visibilityCenter, visibilityCircle.z, ColorRGBA8(98, 170, 230));

		/*
		g_primitiveRenderer->drawLine(
			physCoordToGfx(raft->getPhysicsPosition()),
			physCoordToGfx(raft->getPhysicsPosition() + raft->m_ai.wanderTarget),
			ColorRGBA(1, 0, 1));
		*/
	}

	rafterizer.reset();

	for (auto raft : m_rafts)
	{
		rafterizer.add(*raft);
	}

	const auto& activeDebris = debrisSystem.activeDebris();
	for (auto debris : activeDebris)
	{
		rafterizer.add(*debris, debrisSystem.getSprite());
	}

	rafterizer.sortElements();
	rafterizer.render();

	particles.draw([&](const Particle& p) {
		g_primitiveRenderer->drawCross(vec2(p.position.x, p.position.z), 30 * p.size.x, p.color);
	});

	m_weaponSystem.draw(Gfx_Immediate_Context(g_device), g_primitiveRenderer);
}

void WaterWorld::renderFlocks(aabb2 viewBounds, PrimitiveBatch* prim, Sprite* sprite0, Sprite* sprite1)
{
	for (auto& flock : m_flocks)
		m_peckFX.render(prim, flock, sprite0, sprite1);
}

void WaterWorld::addRaft(Raft* raft, bool isPlayer)
{
	m_rafts.push_back(raft);
	if (isPlayer) {
		m_playerRaft = raft;
	}
}

void WaterWorld::collectPart(RaftPartBlueprint* blueprint)
{
	assert(blueprint);

	if (m_playerRaft->m_parts.size() < 2) {
		std::vector<std::string> parts;
		parts.push_back("door");
		parts.push_back("fence");
		int i = (int)rand()%(int)parts.size();
		blueprint = g_raftPartLibrary->get(parts[i].c_str());
	}

	m_unattachedParts.push_back(blueprint);
	m_unattachedPartsVersion++;
}
vec2 WaterWorld::getObjectWaveSway(vec2 pos) const
{
	return 
		vec2(0, 1.0f) * m_waterSimulator->evalHeightAtPoint(*m_waterGrid, pos)
	+	vec2(0, sin(pos.y + 0.07f * m_waterSimulator->getTimeIndex()));
}

void WaterWorld::playSmashSound()
{
	if (m_smashPlaying)
		return;

	m_smashPlaying = true;
	jamPlaySound(m_smashSound);
}

void WaterWorld::doneSmashSound()
{
	m_smashPlaying = false;
}
