#include "WeaponSystem.h"
#include "Framework/PhysicsSystem.h"
#include "Framework/PrimitiveBatch.h"
#include "Framework/Sprite.h"
#include "Framework/TextureLibrary.h"

void WeaponSystem::spawnExplosion(vec2 center)
{
	Explosion &e = explosions[explosionCount];
	explosionCount++;

	e.angle = randomUnitFloat() * 2 * b2_pi;
	e.center = center;
	e.animationSet = createAnimationSet(getAnimationSetPrefab("Textures/TestExplosion"));
	e.animationSet->playAnimation(AnimationType::Explosion);
}

void WeaponSystem::shotgun(vec2 center, vec2 target, float spread, int bulletCount, bool isPlayer)
{
	b2Vec2 originRay = glm::normalize(target - center);

	BulletInfo info;
	info.center = center;
	info.numRays = bulletCount;
	info.shotByPlayer = isPlayer;

	explode(info, [&](int idx, int numRays) {
		b2Vec2 diagonal = originRay.Skew();
		b2Vec2 rayDir = originRay - (randomUnitFloat() * spread * diagonal);
		rayDir.Normalize();
		rayDir *= glm::mix(1.0f, randomUnitFloat(), sqrtf(spread));
		return rayDir;
	});
}

void WeaponSystem::grenade(vec2 center)
{
	BulletInfo info;
	info.center = center;

	explode(info, [&](int idx, int numRays) {
		float angle = (idx / (float)numRays);
		angle += randomUnitFloat();
		angle *= 2.0f * (float)M_PI;
		angle = fmodf(angle, 2.0f * (float)M_PI);
		return b2Vec2(sinf(angle), cosf(angle));
	});
}

void WeaponSystem::explode(BulletInfo info, std::function<b2Vec2(int, int)> rayGenerator)
{
	for (int i = 0; i < info.numRays; i++)
	{
		b2Vec2 rayDir = rayGenerator(i, info.numRays);

		Bullet& w = bullets[bulletCount];
		bulletCount++;

		w.info = info;
		w.animationSet = createAnimationSet(getAnimationSetPrefab("Textures/TestBullet"));
		w.animationSet->playAnimation(AnimationType::Explosion);

		b2BodyDef bd;
		bd.type = b2_dynamicBody;
		bd.fixedRotation = true;
		bd.bullet = true;
		bd.linearDamping = info.dampening;
		bd.gravityScale = 0;
		bd.position = b2Vec2(info.center.x, info.center.y);
		bd.linearVelocity = info.blastPower * rayDir;

		b2Body* body = g_physicsSystem->physicsWorld->CreateBody(&bd);
		body->SetUserData(&w);
		w.body = body;

		b2CircleShape circleShape;
		circleShape.m_radius = info.size;

		b2FixtureDef fd;
		fd.shape = &circleShape;
		fd.density = info.density / (float)info.numRays;
		fd.friction = 0;
		fd.restitution = info.restitution;
		fd.filter.groupIndex = -1; // todo: filter out shooting entity as well
		fd.filter.categoryBits = info.shotByPlayer
			? COLLISION_MASK_NPC
			: COLLISION_MASK_PLAYER;
		w.fixture = body->CreateFixture(&fd);
	}
}

void WeaponSystem::update(float dt)
{
	for (int i = 0; i < bulletCount; i++)
	{
		Bullet& w = bullets[i];
		if (!w.body->IsAwake() || w.getHealth() < 0 || glm::length(vec2(w.body->GetLinearVelocity())) < 1.f)
		{
			g_physicsSystem->physicsWorld->DestroyBody(w.body);

			bulletCount--;
			bullets[i] = bullets[bulletCount];
		}
	}

	for (int i = 0; i < explosionCount; i++)
	{
		Explosion& e = explosions[i];
		if (e.animationSet->playQueue.empty())
		{
			explosionCount--;
			explosions[i] = explosions[explosionCount];
		}
	}
}

void WeaponSystem::draw(GfxContext* ctx, PrimitiveBatch* prim)
{
	for (int i = 0; i < bulletCount; i++)
	{
		Bullet& w = bullets[i];
		vec2 coord = physCoordToGfx(w.body->GetPosition());
		vec2 d (w.body->GetLinearVelocity());
		d = normalize(d);

		//magic scale factor by pi/4 because sprite is pointed up and box2d & rotate outwards
		float a = atan2(d.y, d.x) + (float)M_PI / 4.0f + (float)M_PI;

		auto sprite = w.animationSet->getCurrentFrame().layers[int(AnimationFrameLayer::Color)];
		drawSprite(g_primitiveRenderer, *sprite, coord, a);
	}

	for (int i = 0; i < explosionCount; i++)
	{
		Explosion& e = explosions[i];
		vec2 coord = physCoordToGfx(e.center);
		auto sprite = e.animationSet->getCurrentFrame().layers[int(AnimationFrameLayer::Color)];

		drawSprite(g_primitiveRenderer, *sprite, coord, e.angle);
	}
}