#include "BitmapFont.h"
#include "Stream.h"
#include "Graphics.h"

#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <algorithm>

enum class BlockType
{
	Info	= 1,
	Common	= 2,
	Pages	= 3,
	Chars	= 4,
	Kerning	= 5
};

#pragma pack(push)
#pragma pack(1)
	
struct InfoBlock
{
	s16	font_size;
 	u8	flags;
	u8	char_set;
	u16	stretch_h;
	u8	antialias;
	u8	padding_up;
	u8	padding_right;
	u8	padding_down;
	u8	padding_left;
	u8	padding_horiz;
	u8	padding_vert;
	u8	outline;
	char	name[1];
};

struct CommonBlock
{
	u16	line_height;
	u16	base;
	u16	scale_width;
	u16	scale_height;
	u16	pages;
	u8	bitfield;
	u8	flags;
	u8	chan_alpha;
	u8	chan_red;
	u8	chan_green;
	u8	chan_blue;
};

struct PagesBlock
{
	char names[1];
};

struct CharInfo
{
	u32	id;
	u16	x;
	u16	y;
	u16	width;
	u16	height;
	s16	offset_x;
	s16	offset_y;
	s16	advance_x;
	u8	page;
	u8	chan;
};

struct CharsBlock
{
	CharInfo chars[1];
};

struct KerningPair
{
	u32 first;
	u32 second;
	u16 amount;
};

struct KerningBlock
{
	KerningPair pairs[1];
};

#pragma pack(pop)

#define ENDIAN_SWAP(x) {(void)sizeof(x);}// nope

static void endian_swap(InfoBlock& x)
{
	ENDIAN_SWAP(x.font_size);
	ENDIAN_SWAP(x.stretch_h);
}

static void endian_swap(CommonBlock& x)
{
	ENDIAN_SWAP(x.line_height);
	ENDIAN_SWAP(x.base);
	ENDIAN_SWAP(x.scale_width);
	ENDIAN_SWAP(x.scale_height);
	ENDIAN_SWAP(x.pages);
}

static void endian_swap(CharInfo& x)
{
	ENDIAN_SWAP(x.id);
	ENDIAN_SWAP(x.x);
	ENDIAN_SWAP(x.y);
	ENDIAN_SWAP(x.width);
	ENDIAN_SWAP(x.height);
	ENDIAN_SWAP(x.offset_x);
	ENDIAN_SWAP(x.offset_y);
	ENDIAN_SWAP(x.advance_x);
}

static void endian_swap(KerningPair& x)
{
	ENDIAN_SWAP(x.first);
	ENDIAN_SWAP(x.second);
	ENDIAN_SWAP(x.amount);
}

inline u32 FourCC(u8 ch0, u8 ch1, u8 ch2, u8 ch3)
{
	return (u32(ch0) | (u32(ch1) << 8) | (u32(ch2) << 16) | (u32(ch3) << 24));
}


bool BitmapFont::read(Stream& stream)
{	
	u32 magic;
	stream.read(&magic, 4);

	bool need_endian_swap = false;
	bool magic_found = false;

	u32 magic_native_endian  = FourCC('B','M','F',3);
	u32 magic_swapped_endian = magic_native_endian; 
	ENDIAN_SWAP(magic_swapped_endian);

	if( magic == magic_native_endian )
	{
		magic_found = true;
		need_endian_swap = false;
	}
	else if( magic == magic_swapped_endian )
	{
		magic_found = true;
		need_endian_swap = true;
	}

	if( magic_found == false )
	{
		SDL_LogError(SDL_LOG_CATEGORY_SYSTEM, "BitmapFont::read() BMF file identifier not found.");
		return false;
	}

	InfoBlock*		info	= nullptr;
	CommonBlock*	common	= nullptr;
	PagesBlock*		pages	= nullptr;
	CharsBlock*		chars	= nullptr;
	KerningBlock*	kerning	= nullptr;
		
	u32 num_chars  = 0;
	u32 num_kpairs = 0;
	u32 num_pages  = 0;
	u32 page_len   = 0;
		
	u8 type;
	while( stream.read(&type, 1) )
	{
		u32 block_size;
		stream.read(&block_size, 4);

		if(need_endian_swap) ENDIAN_SWAP(block_size);

		BlockType block_type = (BlockType)type;

		switch(block_type)
		{
		case BlockType::Info:
			info = (InfoBlock*)malloc(block_size);
			if( info )
			{
				stream.read(info, block_size);
			}
			break;
		case BlockType::Common:
			common = (CommonBlock*)malloc(block_size);
			if( common )
			{
				stream.read(common, block_size);
			}
			break;
		case BlockType::Pages:
			pages = (PagesBlock*)malloc(block_size);
			if( pages )
			{
				stream.read(pages, block_size);
				page_len = (u32)strlen(pages->names);
				num_pages = block_size / (page_len+1);
			}
			break;
		case BlockType::Chars:
			chars = (CharsBlock*)malloc(block_size);
			if( chars )
			{
				stream.read(chars, block_size);
				num_chars = block_size / sizeof(CharInfo);
			}
			break;
		case BlockType::Kerning:
			kerning = (KerningBlock*)malloc(block_size);
			if( kerning )
			{
				stream.read(kerning, block_size);
				num_kpairs = block_size / sizeof(KerningPair);
			}
			break;
		}
	}

	if( info && pages && chars )
	{
		if( need_endian_swap )
		{
			endian_swap(*info);
			endian_swap(*common);
			for(u32 i=0; i<num_chars; ++i) endian_swap(chars->chars[i]);
			if( kerning )
			{
				for(u32 i=0; i<num_kpairs; ++i) endian_swap(kerning->pairs[i]);
			}
		}

		m_size = (u32)abs((float)info->font_size);

		// handle page info
		m_num_pages = num_pages;
		u32 clamped_page_len = std::min<u32>(page_len, sizeof(PageData)-1);
		for( u32 i=0; i<m_num_pages; ++i )
		{
			memcpy(m_pages[i].filename, (char*)pages + (page_len+1)*i, clamped_page_len);
			m_pages[i].filename[clamped_page_len] = 0;
		}

		// handle char into
		for( u32 i=0; i<256; ++i )
		{
			memset(&m_chars[i], 0, sizeof(CharData));
			m_chars[i].page = InvalidPage;
		}

		for( u32 i=0; i<num_chars; ++i )
		{
			u32 c = chars->chars[i].id;

			m_chars[c].x         = chars->chars[i].x;
			m_chars[c].y         = chars->chars[i].y;
			m_chars[c].width     = chars->chars[i].width;
			m_chars[c].height    = chars->chars[i].height;
			m_chars[c].offset_x  = chars->chars[i].offset_x;
			m_chars[c].offset_y  = chars->chars[i].offset_y;
			m_chars[c].advance_x = chars->chars[i].advance_x;
			m_chars[c].page      = chars->chars[i].page;
			m_chars[c].chan      = chars->chars[i].chan;
		}
	}
	else
	{
		SDL_LogError(SDL_LOG_CATEGORY_SYSTEM, "BitmapFont::read() Failed to load font data (info, pages or chars).");
	}

	free(info);
	free(common);
	free(pages);
	free(chars);
	free(kerning);

	return true;
}

BitmapFont::BitmapFont()
	: m_num_pages(0)
	, m_size(0)
{
	memset(&m_pages, 0, sizeof(m_pages));
}

BitmapFontRenderer::BitmapFontRenderer(const BitmapFontData& data)
{
	init(data.header_data, data.header_size, data.pixels_data, data.pixels_size, data.width, data.height, data.format);
	createQuads();
}

BitmapFontRenderer::BitmapFontRenderer(const void* header_data, size_t header_size, const void* pixels_data, size_t pixels_size, u32 width, u32 height, GfxFormat format)
{
	init(header_data, header_size, pixels_data, pixels_size, width, height, format);
	createQuads();
}

BitmapFontRenderer::BitmapFontRenderer(const char* font_filename, const char* texture_path)
{
	FileIn fi(font_filename);
	if (fi.valid())
	{
		m_font_data.read(fi);
		for (u32 i = 0; i < m_font_data.getPageCount(); ++i)
		{
			const char* page_filename = m_font_data.getPage(i);
			std::string page_path = texture_path;
			page_path += "/";
			page_path += page_filename;
			page_path = jamTranslatePath(page_path);

			TextureHandle th = Gfx_Texture_Load(g_device, page_path.c_str());
			m_textures.push_back(th);

			TextureDescr td = Gfx_Texture_Descr(g_device, th);
			m_texture_descr.push_back(td);
		}

		createQuads();
	}
	else
	{
		SDL_LogError(SDL_LOG_CATEGORY_SYSTEM, "Failed to load bitmap font '%s' (texture path '%s')", font_filename, texture_path);
	}
}

BitmapFontRenderer::~BitmapFontRenderer()
{
	for (size_t i = 0; i < m_textures.size(); ++i)
	{
		Gfx_Texture_Destroy(g_device, m_textures[i]);
	}
}
vec2 BitmapFontRenderer::drawString(PrimitiveBatch* batch, const vec2& pos, const char* str, ColorRGBA8 col, bool flush)
{
	if (flush) batch->flush();

	if (m_textures.size() == 0) return pos;

	float x = pos.x;
	float y = pos.y - m_font_data.getSize();
	float orig_x = x;
	const TextureDescr& t = Gfx_Texture_Descr(g_device, m_textures[0]);
	vec2 page_size(float(t.width), float(t.height));
	//vec2 page_size_rcp = page_size.reciprocal();

	while (*str)
	{
		u8 c = *str;
		if (c == '\n')
		{
			y -= m_font_data.getSize();
			x = orig_x;
		}
		else
		{
			const auto& data = m_font_data.getChar(c);
			if (data.page != BitmapFont::InvalidPage)
			{
				TexturedQuad2D q = m_chars[c];

				q.pos[0] += vec2(x, y);
				q.pos[1] += vec2(x, y);
				q.pos[2] += vec2(x, y);
				q.pos[3] += vec2(x, y);

				batch->setTexture(m_textures[data.page]);
				batch->drawTexturedQuad(&q, col);
				x += float(data.advance_x);
			}
		}
		++str;
	}

	if (flush) batch->flush();

	return vec2(x, y);
}

vec2 BitmapFontRenderer::measureString(const char* str)
{
	if (m_textures.size() == 0) return vec2(0, 0);

	float x = 0;
	float y = 0;
	float orig_x = x;
	float max_x = 0;
	TextureDescr t = Gfx_Texture_Descr(g_device, m_textures[0]);
	vec2 page_size(float(t.width), float(t.height));
	//vec2 page_size_rcp = page_size.reciprocal();

	float char_w = 0;
	float char_h = float(m_font_data.getSize());

	while (*str)
	{
		u8 c = *str;
		if (c == '\n')
		{
			max_x = std::max(max_x, x);
			y += char_h;
			x = orig_x;
		}
		else
		{
			const BitmapFont::CharData& data = m_font_data.getChar(c);
			if (data.page != BitmapFont::InvalidPage)
			{
				char_w = data.advance_x;
				x += float(char_w);
			}
		}
		++str;
	}
	max_x = std::max(max_x, x);

	return vec2(max_x, y + char_h);

}

void BitmapFontRenderer::init(const void *header_data, size_t header_size, const void *pixels_data, size_t pixels_size, u32 width, u32 height, GfxFormat format)
{
	MemStream fi((void*)header_data, (u32)header_size);
	m_font_data.read(fi);

	TextureDescr td = TextureDescr::create_2D(width, height, format, false, false);
	m_texture_descr.push_back(td);

	TextureHandle th = Gfx_Texture_Create(g_device, td, pixels_data, pixels_size);
	m_textures.push_back(th);
}

void BitmapFontRenderer::createQuads()
{
	for (u8 c = 0; c < 255; ++c)
	{
		u32 size = m_font_data.getSize();

		const BitmapFont::CharData& data = m_font_data.getChar(c);
		if (data.page != BitmapFont::InvalidPage)
		{
			TextureDescr& texture_descr = m_texture_descr[data.page];
			vec2 page_size(float(texture_descr.width), float(texture_descr.height));

			float w = float(data.width);
			float h = float(data.height);

			float px = float(data.offset_x);
			float py = float(data.offset_y);

			float tx = float(data.x) / page_size.x;
			float ty = float(data.y) / page_size.y;
			float tw = float(data.width) / page_size.x;
			float th = float(data.height) / page_size.y;

			TexturedQuad2D& q = m_chars[c];

			q.pos[0] = vec2(px, py);
			q.pos[1] = vec2(px + w, py);
			q.pos[2] = vec2(px + w, py + h);
			q.pos[3] = vec2(px, py + h);

			q.tex[0] = vec2(tx, ty);
			q.tex[1] = vec2(tx + tw, ty);
			q.tex[2] = vec2(tx + tw, ty + th);
			q.tex[3] = vec2(tx, ty + th);

			for (vec2& p : q.pos)
			{
				p.y = size - p.y;
			}

			for (vec2& t : q.tex)
			{
				t.y = 1.0f - t.y;
			}
		}
	}
}
