#pragma once
#include "Reflection.h"
#include <unordered_map>
#include <iostream>
#include <functional>

#if 0
struct Entity {
	template <typename T>
	T& comp() {
		return T::getForEntity(id);
	};

	template <typename T>
	T* findComp() {
		return T::findForEntity(id);
	};

	static Entity fromId(u32 id) {
		Entity ent;
		ent.id = id;
		return ent;
	}

	u32 getId() const {
		return id;
	}

	bool isValid() const {
		return id != 0xffffffffu;
	}

private:
	u32 id = 0xffffffffu;
};

// --------------------------------------------------------

template <typename Comp, typename... Ts>
struct DependencyInit {
	void initialize(Comp*, Entity) {}
};

template <typename Comp, typename T, typename... Ts>
struct DependencyInit<Comp, T, Ts...> : DependencyInit<Comp, Ts...> {
	void initialize(Comp* thiscomp, Entity ent) {
		T* component = ent.findComp<T>();
		if (!component) {
			SDL_LogCritical(SDL_LOG_CATEGORY_APPLICATION, "Dependent component %s not found for %s!", T::info().name(), Comp::info().name());
			abort();
		}
		thiscomp->dependencies.template set<T>(component);

		DependencyInit<Comp, Ts...>::initialize(thiscomp, ent);
	}
};

template <typename Comp, typename ... T>
struct DependencyStore : DependencyInit<Comp, T...> {
	std::tuple<T*...> pointers;

	template <typename X>
	X& as() {
		return *std::get<X*>(pointers);
	}

	template <typename X>
	const X& as() const {
		return *std::get<X*>(pointers);
	}

	template <typename X>
	void set(void* ptr) {
		std::get<X*>(pointers) = (X*)ptr;
	}
};

struct IComponentFactory;
struct IComponent : IReflectionSupport {
	virtual IComponentFactory* createFactory() const = 0;
	virtual IComponentFactory* createInitializingFactory() const = 0;
	virtual IComponent* createForEntity(Entity ent) const = 0;
	virtual bool drawEditorGui() = 0;
	virtual Entity getEntity() const = 0;

	virtual void onEdited() {}
	virtual void onStart() {}
};

bool drawDefaultComponentEditorGui(IComponent*);

struct ComponentInfo {
	std::unordered_map<u32, IComponent*> entityComponentMap;
	std::vector<IComponent*> instanceList;
};

template <typename T>
struct ComponentInfoT {
	std::unordered_map<u32, T*> entityComponentMap;
	std::vector<T*> instanceList;
};

template <typename T, typename ... Dependencies>
struct ComponentT : IComponent {
	DependencyStore<T, Dependencies ...> dependencies;
	Entity owningEntity;

	template <typename X>
	X& as() { return dependencies. template as<X>(); }

	template <typename X>
	const X& as() const { return dependencies. template as<X>(); }

	static T* findForEntity(u32 id) {
		auto it = T::componentInfo.entityComponentMap.find(id);
		if (it != T::componentInfo.entityComponentMap.end()) {
			return (T*)it->second;
		}
		else {
			return nullptr;
		}
	}

	static T& getForEntity(u32 id) {
		T* comp = findForEntity(id);
		if (!comp) abort();
		return *comp;
	}

	const ReflectionBase& getReflection() const override {
		return T::info();
	}

	virtual bool drawEditorGui() override {
		return drawDefaultComponentEditorGui(this);
	}

	Entity getEntity() const override {
		return owningEntity;
	}
};

std::vector<ComponentInfo*>& g_componentInfo();
std::unordered_map<std::string, std::function<IComponent*()>>& g_componentProtoFactories();


#define DECLARE_COMPONENT(ComponentName, ...) \
	static ComponentInfoT<ComponentName> componentInfo; \
	static u32 componentId; \
	/* if random access is not required, this should be a chunked linked list */ \
	/* or maybe a flat array, but then pointers to components can't be retained */ \
	IComponentFactory* createFactory() const override; \
	IComponentFactory* createInitializingFactory() const override; \
	IComponent* createForEntity(Entity ent) const override; \
	RTTI(ComponentName, __VA_ARGS__)

#define DEFINE_COMPONENT(ComponentName) \
	ComponentInfoT<ComponentName> ComponentName::componentInfo; \
	u32 ComponentName::componentId = registerComponent<ComponentName>(); \
	IComponentFactory* ComponentName::createFactory() const { return new ComponentFactory<ComponentName>(ComponentName()); }; \
	IComponentFactory* ComponentName::createInitializingFactory() const { return new ComponentFactory<ComponentName>(*this); }; \
	IComponent* ComponentName::createForEntity(Entity ent) const { return ComponentFactory<ComponentName>::createDefaultComponentForEntity(ent); }

// Fn accepts IComponent*
template <typename CallbackType>
void iterateEntityComponents(Entity ent, CallbackType callback)
{
	for (ComponentInfo* componentInfo : g_componentInfo()) {
		auto compIt = componentInfo->entityComponentMap.find(ent.getId());
		if (compIt != componentInfo->entityComponentMap.end()) {
			callback(compIt->second);
		}
	}
}

template <typename Fn>
void iterateAllComponents(Fn fn) {
	for (Entity ent : g_entities) {
		iterateEntityComponents(ent, fn);
	}
}

struct IComponentFactory {
	virtual void createComponent(Entity ent) = 0;
	virtual ReflectionBase& componentInfo() const = 0;
	virtual cJSON* marshall() const = 0;
	virtual const IComponent* getInitiailzer() const = 0;
};

inline IComponent* createComponentProtoByName(const std::string& name) {
	return g_componentProtoFactories()[name]();
}


template <typename T>
struct ComponentFactory : IComponentFactory {
	T initVal;

	ComponentFactory() {};

	explicit ComponentFactory(const T& initVal)
		: initVal(initVal)
	{}

	static T* createDefaultComponentForEntity(Entity ent) {
		auto existing = T::componentInfo.entityComponentMap.find(ent.getId());
		if (T::componentInfo.entityComponentMap.end() == existing) {
			T* comp = new T;
			comp->owningEntity = ent;
			T::componentInfo.entityComponentMap[ent.getId()] = comp;
			T::componentInfo.instanceList.push_back(comp);
			comp->dependencies.initialize(comp, ent);
			return comp;
		}
		else {
			return existing->second;
		}
	}

	void createComponent(Entity ent) override {
		T* comp = new T(initVal);
		comp->owningEntity = ent;
		T::componentInfo.entityComponentMap[ent.getId()] = comp;
		T::componentInfo.instanceList.push_back(comp);
		comp->dependencies.initialize(comp, ent);
	}

	ReflectionBase& componentInfo() const override {
		return T::info();
	}

	cJSON* marshall() const override {
		cJSON* res = cJSON_CreateObject();
		cJSON_AddItemToObject(res, "componentName", cJSON_CreateString(componentInfo().name()));
		cJSON_AddItemToObject(res, "initializer", marshallObjectFields(initVal));
		return res;
	}

	const IComponent* getInitiailzer() const override {
		return &initVal;
	}
};

template <typename T>
u32 registerComponent() {
	u32 id = u32(g_componentInfo().size());
	g_componentInfo().push_back((ComponentInfo*)&T::componentInfo);
	g_componentProtoFactories()[T::info().name()] = []() -> IComponent* { return new T; };
	return id;
}


extern std::vector<Entity> g_entities;
extern std::vector<int> g_entityConnectedPrefabs;

struct EntityPool {
	Entity allocate() {
		Entity ent = Entity::fromId(next++);
		g_entities.push_back(ent);
		g_entityConnectedPrefabs.push_back(-1);
		return ent;
	}

	Entity allocateConnectedToPrefab(int prefabId) {
		Entity ent = Entity::fromId(next++);
		g_entities.push_back(ent);
		g_entityConnectedPrefabs.push_back(prefabId);
		return ent;
	}

	u32 next = 0;
};

extern EntityPool g_entityPool;

struct EntityPrefab {
	const std::string& getName() {
		return name;
	}

	template <typename T>
	void addComponent() {
		componentFactories.push_back(new ComponentFactory<T>(T()));
	}

	template <typename T>
	void addComponent(const T& initVal) {
		componentFactories.push_back(new ComponentFactory<T>(initVal));
	}

	Entity create();

	cJSON* marshall() const {
		cJSON* prefab = cJSON_CreateArray();
		for (auto it : componentFactories) {
			cJSON_AddItemToArray(prefab, it->marshall());
		}
		return prefab;
	}

	void demarshall(cJSON *const json) {
		const int componentCount = cJSON_GetArraySize(json);
		for (int i = 0; i < componentCount; ++i) {
			cJSON* prefabJson = cJSON_GetArrayItem(json, i);
			IComponentFactory* factory = nullptr;
			IComponent* initializer = createComponentProtoByName(cJSON_GetObjectItem(prefabJson, "componentName")->valuestring);

			if (cJSON* initializerJson = cJSON_GetObjectItem(prefabJson, "initializer")) {
				demarshallObjectFields(initializerJson, initializer, initializer->getReflection());
				factory = initializer->createInitializingFactory();
			} else {
				factory = initializer->createFactory();
			}

			componentFactories.push_back(factory);
		}
	}

	friend EntityPrefab* createPrefab(const std::string& name);

	const IComponent* getComponentInitializer(const ReflectionBase* componentType) {
		for (IComponentFactory* fact : componentFactories) {
			if (&fact->componentInfo() == componentType) {
				return fact->getInitiailzer();
			}
		}

		return nullptr;
	}

private:
	std::vector<IComponentFactory*> componentFactories;
	int thisPrefabId;
	std::string name;

	EntityPrefab() = delete;
	EntityPrefab(int thisPrefabId) 
		: thisPrefabId(thisPrefabId)
	{}
};

template <typename T>
struct function_traits
	: public function_traits<decltype(&T::operator())>
{};
// For generic types, directly use the result of the signature of its 'operator()'

template <typename ClassType, typename ReturnType, typename... Args>
struct function_traits<ReturnType(ClassType::*)(Args...) const>
	// we specialize for pointers to member function
{
	enum { arity = sizeof...(Args) };
	// arity is the number of arguments.

	typedef ReturnType result_type;

	template <size_t i>
	struct arg
	{
		typedef typename std::tuple_element<i, std::tuple<Args...>>::type type;
		// the i-th argument is equivalent to the i-th tuple element of a tuple
		// composed of those arguments.
	};
};

/* Does not compile with clang / gcc
template <typename T>
inline void withComp(Entity ent, T componentHandler) {
	componentHandler(ent.comp<std::decay<function_traits<T>::arg<0>::type>::type>());
}
*/

template <typename T>
T* createComponent(Entity ent) {
	return ComponentFactory<T>::createDefaultComponentForEntity(ent);
}

template <typename T>
inline void dumpObject(const T& obj) {
	dumpObject(&obj, obj.info());
}

void dumpObject(const void *const obj, const ReflectionBase& info);
void dumpEntity(Entity ent);
void dumpAllEntities();

cJSON* marshallEntity(Entity ent);
cJSON* marshallAllEntities();
void demarshallEntities(cJSON* json);

extern std::unordered_map<std::string, EntityPrefab*> g_entityPrefabs;
extern std::vector<EntityPrefab*> g_entityPrefabsById;

EntityPrefab* createPrefab(const std::string& name);
cJSON* marshallAllPrefabs();
void demarshallPrefabs(cJSON*);
#endif
