#include "Framework.h"
#include "Graphics.h"
#include "Gui.h"
#include "PrimitiveBatch.h"

#define STB_IMAGE_IMPLEMENTATION
#pragma warning(push, 0)
#include "stb_image.h"
#pragma warning(pop)

#include <mutex>

#include <libxmp-lite/xmp.h>

#ifdef _WIN32
#include <windows.h>
#endif

#include <chrono>

static JamConfig s_config;

SDL_Window* g_window = nullptr;
SDL_Renderer* g_renderer = nullptr;
GfxDevice* g_device = nullptr;
GameWorld* g_world = nullptr;
PrimitiveBatch* g_primitiveRenderer = nullptr;
bool g_enableSound = true;
bool g_enableEditor = false;
float g_deltaTime = 1.0f / 60.0f; // updated during game loop
float g_elapsedTime = 0.0f; // updated during game loop

struct MixerChannel
{
	bool playing = false;
	bool loop = false;
	SoundClip clip;
	u32 pos = 0;
	SDL_AudioCallback callback = nullptr;
	void* userData = nullptr;
};

struct Mixer
{
	enum { MixerChannelCount = 16 };
	bool enabled = false;
	SDL_AudioSpec specDesired;
	SDL_AudioSpec specObtained;
	MixerChannel channels[MixerChannelCount];
	int volume = SDL_MIX_MAXVOLUME;
	std::mutex mutex;
};

static Mixer s_mixer;

static void jamAudioCallback(void *userdata, Uint8 *stream, int len);

char * jamTranslatePath(const char *filename)
{
	char path[512];

	path[0] = 0;
	SDL_strlcat(path, s_config.dataDirectory, sizeof path);
	SDL_strlcat(path, filename, sizeof path);

	return SDL_strdup(path);
}

std::string jamTranslatePath(const std::string& filename)
{
	return std::string(s_config.dataDirectory) + filename;
}

SoundClip jamLoadSoundClip(const char* filename)
{
	char *realFilename = jamTranslatePath(filename);

	SoundClip clip = {};

	if (SDL_LoadWAV(realFilename, &clip.spec, &clip.buffer, &clip.length) == nullptr)
	{
		SDL_free(realFilename);
		return clip;
	}

	SDL_AudioCVT cvt;
	SDL_BuildAudioCVT(&cvt,
		clip.spec.format, clip.spec.channels, clip.spec.freq,
		s_mixer.specObtained.format, s_mixer.specObtained.channels, s_mixer.specObtained.freq);
	
	if (cvt.needed)
	{
		cvt.len = clip.length;
		cvt.buf = (Uint8*)SDL_malloc(cvt.len * cvt.len_mult);
		SDL_memcpy(cvt.buf, clip.buffer, clip.length);
		int err = SDL_ConvertAudio(&cvt);

		SDL_free(clip.buffer);
		clip.buffer = cvt.buf;

		clip.spec.freq = s_mixer.specObtained.freq;
		clip.spec.format = s_mixer.specObtained.format;
		clip.spec.channels = s_mixer.specObtained.channels;
		clip.spec.size = cvt.len_cvt;
		clip.length = cvt.len_cvt;

		printf("Converted audio format for '%s': %d\n", filename, err);
	}

	SDL_free(realFilename);

	return clip;
}


SDL_Texture * jamLoadTexture(const char *filename, Uint32 colorKey)
{
	char *textureFile = jamTranslatePath(filename);

	int width, height, components;

	u8 *const pixels = stbi_load(textureFile, &width, &height, &components, 4);
	SDL_free(textureFile);

	SDL_assert(pixels && "Can't load texture");

	SDL_Texture *const texture = SDL_CreateTexture(
		g_renderer, SDL_PIXELFORMAT_ABGR8888, SDL_TEXTUREACCESS_STATIC, width, height
	);
	SDL_assert(texture && "Failed to create texture");

	if (pixels)
	{
		if (colorKey)
		{
			const u32 numPixels = width * height;
			for (u32 i = 0; i < numPixels; ++i)
			{
				u8  r = pixels[i * 4 + 0];
				u8  g = pixels[i * 4 + 1];
				u8  b = pixels[i * 4 + 2];
				u8& a = pixels[i * 4 + 3];

				if (((u32(r) << 16u) + (u32(g) << 8u) + u32(b)) == colorKey) {
					a = 0u;
				}
			}
		}

		if (colorKey || 4 == components)
		{
			SDL_SetTextureBlendMode(texture, SDL_BLENDMODE_BLEND);
		}

		const int updateStatus = SDL_UpdateTexture(texture, nullptr, pixels, width * 4);
		SDL_assert(0 == updateStatus);

		stbi_image_free(pixels);
	}

	return texture;
}

#if EMSCRIPTEN
static void LogOutputFunction(void* userdata, int category, SDL_LogPriority priority, const char* message)
{
	(void)userdata;
	(void)category;
	(void)priority;

	static int id = 0;
	printf("SDL %d: %s\n", id++, message);
}
#endif

static const char* audioFormatToString(int id)
{
	switch (id)
	{
	case AUDIO_U8: return "AUDIO_U8";
	case AUDIO_S8: return "AUDIO_S8";
	case AUDIO_U16LSB: return "AUDIO_U16LSB";
	case AUDIO_S16LSB: return "AUDIO_S16LSB";
	case AUDIO_U16MSB: return "AUDIO_U16MSB";
	case AUDIO_S16MSB: return "AUDIO_S16MSB";
	//case AUDIO_U16: return "AUDIO_U16";
	//case AUDIO_S16: return "AUDIO_S16";
	case AUDIO_S32LSB: return "AUDIO_S32LSB";
	case AUDIO_S32MSB: return "AUDIO_S32MSB";
	//case AUDIO_S32: return "AUDIO_S32";
	case AUDIO_F32LSB: return "AUDIO_F32LSB";
	case AUDIO_F32MSB: return "AUDIO_F32MSB";
	//case AUDIO_F32: return "AUDIO_F32";
	}
	return "UNKNOWN";
}

static void jamStartupInternal(const JamConfig& config)
{	
	printf("SDL Init\n");

	s_config = config;

	SDL_LogSetAllPriority(SDL_LOG_PRIORITY_VERBOSE);

#if EMSCRIPTEN

	printf("Setting Emscripten log output function\n");
	SDL_LogSetOutputFunction(LogOutputFunction, nullptr);

	SDL_Init(SDL_INIT_VIDEO);

	SDL_GL_SetAttribute(SDL_GL_ACCELERATED_VISUAL, 1);
	SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);

	SDL_CreateWindowAndRenderer(config.width, config.height, 0, &g_window, &g_renderer);

#else

#ifdef _WIN32
	if (config.dpiAware)
	{
		SetProcessDPIAware();
	}
#endif

	uint32_t initFlags = SDL_INIT_EVERYTHING & (~(SDL_INIT_HAPTIC | SDL_INIT_AUDIO));
	if (SDL_Init(initFlags) != 0)
	{
		const char* errorMsg = SDL_GetError();
		SDL_LogError(SDL_LOG_CATEGORY_SYSTEM, "SDL_Init failed: %s", errorMsg);
		exit(1);
	}

	SDL_GL_SetAttribute(SDL_GL_ACCELERATED_VISUAL, 1);
	SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);

	SDL_assert(!g_window);
	SDL_assert(!g_renderer);

	g_window = SDL_CreateWindow(config.name, 100, 100, config.width, config.height,
								SDL_WINDOW_SHOWN
								| SDL_WINDOW_OPENGL
								//| SDL_WINDOW_INPUT_GRABBED
								| (config.resizable ? SDL_WINDOW_RESIZABLE : 0)
								);

	g_renderer = SDL_CreateRenderer(g_window, -1, SDL_RENDERER_ACCELERATED | SDL_RENDERER_PRESENTVSYNC);
	if (g_renderer == nullptr)
	{
		SDL_LogError(SDL_LOG_CATEGORY_VIDEO, "SDL_CreateRenderer failed");
		exit(1);
	}

#endif

	// initialize sound

	s_mixer.specDesired.freq = 44100;
	s_mixer.specDesired.format = AUDIO_S16LSB;
	s_mixer.specDesired.channels = 2;
	s_mixer.specDesired.silence = 0;
	s_mixer.specDesired.samples = 1024;
	s_mixer.specDesired.size = SDL_AUDIO_BITSIZE(s_mixer.specDesired.format) *s_mixer.specDesired.channels * s_mixer.specDesired.samples / 8;
	s_mixer.specDesired.callback = jamAudioCallback;
	s_mixer.specDesired.userdata = nullptr;

	if (SDL_OpenAudio(&s_mixer.specDesired, &s_mixer.specObtained) == 0)
	{
		printf("Audio successfully initialized\n");
		printf("Freq: %d\n", s_mixer.specObtained.freq);
		printf("Format: %s (%d) \n", audioFormatToString(s_mixer.specObtained.format), s_mixer.specObtained.format);
		printf("Channels: %d\n", s_mixer.specObtained.channels);
		printf("Samples: %d\n", s_mixer.specObtained.samples);
		printf("Size: %d\n", s_mixer.specObtained.size);
		s_mixer.enabled = true;
		SDL_PauseAudio(0);
	}
	else
	{
		printf("Failed to initialize audio. This is non-fatal, proceeding without sound.\n");
	}

	// initialize low-level graphics api
	printf("Initializing low-level GfxAPI\n");
	GfxDeviceConfig gfxConfig; // not used, so defaults are fine
	g_device = Gfx_CreateDevice(g_window, gfxConfig);

	g_primitiveRenderer = new PrimitiveBatch(g_device);

	SDL_GL_SetSwapInterval(1);

	printf("Initializing ImGui\n");
	initGui();

	printf("Startup done!\n");
}

static void jamShutdownInternal()
{
	SDL_PauseAudio(1);

	if (g_primitiveRenderer)
	{
		delete g_primitiveRenderer;
		g_primitiveRenderer = nullptr;
	}

	if (g_device)
	{
		Gfx_DestroyDevice(g_device);
		g_device = nullptr;
	}

	if (g_renderer)
	{
		SDL_DestroyRenderer(g_renderer);
		g_renderer = nullptr;
	}

	if (g_window)
	{
		SDL_DestroyWindow(g_window);
		g_window = nullptr;
	}

	if (s_mixer.enabled)
	{
		SDL_CloseAudio();
		s_mixer.enabled = false;
	}

	SDL_Quit();
}

void jamShutdown()
{
	jamShutdownInternal();
}

void jamStartup(const JamConfig& config)
{
	jamStartupInternal(config);
}

void jamUpdateGuiBegin()
{
	updateGuiBegin();
}

bool jamGuiEvent(SDL_Event& event)
{
	return updateGuiEvent(event);
}

void jamUpdateGuiEnd(PrimitiveBatch *prim)
{
	if (prim)
	{
		updateGuiEnd(prim);
	}
}

int jamPlaySoundWithCallback(SDL_AudioCallback callback, void* userData)
{
	int result = -1;

	if (!s_mixer.enabled)
		return result;

	NOT_EMSCRIPTEN(s_mixer.mutex.lock());

	for (MixerChannel& channel : s_mixer.channels)
	{
		++result;

		if (channel.playing)
			continue;

		channel.playing = true;
		channel.callback = callback;
		channel.userData = userData;

		break;
	}

	NOT_EMSCRIPTEN(s_mixer.mutex.unlock());

	return result;
}


int jamPlaySound(SoundClip clip, bool loop)
{
	int result = -1;
	
	if(clip.length == 0)
		return result;

	if(!s_mixer.enabled)
		return result;
	
	if (!g_enableSound)
		return result;
	
	NOT_EMSCRIPTEN(s_mixer.mutex.lock());

	SDL_assert(clip.spec.freq == s_mixer.specObtained.freq);
	SDL_assert(clip.spec.channels == s_mixer.specObtained.channels);
	SDL_assert(clip.spec.format == s_mixer.specObtained.format);

	for (MixerChannel& channel : s_mixer.channels)
	{
		++result;

		if (channel.playing)
			continue;

		channel.playing = true;
		channel.pos = 0;
		channel.clip = clip;
		channel.loop = loop;

		break;
	}

	NOT_EMSCRIPTEN(s_mixer.mutex.unlock());

	return result;
}

void jamStopSound(int channel)
{
	if (channel < 0)
		return;

	NOT_EMSCRIPTEN(s_mixer.mutex.lock());
	s_mixer.channels[channel] = MixerChannel();
	NOT_EMSCRIPTEN(s_mixer.mutex.unlock());
}

static void jamAudioCallback(void *userdata, Uint8 *stream, int len)
{
	NOT_EMSCRIPTEN(s_mixer.mutex.lock());

	SDL_memset(stream, 0, len);

	for (MixerChannel& channel : s_mixer.channels)
	{
		if (!channel.playing)
			continue;

		if (channel.callback)
		{
			channel.callback(channel.userData, stream, len);
			continue;
		}

		SDL_assert(channel.pos < channel.clip.length);
		int mixLength = glm::min<int>(len, channel.clip.length - channel.pos);

		SDL_MixAudio(stream, channel.clip.buffer + channel.pos, mixLength, s_mixer.volume);

		channel.pos += mixLength;

		if (channel.pos == channel.clip.length)
		{
			if (channel.loop)
			{
				channel.pos = 0;
			}
			else
			{
				channel.playing = false;
			}

			if (channel.clip.doneCallback)
			{
				channel.clip.doneCallback();
			}
		}
	}

	NOT_EMSCRIPTEN(s_mixer.mutex.unlock());
}

char *jamReadFile(const char* filename, int *outSize)
{
	char* translatedFilename = jamTranslatePath(filename);

	SDL_RWops *f = SDL_RWFromFile(translatedFilename, "r");
	SDL_assert(f && "Failed to find file, set the proper working directory");

	Sint64 size = SDL_RWsize(f) + 1;
	SDL_assert(size >= 0 && "Failed to get filesize");

	char *content = (char*)SDL_malloc(size);
	SDL_assert(content && "Failed to allocate memory");

	size_t r = SDL_RWread(f, content, size, 1);
	SDL_assert(r == 0 && "Failed to read entire file");

	content[size - 1] = 0;

	r = SDL_RWclose(f);
	SDL_assert(r == 0 && "Failed to close file");

	if (outSize)
		*outSize = (int)size;

	SDL_free(translatedFilename);

	return content;
}

void jamSetMasterVolume(float volume)
{
	SDL_assert(volume >= 0.0f);
	SDL_assert(volume <= 1.0f);
	volume *= SDL_MIX_MAXVOLUME;
	s_mixer.volume = (int)volume;
}

float jamGetMasterVolume()
{
	return (float)s_mixer.volume / SDL_MIX_MAXVOLUME;
}

void emscriptenMainLoop(void* param)
{
	std::function<void()> mainLoop = *(std::function<void()>*)param;
	mainLoop();
}

void jamGetMixerAudioSpec(SDL_AudioSpec* outDesired, SDL_AudioSpec* outObtained)
{
	if (outDesired) *outDesired = s_mixer.specDesired;
	if (outObtained) *outObtained = s_mixer.specObtained;
}

void jamXmpCallback(void *userdata, Uint8 *stream, int len)
{
	void* buffer = SDL_malloc(len);
	SDL_assert(buffer);

	SDL_AudioSpec mixerSpec;
	jamGetMixerAudioSpec(nullptr, &mixerSpec);

	u8 sourceChannels = 1; // assumes xmp player was created with XMP_FORMAT_MONO flag
	SDL_AudioCVT cvt;
	SDL_BuildAudioCVT(&cvt,
		AUDIO_S16LSB, sourceChannels, mixerSpec.freq,
		mixerSpec.format, mixerSpec.channels, mixerSpec.freq);

	cvt.len = len / cvt.len_mult;
	cvt.buf = (Uint8*)SDL_malloc(cvt.len * cvt.len_mult);

	xmp_play_buffer((xmp_context)userdata, buffer, cvt.len, 0);

	SDL_memcpy(cvt.buf, buffer, len);
	int err = SDL_ConvertAudio(&cvt);

	SDL_MixAudio(stream, (u8*)cvt.buf, len, (int)(jamGetMasterVolume() * SDL_MIX_MAXVOLUME));

	SDL_free(buffer);
	SDL_free(cvt.buf);
}

int jamMainLoop(const JamConfig& config)
{
	SDL_assert(config.gameFactory);

	jamStartupInternal(config);

	// process command line args

	for (int i = 1; i < config.argc; ++i) {
		const char* arg = config.argv[i];
		if (!strcmp(arg, "-nosound")) {
			g_enableSound = false;
		}
		if (!strcmp(arg, "-editor")) {
			g_enableEditor = true;
		}
	}

	IGame* game = config.gameFactory();

	// run main loop

	bool isRunning = true;

	float dt = 1000.0f / 60.f;
	std::function<void()> mainLoop = [&]()
	{
		g_deltaTime = dt;

		jamUpdateGuiBegin();

		SDL_Event sdlEvent;
		while (SDL_PollEvent(&sdlEvent))
		{
			if (jamGuiEvent(sdlEvent)) {
				continue;
			}

			if (game->onEvent(sdlEvent)) {
				continue;
			}

			switch (sdlEvent.type)
			{
			case SDL_QUIT:
				isRunning = false;
				break;
			case SDL_KEYDOWN:
				switch (sdlEvent.key.keysym.sym)
				{
				/*case SDLK_ESCAPE:
					isRunning = false;
					break;*/
				case SDLK_F4:
					isRunning = sdlEvent.key.keysym.mod != KMOD_LALT;
					break;
				}
			}
		}

		game->update(dt);

		jamUpdateGuiEnd(g_primitiveRenderer);

		Gfx_Present(g_device);

		Gfx_FrameEnd(g_device);

	};

#if EMSCRIPTEN
	emscripten_set_main_loop_arg(loop, &mainLoop, cfg.targetFPS, 1);
#else
	while (isRunning)
	{
		mainLoop();
	}
#endif

	delete game;

	jamShutdownInternal();

	return 0;
}

