#pragma once

#include <stddef.h>
#include <stdlib.h>

#include <SDL.h>
#include <SDL_log.h>

#include <string>
#include <algorithm>
#include <functional>
#include <cstring>

#if defined(__GNUC__)
#define __forceinline inline
#endif

#if EMSCRIPTEN
#include <html5.h>
#include <emscripten.h>
#define NOT_EMSCRIPTEN(x)
#else
#define NOT_EMSCRIPTEN(x) x
#endif

typedef unsigned char			u8;
typedef unsigned short			u16;
typedef unsigned int			u32;
typedef unsigned long long		u64;

typedef signed char				s8;
typedef signed short			s16;
typedef signed int				s32;
typedef signed long long		s64;

typedef unsigned int			uint;

class PrimitiveBatch;
class GfxDevice;
class GameWorld;

class IGame
{
public:
	virtual ~IGame() {};
	virtual void update(float dt) = 0;
	virtual bool onEvent(const SDL_Event& e) { (void)e; return false; }
};

using GameFactory = IGame* (*)(void);

struct JamConfig
{
	int argc = 0;
	char** argv = nullptr;
	const char* dataDirectory = "Data/";
	const char* name = "Window";
	u32 width = 1280;
	u32 height = 720;
	int targetFPS = 60; // mostly for emscripten rendering
	bool resizable = false;
	bool dpiAware = true;
	GameFactory gameFactory = nullptr;
};

int jamMainLoop(const JamConfig& config);

[[deprecated("Deprecated. Use jamMainLoop instead.")]] void jamStartup(const JamConfig& config);
/*[[deprecated("Deprecated. Use jamMainLoop instead.")]]*/ void jamUpdateGuiEnd(PrimitiveBatch *prim);
/*[[deprecated("Deprecated. Use jamMainLoop instead.")]]*/ bool jamGuiEvent(SDL_Event& event); // returns 'true' if event was handled by GUI
/*[[deprecated("Deprecated. Use jamMainLoop instead.")]]*/ void jamUpdateGuiBegin();
[[deprecated("Deprecated. Use jamMainLoop instead.")]] void jamShutdown();

char* jamTranslatePath(const char *filename);
std::string jamTranslatePath(const std::string& filename);
SDL_Texture* jamLoadTexture(const char *filename, Uint32 colorKey = 0);

struct SoundClip
{
	SDL_AudioSpec spec;
	u32 length = 0;
	u8* buffer = nullptr;
	std::function<void()> doneCallback;
};

SoundClip jamLoadSoundClip(const char* filename);

// returns integer ID of the audio channel/voice, which can be used to stop a sound
int jamPlaySound(SoundClip clip, bool loop = false);
int jamPlaySoundWithCallback(SDL_AudioCallback callback, void* userData);
void jamStopSound(int channel);
void jamSetMasterVolume(float volume);
float jamGetMasterVolume();
void jamGetMixerAudioSpec(SDL_AudioSpec* outDesired, SDL_AudioSpec* outObtained);
void jamXmpCallback(void *userdata, Uint8 *stream, int len);
char *jamReadFile(const char* filename, int *outSize = nullptr);

template <typename T>
T stringToEnum(const char* const inputName, T defaultVal) {
	for (int i = 0; i < int(T::NumItems); ++i) {
		const char* const name = enumToString(T(i));
		if (0 == strcmp(name, inputName)) {
			return T(i);
		}
	}

	SDL_assert(!"Couldn't parse enum");
	return defaultVal;
}

#ifdef _MSC_VER
#define stricmp _stricmp
#else
#define stricmp strcasecmp
#endif

template <typename T>
T stringToEnumInsensitive(const char* const inputName, T defaultVal) {
	for (int i = 0; i < int(T::NumItems); ++i) {
		const char* const name = enumToString(T(i));
		if (0 == stricmp(name, inputName)) {
			return T(i);
		}
	}

	SDL_assert(!"Couldn't parse enum");
	return defaultVal;
}

inline float randomUnitFloat()
{
	return (float)rand() / (float)RAND_MAX;
}

void emscriptenMainLoop(void* param);

#define countof(x) (sizeof(x)/sizeof(*x))
#define BASIC_REFLECT_MEMBER(reflector, member) reflector(member, #member)

//////////////////////////////////////////////////////////////////////////
// Jam framework globals

extern float g_deltaTime;
extern float g_elapsedTime;
extern float g_worldScale; // todo: remove this
extern SDL_Window* g_window;
extern SDL_Renderer* g_renderer;
extern GfxDevice* g_device;
extern PrimitiveBatch* g_primitiveRenderer;
extern bool g_enableSound;
extern bool g_enableEditor;
extern GameWorld* g_world;

//////////////////////////////////////////////////////////////////////////

