 #include "Graphics.h"

#include "stb_image.h"

#include <stdio.h> // for sprintf

#ifdef JAM_GLEW
#include <GL/glew.h>
#endif

#if defined (WIN32)
#include <windows.h>
#ifdef JAM_GLEW
#include <GL/wglew.h>
#endif
#endif // JAM_PLATFORM_WINDOWS

u32 bits_per_pixel(GfxFormat fmt)
{
	return bits_per_pixel(GfxFormatExtractStorage(fmt));
}

u32 bits_per_pixel(GfxFormatStorage fmt)
{
	switch (fmt)
	{
	case GfxFormatStorage_BC1: 			return 4;
	case GfxFormatStorage_BC2: 			return 8;
	case GfxFormatStorage_BC3: 			return 8;
	case GfxFormatStorage_BC4: 			return 4;
	case GfxFormatStorage_BC5: 			return 8;
	case GfxFormatStorage_BC6H: 		return 8;
	case GfxFormatStorage_BC7: 			return 8;
	case GfxFormatStorage_R16: 			return 16;
	case GfxFormatStorage_R24G8: 		return 32;
	case GfxFormatStorage_R32: 			return 32;
	case GfxFormatStorage_R32G8X24: 	return 64;
	case GfxFormatStorage_R8: 			return 8;
	case GfxFormatStorage_RG16: 		return 32;
	case GfxFormatStorage_RG32: 		return 64;
	case GfxFormatStorage_RG8: 			return 16;
	case GfxFormatStorage_RGB16: 		return 48;
	case GfxFormatStorage_RGB32: 		return 96;
	case GfxFormatStorage_RGB8: 		return 24;
	case GfxFormatStorage_RGBA16: 		return 64;
	case GfxFormatStorage_RGBA32: 		return 128;
	case GfxFormatStorage_RGBA8: 		return 32;

	default:
		SDL_assert(0);
		return 0;
	}
}

const char* to_string(GfxFormatStorage storage)
{
	switch (storage)
	{
	case GfxFormatStorage_BC1: 			return "BC1";
	case GfxFormatStorage_BC2: 			return "BC2";
	case GfxFormatStorage_BC3: 			return "BC3";
	case GfxFormatStorage_BC4: 			return "BC4";
	case GfxFormatStorage_BC5: 			return "BC5";
	case GfxFormatStorage_BC6H: 		return "BC6H";
	case GfxFormatStorage_BC7: 			return "BC7";
	case GfxFormatStorage_R16: 			return "R16";
	case GfxFormatStorage_R24G8: 		return "R24G8";
	case GfxFormatStorage_R32: 			return "R32";
	case GfxFormatStorage_R32G8X24: 	return "R32G8X24";
	case GfxFormatStorage_R8: 			return "R8";
	case GfxFormatStorage_RG16: 		return "RG16";
	case GfxFormatStorage_RG32: 		return "RG32";
	case GfxFormatStorage_RG8: 			return "RG8";
	case GfxFormatStorage_RGB16: 		return "RGB16";
	case GfxFormatStorage_RGB32: 		return "RGB32";
	case GfxFormatStorage_RGB8: 		return "RGB8";
	case GfxFormatStorage_RGBA16: 		return "RGBA16";
	case GfxFormatStorage_RGBA32: 		return "RGBA32";
	case GfxFormatStorage_RGBA8: 		return "RGBA8";

	default:
		SDL_assert(0);
		return "Unknown";
	}
}

const char* to_string(GfxFormatType type)
{
	switch (type)
	{
	case GfxFormatType_Float: 				return "Float";
	case GfxFormatType_Sint: 				return "Sint";
	case GfxFormatType_Snorm: 				return "Snorm";
	case GfxFormatType_sRGB: 				return "sRGB";
	case GfxFormatType_Typeless: 			return "Typeless";
	case GfxFormatType_Typeless_Uint: 		return "Typeless_Uint";
	case GfxFormatType_Ufloat: 				return "Ufloat";
	case GfxFormatType_Uint: 				return "Uint";
	case GfxFormatType_Unorm: 				return "Unorm";
	case GfxFormatType_Unorm_Typeless: 		return "Unorm_Typeless";
	case GfxFormatType_Unorm_Uint: 			return "Unorm_Uint";

	default:
		SDL_assert(0);
		return "Unknown";
	}
}

static u16 data_type_to_size(VertexFormatDescr::DataType type)
{
	switch (type)
	{
	case VertexFormatDescr::DataType::Float1:		return 4;
	case VertexFormatDescr::DataType::Float2:		return 8;
	case VertexFormatDescr::DataType::Float3:		return 12;
	case VertexFormatDescr::DataType::Float4:		return 16;
	case VertexFormatDescr::DataType::Half2:		return 4;
	case VertexFormatDescr::DataType::Half4:		return 8;
	case VertexFormatDescr::DataType::Short2:		return 4;
	case VertexFormatDescr::DataType::Short2N:		return 4;
	case VertexFormatDescr::DataType::UByte4:		return 4;
	case VertexFormatDescr::DataType::Dec3N:		return 4;
	case VertexFormatDescr::DataType::Color:		return 4;
	default:
		return 0;
	}
}

VertexFormatDescr::VertexFormatDescr()
	: m_has_position(false)
	, m_has_normal(false)
	, m_has_color(false)
{
	for (u32 i = 0; i < MaxStreams; ++i)
	{
		m_stream_offset[i] = 0;
	}
	for (u32 i = 0; i < MaxElements; ++i)
	{
		m_elements[i] = Element(0, DataType::Unused, UsageType::Unused, 0);
	}
}

void VertexFormatDescr::add(Element element)
{
	if (element.usage == UsageType::Position) m_has_position = true;
	if (element.usage == UsageType::Normal)   m_has_normal = true;
	if (element.usage == UsageType::Color)    m_has_color = true;

	element.offset = m_stream_offset[element.stream];

	m_stream_offset[element.stream] = m_stream_offset[element.stream] + element.size;

	m_elements.push_back(element);
}

VertexFormatDescr::Element::Element(u16 _stream, DataType _type, UsageType _usage, u8 _index)
	: stream(_stream)
	, size(data_type_to_size(_type))
	, offset(0)
	, index(_index)
	, type(_type)
	, usage(_usage)
{

}

BlendStateDescr BlendStateDescr::make_solid()
{
	BlendStateDescr res;

	res.enable = false;
	res.src = Blend::One;
	res.dst = Blend::Zero;
	res.op = BlendOp::Add;
	res.alpha_separate = false;
	res.alpha_src = Blend::One;
	res.alpha_dst = Blend::Zero;
	res.alpha_op = BlendOp::Add;

	return res;
}

BlendStateDescr BlendStateDescr::make_alphablend()
{
	BlendStateDescr res;

	res.enable = true;
	res.src = Blend::SrcAlpha;
	res.dst = Blend::InvSrcAlpha;
	res.op = BlendOp::Add;
	res.alpha_separate = false;
	res.alpha_src = Blend::SrcAlpha;
	res.alpha_dst = Blend::InvSrcAlpha;
	res.alpha_op = BlendOp::Add;

	return res;
}

BlendStateDescr BlendStateDescr::make_additive()
{
	BlendStateDescr res;

	res.enable = true;
	res.src = Blend::SrcAlpha;
	res.dst = Blend::One;
	res.op = BlendOp::Add;
	res.alpha_separate = false;
	res.alpha_src = Blend::One;
	res.alpha_dst = Blend::One;
	res.alpha_op = BlendOp::Add;

	return res;
}


BlendStateDescr BlendStateDescr::make_premultiplied_alpha()
{
	BlendStateDescr res;

	res.enable = true;
	res.src = Blend::One;
	res.dst = Blend::InvSrcAlpha;
	res.op = BlendOp::Add;
	res.alpha_separate = true;
	res.alpha_src = Blend::One;
	res.alpha_dst = Blend::One;
	res.alpha_op = BlendOp::Add;

	return res;
}

SamplerStateDescr SamplerStateDescr::make_linear()
{
	SamplerStateDescr res;

	res.filter_min = TextureFilter::Linear;
	res.filter_mag = TextureFilter::Linear;
	res.filter_mip = TextureFilter::Linear;

	return res;
}

SamplerStateDescr SamplerStateDescr::make_point()
{
	SamplerStateDescr res;

	res.filter_min = TextureFilter::Point;
	res.filter_mag = TextureFilter::Point;
	res.filter_mip = TextureFilter::Point;

	return res;
}

#ifdef _DEBUG
static void gl_check_error(const char* expr)
{
	auto err = glGetError();
	switch (err)
	{
	case GL_NO_ERROR:
		return;
	case GL_INVALID_ENUM:
		SDL_LogError(SDL_LOG_CATEGORY_RENDER, "%s -> GL Error: GL_INVALID_ENUM", expr);
		return;
	case GL_INVALID_VALUE:
		SDL_LogError(SDL_LOG_CATEGORY_RENDER, "%s -> GL Error: GL_INVALID_VALUE", expr);
		return;
	case GL_INVALID_OPERATION:
		SDL_LogError(SDL_LOG_CATEGORY_RENDER, "%s -> GL Error: GL_INVALID_OPERATION", expr);
		return;
	case GL_INVALID_FRAMEBUFFER_OPERATION:
		SDL_LogError(SDL_LOG_CATEGORY_RENDER, "%s -> GL Error: GL_INVALID_FRAMEBUFFER_OPERATION", expr);
		return;
	case GL_OUT_OF_MEMORY:
		SDL_LogError(SDL_LOG_CATEGORY_RENDER, "%s -> GL Error: GL_OUT_OF_MEMORY", expr);
		return;
	case GL_STACK_UNDERFLOW:
		SDL_LogError(SDL_LOG_CATEGORY_RENDER, "%s -> GL Error: GL_STACK_UNDERFLOW", expr);
		return;
	case GL_STACK_OVERFLOW:
		SDL_LogError(SDL_LOG_CATEGORY_RENDER, "%s -> GL Error: GL_STACK_OVERFLOW", expr);
		return;
	default:
		SDL_LogError(SDL_LOG_CATEGORY_RENDER, "%s -> GL Error: %d", expr, err);
		return;
	}
}
#define GL_CALL(x) x; gl_check_error(#x)
#else
#define GL_CALL(x) x
#endif

void GLAPIENTRY gl_debug_callback(
	GLenum source,
	GLenum type,
	GLuint id,
	GLenum severity,
	GLsizei length,
	const GLchar* message,
	const void* userParam)
{
	SDL_LogError(SDL_LOG_CATEGORY_RENDER, "OpenGL: %s", message);
}

void GLAPIENTRY gl_debug_callback_amd(
	GLuint id,
	GLenum category,
	GLenum severity,
	GLsizei length,
	const GLchar* message,
	void* userParam)
{
	SDL_LogError(SDL_LOG_CATEGORY_RENDER, "OpenGL: %s", message);
}


void GLAPIENTRY gl_debug_callback_arb(
	GLenum source,
	GLenum type,
	GLuint id,
	GLenum severity,
	GLsizei length,
	const GLchar* message,
	const void* userParam)
{
	SDL_LogError(SDL_LOG_CATEGORY_RENDER, "OpenGL: %s", message);
}

enum StateDirtyFlag
{
	StateDirtyFlag_Technique = 1 << 0,
	StateDirtyFlag_VertexStream = 1 << 1,
	StateDirtyFlag_PixelTexture = 1 << 2,
};

inline void* gl_buffer_offset(u32 offset)
{
	return ((char*)(nullptr)) + offset;
}

struct PixelFormatGL
{
	GLenum internal_format;
	GLenum format;
	GLenum type;
};

inline bool has_stencil(GfxFormat fmt)
{
	return (fmt & GfxFormatComponent_Stencil) != 0;
}

inline PixelFormatGL gl_convert(GfxFormat fmt)
{
	switch (fmt)
	{
	case GfxFormat_D32_Float:		return{ GL_DEPTH_COMPONENT32, GL_DEPTH_COMPONENT, GL_FLOAT };
	case GfxFormat_RGBA32_Float:	return{ GL_RGBA32F, GL_RGBA, GL_FLOAT };
	case GfxFormat_RGBA8_Unorm:		return{ GL_RGBA, GL_RGBA, GL_UNSIGNED_BYTE };
	default:						SDL_assert(0); return{ 0, 0 };
	}
}

inline GLenum gl_convert(CompareFunc usage)
{
	switch (usage)
	{
	case CompareFunc::Less:				return GL_LESS;
	case CompareFunc::Equal:			return GL_EQUAL;
	case CompareFunc::LessEqual:		return GL_LEQUAL;
	case CompareFunc::Greater:			return GL_GREATER;
	case CompareFunc::NotEqual:			return GL_NOTEQUAL;
	case CompareFunc::GreaterEqual:		return GL_GEQUAL;
	case CompareFunc::Always:			return GL_ALWAYS;
	case CompareFunc::Never:			return GL_NEVER;
	default:							SDL_assert(0); return 0;
	}
}

inline GLenum gl_convert(BufferMode usage)
{
	switch (usage)
	{
	case BufferMode::WriteOnce:		return GL_STATIC_DRAW;
	case BufferMode::WriteOften:	return GL_STREAM_DRAW;
	case BufferMode::WriteRarely:	return GL_DYNAMIC_DRAW;
	default:						SDL_assert(0); return GL_DYNAMIC_DRAW;
	}
}

static inline GLenum gl_convert(BlendOp op)
{
	switch (op)
	{
	case BlendOp::Add:			return GL_FUNC_ADD;
	case BlendOp::Subtract:		return GL_FUNC_SUBTRACT;
	case BlendOp::RevSubtract:	return GL_FUNC_REVERSE_SUBTRACT;
	case BlendOp::Min:			return GL_MIN_EXT;
	case BlendOp::Max:			return GL_MAX_EXT;
	default:					SDL_assert(0); return 0;
	}
}

static inline GLenum gl_convert(Blend blend)
{
	switch (blend)
	{
	case Blend::Zero:				return GL_ZERO;
	case Blend::One:				return GL_ONE;
	case Blend::SrcColor:			return GL_SRC_COLOR;
	case Blend::InvSrcColor:		return GL_ONE_MINUS_SRC_COLOR;
	case Blend::SrcAlpha:			return GL_SRC_ALPHA;
	case Blend::InvSrcAlpha:		return GL_ONE_MINUS_SRC_ALPHA;
	case Blend::DestAlpha:			return GL_DST_ALPHA;
	case Blend::InvDestAlpha:		return GL_ONE_MINUS_DST_ALPHA;
	case Blend::DestColor:			return GL_DST_COLOR;
	case Blend::InvDestColor:		return GL_ONE_MINUS_DST_COLOR;
	default:						SDL_assert(0); return GL_NEVER;
	}
}

static inline GLenum gl_convert(TextureFilter filter, TextureFilter mip)
{
	switch (filter)
	{
	case TextureFilter::Point:		return GL_NEAREST;
	case TextureFilter::Linear:		return GL_LINEAR;
	default:						SDL_assert(0); return 0;
	}
}

static inline GLint vertex_data_type_to_count(VertexFormatDescr::DataType type)
{
	switch (type)
	{
	case VertexFormatDescr::DataType::Float1:		return 1;
	case VertexFormatDescr::DataType::Float2:		return 2;
	case VertexFormatDescr::DataType::Float3:		return 3;
	case VertexFormatDescr::DataType::Float4:		return 4;
	case VertexFormatDescr::DataType::Half2:		return 2;
	case VertexFormatDescr::DataType::Half4:		return 4;
	case VertexFormatDescr::DataType::Short2:		return 2;
	case VertexFormatDescr::DataType::Short2N:		return 2;
	case VertexFormatDescr::DataType::UByte4:		return 4;
	case VertexFormatDescr::DataType::Dec3N:		return 3;
	case VertexFormatDescr::DataType::Color:		return 4;
	default:										SDL_assert(0); return 0;
	}
}

static inline GLenum vertex_data_type_to_gl(VertexFormatDescr::DataType type)
{
	switch (type)
	{
	case VertexFormatDescr::DataType::Float1:		return GL_FLOAT;
	case VertexFormatDescr::DataType::Float2:		return GL_FLOAT;
	case VertexFormatDescr::DataType::Float3:		return GL_FLOAT;
	case VertexFormatDescr::DataType::Float4:		return GL_FLOAT;
	case VertexFormatDescr::DataType::Half2:		return 0; // not supported
	case VertexFormatDescr::DataType::Half4:		return 0; // not supported
	case VertexFormatDescr::DataType::Short2:		return GL_SHORT;
	case VertexFormatDescr::DataType::Short2N:		return GL_SHORT;
	case VertexFormatDescr::DataType::UByte4:		return GL_UNSIGNED_BYTE;
	case VertexFormatDescr::DataType::Dec3N:		return 0; // not supported
	case VertexFormatDescr::DataType::Color:		return GL_UNSIGNED_BYTE;
	default:										SDL_assert(0); return 0;
	}
}

static inline GLboolean vertex_data_type_to_normalized(VertexFormatDescr::DataType type)
{
	switch (type)
	{
	case VertexFormatDescr::DataType::Float1:		return GL_FALSE;
	case VertexFormatDescr::DataType::Float2:		return GL_FALSE;
	case VertexFormatDescr::DataType::Float3:		return GL_FALSE;
	case VertexFormatDescr::DataType::Float4:		return GL_FALSE;
	case VertexFormatDescr::DataType::Half2:		return GL_FALSE;
	case VertexFormatDescr::DataType::Half4:		return GL_FALSE;
	case VertexFormatDescr::DataType::Short2:		return GL_FALSE;
	case VertexFormatDescr::DataType::Short2N:		return GL_TRUE;
	case VertexFormatDescr::DataType::UByte4:		return GL_FALSE;
	case VertexFormatDescr::DataType::Dec3N:		return GL_TRUE;
	case VertexFormatDescr::DataType::Color:		return GL_TRUE;
	default:										SDL_assert(0); return GL_FALSE;
	}
}

inline const char* usage_type_to_string(VertexFormatDescr::UsageType type)
{
	switch (type)
	{
	case VertexFormatDescr::UsageType::Position:	return "a_pos";
	case VertexFormatDescr::UsageType::Normal:		return "a_nor";
	case VertexFormatDescr::UsageType::Binormal:	return "a_bin";
	case VertexFormatDescr::UsageType::Tangent:		return "a_tan";
	case VertexFormatDescr::UsageType::Color:		return "a_col";
	case VertexFormatDescr::UsageType::Texcoord:	return "a_tex";
	default:										return "UNUSED";
	}
}

struct VertexFormatGL
{
	VertexFormatDescr descr;
};

struct VertexShaderGL
{
	GLuint native;
};

struct PixelShaderGL
{
	GLuint native;
};

struct TextureGL
{
	TextureDescr descr;
	GLuint native;
};

struct BufferGL
{
	BufferDescr descr;
	GLuint native;
};

struct SamplerStateGL
{
	SamplerStateDescr descr;
	GLuint native;
};

struct BlendStateGL
{
	BlendStateDescr descr;
};

struct DepthStencilStateGL
{
	DepthStencilStateDescr descr;
};

struct RasterizerStateGL
{
	RasterizerStateDescr descr;
};


struct TechniqueGL
{
	struct UniformBinding : public ShaderConstantIndex
	{
		ShaderConstantBindings::BindingType type;
		const void* data;
	};

	struct SamplerBinding
	{
		u32 index;
		GLint location;
	};

	struct AttributeBinding
	{
		GLint location;
		GLint count;
		GLenum type;
		GLsizei stride;
		u16 offset;
		u16 stream;
		GLboolean normalized;
	};

	GLuint native;
	GLuint vertex_array_object;

	std::vector<UniformBinding>    uniforms;
	std::vector<AttributeBinding>  attributes;
	std::vector<SamplerBinding>    samplers;
};

struct RenderResourcesGL
{
	DynamicResourcePool<VertexFormatGL, VertexFormatHandle>			vertex_formats;
	DynamicResourcePool<VertexShaderGL, VertexShaderHandle>			vertex_shaders;
	DynamicResourcePool<PixelShaderGL, PixelShaderHandle>			pixel_shaders;
	DynamicResourcePool<TextureGL, TextureHandle>				textures;
	DynamicResourcePool<BufferGL, BufferHandle>				buffers;
	DynamicResourcePool<SamplerStateGL, SamplerStateHandle>			sampler_states;
	DynamicResourcePool<BlendStateGL, BlendStateHandle>			blend_states;
	DynamicResourcePool<TechniqueGL, TechniqueHandle>			techniques;
	DynamicResourcePool<DepthStencilStateGL, DepthStencilStateHandle>	depth_stencil_states;
	DynamicResourcePool<RasterizerStateGL, RasterizerStateHandle>		rasterizer_states;

	//GLuint vao; // dynamic vertex array object
	GLuint fbo; // dynamic frame buffer object
};

class GfxContext
{
public:
	enum
	{
		MaxTextures = 16,
		MaxVertexStreams = 8
	};

	GfxContext()
		: primitive_type(GL_TRIANGLES)
		, index_type(0)
		, index_size(0)
		, index_buffer(0)
		, resources(nullptr)
		, stats(nullptr)
		, dirty_state_flags(0)
		, dirty_vertex_streams(0)
		, line_width(1.0f)
	{
		enabled_vertex_attributes.reserve(128);
	}

	GLint 	primitive_type;
	GLint 	index_type;
	int 	index_size;
	GLuint 	index_buffer;
	float	line_width;

	RenderResourcesGL* resources;
	GfxDeviceStats* stats;

	u32					dirty_state_flags;
	TechniqueHandle		pending_technique;
	TextureHandle		pending_pixel_textures[MaxTextures];
	SamplerStateHandle	pending_pixel_samplers[MaxTextures];
	BufferHandle		pending_vertex_streams[MaxVertexStreams];
	u32					dirty_vertex_streams;

	std::vector<u32>	enabled_vertex_attributes;
};

class GfxDevice
{
public:

	ReferenceCount refct;
	RenderResourcesGL resources;
	GfxContext context;
	GfxDeviceConfig config;
	SDL_Window* window;
	GfxDeviceStats stats;
	SDL_GLContext gl_context;
	u32 gl_version_number;	// (MAJOR << 8) | MINOR
};

static GLuint gl_compile_shader(GfxDevice* dev, const char* source, ShaderType type)
{
	// copypaste from 
	GLuint shader;
	GLint compiled;

	GLenum gl_shader_type;// = (type==ShaderType::Pixel) ? GL_FRAGMENT_SHADER : GL_VERTEX_SHADER ;
	switch (type) {
	case ShaderType::Pixel:
		gl_shader_type = GL_FRAGMENT_SHADER;
		break;
	case ShaderType::Vertex:
		gl_shader_type = GL_VERTEX_SHADER;
		break;
	default:
		SDL_LogError(SDL_LOG_CATEGORY_RENDER, "Unsupported shader type (%d)", (int)type);
		return 0;
	}

	shader = glCreateShader(gl_shader_type);

	if (shader == 0)
		return 0;

	std::string patched_source;

#if JAM_RENDER_API==JAM_RENDER_API_GLES2
	if (!strstr(source, "#version"))
	{
		if (dev->gl_version_number > 0x200)
		{
			//patched_source = "#version 100\n";
		}

		#if EMSCRIPTEN
		patched_source = "#version 100\n";
		#endif 

		if (gl_shader_type == GL_FRAGMENT_SHADER)
		{
			if (dev->gl_version_number > 0x200)
			{
				//patched_source += "precision mediump float;\n";
			}

			#if EMSCRIPTEN
			patched_source += "precision mediump float;\n";
			#endif 

			patched_source +=
				"#define JAM_TEX2D(s,t) texture2D(s,t)\n"
				"#define JAM_IN varying\n"
				"#define JAM_FRAGCOLOR gl_FragColor\n";
		}
		else if (gl_shader_type == GL_VERTEX_SHADER)
		{
			patched_source +=
				"#define JAM_IN attribute\n"
				"#define JAM_OUT varying\n";
		}
	}
#else
	if (!strstr(source, "#version"))
	{
		patched_source = "#version 150\n"
			"#define JAM_IN in\n"
			"#define JAM_OUT out\n"
			"#define JAM_TEX2D(s,t) texture(s,t)\n";
		if (gl_shader_type == GL_FRAGMENT_SHADER)
		{
			patched_source +=
				"out vec4 JAM_FRAGCOLOR;\n";
		}
	}
#endif

	patched_source += source;

	const char* patched_source_cstr = patched_source.c_str();

	// Load the shader source
	GL_CALL(glShaderSource(shader, 1, &patched_source_cstr, nullptr));
	GL_CALL(glCompileShader(shader));

	// Check the compile status
	GL_CALL(glGetShaderiv(shader, GL_COMPILE_STATUS, &compiled));

	if (!compiled)
	{
		GLint infoLen = 0;

		GL_CALL(glGetShaderiv(shader, GL_INFO_LOG_LENGTH, &infoLen));

		if (infoLen > 1)
		{
			char* infoLog = (char*)malloc(sizeof(char) * infoLen);

			GL_CALL(glGetShaderInfoLog(shader, infoLen, nullptr, infoLog));
			SDL_LogError(SDL_LOG_CATEGORY_RENDER, "Error compiling shader:\n%s\n%s\n", patched_source_cstr, infoLog);

			free(infoLog);
		}

		GL_CALL(glDeleteShader(shader));

		return 0;
	}

	return shader;
}

void Gfx_AddReference(GfxDevice* dev)
{
	dev->refct.add_reference();
}

void Gfx_RemoveReference(GfxDevice* dev)
{
	dev->refct.remove_reference();
}

void Gfx_Clear(GfxContext* /*rc*/, ColorRGBA color, ClearFlags clear_flags, float depth, u32 stencil)
{
	#if EMSCRIPTEN //TODO: @Yuriy
	GLenum status = glCheckFramebufferStatus(GL_FRAMEBUFFER);
	#else
	GLenum status = glCheckFramebufferStatusEXT(GL_FRAMEBUFFER);
	#endif

	if (status != GL_FRAMEBUFFER_COMPLETE)
		SDL_LogError(SDL_LOG_CATEGORY_RENDER, "Incomplete frame buffer");

	GLbitfield clear_bits = 0;
	if ((u32)clear_flags & (u32)ClearFlags::Color)
	{
		clear_bits |= GL_COLOR_BUFFER_BIT;
		GL_CALL(glClearColor(color.r, color.g, color.b, color.a));
	}
	if ((u32)clear_flags & (u32)ClearFlags::Depth)
	{
		clear_bits |= GL_DEPTH_BUFFER_BIT;
		GL_CALL(glClearDepth(depth));
	}
	if ((u32)clear_flags & (u32)ClearFlags::Stencil)
	{
		clear_bits |= GL_STENCIL_BUFFER_BIT;
		GL_CALL(glClearStencil(stencil));
	}
	if (clear_bits)
	{
		GL_CALL(glClear(clear_bits));
	}
}

const TextureDescr& Gfx_Texture_Descr(GfxDevice* dev, TextureHandle h)
{
	if (h.valid())
	{
		return dev->resources.textures[h].descr;
	}
	else
	{
		static const TextureDescr invalid_texture = TextureDescr::create_2D(1, 1, GfxFormat_Unknown, false, false);
		return invalid_texture;
	}
}

inline void Gfx_Commit_GL(GfxContext* rc)
{
	const u32 dirty = rc->dirty_state_flags;

	if (dirty & (StateDirtyFlag_Technique | StateDirtyFlag_VertexStream | StateDirtyFlag_PixelTexture))
	{
		const TechniqueGL& t = rc->resources->techniques[rc->pending_technique];

		if (dirty&StateDirtyFlag_Technique)
		{
			GL_CALL(glUseProgram(t.native));
		}

		if ((dirty&StateDirtyFlag_Technique) || (dirty&StateDirtyFlag_VertexStream))
		{
			//GL_CALL(glBindVertexArray(rc->resources->vao));

			for (u32 i = 0; i < rc->enabled_vertex_attributes.size(); ++i)
			{
				GL_CALL(glDisableVertexAttribArray(rc->enabled_vertex_attributes[i]));
			}
			rc->enabled_vertex_attributes.clear();

			for (u32 i = 0; i < t.attributes.size(); ++i)
			{
				const TechniqueGL::AttributeBinding& attr = t.attributes[i];
				const BufferHandle vbh = rc->pending_vertex_streams[attr.stream];
				if (vbh.valid())
				{
					const BufferGL& vb = rc->resources->buffers[vbh];
					GL_CALL(glBindBuffer(GL_ARRAY_BUFFER, vb.native));
					GL_CALL(glVertexAttribPointer(attr.location, attr.count, attr.type, attr.normalized, attr.stride, gl_buffer_offset(attr.offset)));
					GL_CALL(glEnableVertexAttribArray(attr.location));
					rc->enabled_vertex_attributes.push_back(attr.location);
				}
			}
		}

		// TODO: Need to unset previously set but now unused textures
		if (dirty&StateDirtyFlag_PixelTexture)
		{
			for (u32 i = 0; i < t.samplers.size(); ++i)
			{
				const TechniqueGL::SamplerBinding& b = t.samplers[i];

				TextureHandle texture_handle = rc->pending_pixel_textures[b.index];
				SamplerStateHandle sampler_handle = rc->pending_pixel_samplers[b.index];

				const TextureGL& texture = rc->resources->textures[texture_handle];
				const SamplerStateGL& sampler = rc->resources->sampler_states[sampler_handle];

				GL_CALL(glActiveTexture(GL_TEXTURE0 + b.index));
				GL_CALL(glBindTexture(GL_TEXTURE_2D, texture.native));
#if JAM_RENDER_API!=JAM_RENDER_API_GLES2
				GL_CALL(glBindSampler(b.index, sampler.native));
#endif
				GL_CALL(glUniform1i(b.location, b.index));
			}
		}
	}

	// TODO: This is temporary here, until constant buffers are added
	if (rc->pending_technique.valid())
	{
		const TechniqueGL& t = rc->resources->techniques[rc->pending_technique];
		for (u32 i = 0; i < t.uniforms.size(); ++i)
		{
			const TechniqueGL::UniformBinding& b = t.uniforms[i];
			if (b.type == ShaderConstantBindings::BindingType_Matrix)
			{
				GL_CALL(glUniformMatrix4fv(b.index, 1, false, (const GLfloat*)b.data));
			}
			else if (b.type == ShaderConstantBindings::BindingType_Scalar)
			{
				GL_CALL(glUniform1fv(b.index, b.count, (const GLfloat*)b.data));
			}
			else if (b.type == ShaderConstantBindings::BindingType_Vector)
			{
				GL_CALL(glUniform4fv(b.index, b.count, (const GLfloat*)b.data));
			}
		}
	}

	if (rc->primitive_type == GL_LINES || rc->primitive_type == GL_LINE_LOOP || rc->primitive_type == GL_LINE_STRIP)
	{
		GL_CALL(glLineWidth(rc->line_width)); // this shit makes me sad, but it works.
	}

	rc->dirty_state_flags = 0;
}

void Gfx_Draw(GfxContext* rc, u32 first_vertex, u32 vertex_count)
{
	Gfx_Commit_GL(rc);
	GL_CALL(glDrawArrays(rc->primitive_type, first_vertex, vertex_count));

	rc->stats->draw_calls++;
	rc->stats->vertices += vertex_count;
}

void Gfx_DrawIndexed(GfxContext* rc, u32 index_count, u32 first_index, u32 base_vertex, u32 num_vertices)
{
	SDL_assert(rc->index_buffer && "Index buffer must be bound before Gfx_DrawIndexed is called.");

	(void)num_vertices;
	(void)base_vertex;

	Gfx_Commit_GL(rc);
	GL_CALL(glDrawElements(rc->primitive_type, index_count, rc->index_type, gl_buffer_offset(first_index*rc->index_size)));

	rc->stats->draw_calls++;
	rc->stats->vertices += index_count;
}

GfxContext* Gfx_FrameBegin(GfxDevice* dev)
{
	return Gfx_Immediate_Context(dev);
}

void Gfx_FrameEnd(GfxDevice*)
{
	// nothing
}

void Gfx_Present(GfxDevice* dev)
{
	SDL_GL_SwapWindow(dev->window);
}

const GfxDeviceStats& Gfx_Stats(GfxDevice* dev)
{
	return dev->stats;
}

void Gfx_ResetStats(GfxDevice* dev)
{
	dev->stats.draw_calls = 0;
	dev->stats.vertices = 0;
}

void Gfx_SetPixelSamplerState(GfxContext* rc, u32 idx, SamplerStateHandle h)
{
	if (idx < GfxContext::MaxTextures)
	{
		rc->pending_pixel_samplers[idx] = h;
		rc->dirty_state_flags |= StateDirtyFlag_PixelTexture; // TODO: dedicated dirty flag
	}
	else
	{
		SDL_assert(false && "Sampler index out of bounds");
	}
}

void Gfx_SetDepthStencilState(GfxContext* rc, DepthStencilStateHandle next_state)
{
	const auto& descr = rc->resources->depth_stencil_states[next_state].descr;
	if (descr.enable)
	{
		GL_CALL(glEnable(GL_DEPTH_TEST));
		GL_CALL(glDepthMask(descr.write_enable ? GL_TRUE : GL_FALSE));
		GL_CALL(glDepthFunc(gl_convert(descr.cmp_func)));
	}
	else
	{
		GL_CALL(glDisable(GL_DEPTH_TEST));
	}
}

void Gfx_SetRasterizerState(GfxContext* rc, RasterizerStateHandle next_state)
{
	const auto& descr = rc->resources->rasterizer_states[next_state].descr;

#if !EMSCRIPTEN //TODO: @Yuriy
	if (descr.fill_mode == FillMode::Solid)
	{
		GL_CALL(glPolygonMode(GL_FRONT_AND_BACK, GL_FILL));
	}
	else
	{
		GL_CALL(glPolygonMode(GL_FRONT_AND_BACK, GL_LINE));
	}
#endif

	if (descr.cull_mode == CullMode::CW)
	{
		GL_CALL(glCullFace(GL_FRONT));
	}
	else
	{
		GL_CALL(glCullFace(GL_BACK));
	}

	if (descr.depth_bias)
	{
		GL_CALL(glEnable(GL_POLYGON_OFFSET_FILL));
		GL_CALL(glPolygonOffset(descr.depth_bias_slope_scale, descr.depth_bias));
	}
	else
	{
		GL_CALL(glDisable(GL_POLYGON_OFFSET_FILL));
	}


	if (descr.scissor_enable)
	{
		GL_CALL(glEnable(GL_SCISSOR_TEST));
	}
	else
	{
		GL_CALL(glDisable(GL_SCISSOR_TEST));
	}
}

void Gfx_DepthStencilState_Destroy(GfxDevice* dev, DepthStencilStateHandle h)
{
	if (h.valid())
	{
		dev->resources.depth_stencil_states.remove(h);
	}
}

void Gfx_SetVertexStream(GfxContext* rc, u32 idx, BufferHandle h)
{
	if (idx < GfxContext::MaxVertexStreams)
	{
		rc->pending_vertex_streams[idx] = h;
		rc->dirty_vertex_streams |= 1 << idx;
		rc->dirty_state_flags |= StateDirtyFlag_VertexStream;
	}
	else
	{
		SDL_assert(false && "Vertex stream index out of bounds");
	}
}

void Gfx_SetPixelTexture(GfxContext* rc, u32 idx, TextureHandle h)
{
	if (idx < GfxContext::MaxTextures)
	{
		rc->pending_pixel_textures[idx] = h;
		rc->dirty_state_flags |= StateDirtyFlag_PixelTexture;
	}
	else
	{
		SDL_assert(false && "Texture index out of bounds");
	}
}

void Gfx_SetIndexStream(GfxContext* rc, BufferHandle h)
{
	const BufferGL ib = rc->resources->buffers[h];
	GL_CALL(glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, ib.native));
	rc->index_buffer = ib.native;
	rc->index_type = (ib.descr.format == GfxFormat_R16_Uint) ? GL_UNSIGNED_SHORT : GL_UNSIGNED_INT;
	rc->index_size = (ib.descr.format == GfxFormat_R16_Uint) ? 2 : 4;
}

void Gfx_SetBlendState(GfxContext* rc, BlendStateHandle h)
{
	const BlendStateGL& state = rc->resources->blend_states[h];
	const BlendStateDescr& descr = state.descr;

	//TODO: only color blend states are implemented, still need to support separate alpha states

	if (descr.enable)
	{
		GL_CALL(glEnable(GL_BLEND));
		GL_CALL(glBlendFuncSeparate(gl_convert(descr.src), gl_convert(descr.dst), gl_convert(descr.alpha_src), gl_convert(descr.alpha_dst)));
		GL_CALL(glBlendEquationSeparate(gl_convert(descr.op), gl_convert(descr.alpha_op)));
	}
	else
	{
		GL_CALL(glDisable(GL_BLEND));
	}
}

void Gfx_VertexShader_Destroy(GfxDevice* dev, VertexShaderHandle& h)
{
	(void)h;
	(void)dev;
	// TODO
};

void Gfx_VertexFormat_Destroy(GfxDevice* dev, VertexFormatHandle h)
{
	(void)h;
	(void)dev;
	// TODO
}

void Gfx_Buffer_Destroy(GfxDevice* dev, BufferHandle h)
{
	(void)h;
	(void)dev;
	// TODO
}

void Gfx_SamplerState_Destroy(GfxDevice* dev, SamplerStateHandle h)
{
	(void)h;
	(void)dev;
	// TODO
}

void Gfx_SetTechnique(GfxContext* rc, TechniqueHandle h)
{
	rc->pending_technique = h;
	rc->dirty_state_flags |= StateDirtyFlag_Technique;
}

void Gfx_SetLineWidth(GfxContext* rc, float width)
{
	rc->line_width = width;
}

void Gfx_SetPrimitive(GfxContext* rc, PrimitiveType type)
{
	switch (type)
	{
	case PrimitiveType::PointList:
		rc->primitive_type = GL_POINTS;
		break;
	case PrimitiveType::LineList:
		rc->primitive_type = GL_LINES;
		break;
	case PrimitiveType::LineStrip:
		rc->primitive_type = GL_LINE_STRIP;
		break;
	case PrimitiveType::LineLoop:
		rc->primitive_type = GL_LINE_LOOP;
		break;
	case PrimitiveType::TriangleList:
		rc->primitive_type = GL_TRIANGLES;
		break;
	case PrimitiveType::TriangleStrip:
		rc->primitive_type = GL_TRIANGLE_STRIP;
		break;
	default:
		rc->primitive_type = 0;
		break;
	}
}

void Gfx_Buffer_Update(GfxDevice* dev, BufferHandle h, void* data, u32 offset, u32 size, bool discard)
{
	BufferGL& buf = dev->resources.buffers[h];
	SDL_assert(buf.native);

	GLint gl_type = buf.descr.type == BufferType::Vertex ? GL_ARRAY_BUFFER : GL_ELEMENT_ARRAY_BUFFER;

	GL_CALL(glBindBuffer(gl_type, buf.native));

	u32 buf_size = buf.descr.stride * buf.descr.count;

	if (offset == 0 && ((size == 0 || size == buf_size) || discard))
	{
		if (size == 0) size = buf_size;
		GL_CALL(glBufferData(gl_type, size, data, gl_convert(buf.descr.mode)));
	}
	else
	{
		GL_CALL(glBufferSubData(gl_type, offset, size, data));
	}
}

GfxContext* Gfx_Immediate_Context(GfxDevice* dev)
{
	return &dev->context;
}

void Gfx_PixelShader_Destroy(GfxDevice* dev, PixelShaderHandle& h)
{
	(void)h;
	(void)dev;
	// TODO
}

void Gfx_BlendState_Destroy(GfxDevice* dev, BlendStateHandle h)
{
	(void)h;
	(void)dev;
	// TODO
}

void Gfx_Technique_Destroy(GfxDevice* /*dev*/, TechniqueHandle /*h*/)
{
	// TODO
}

void Gfx_Texture_Destroy(GfxDevice* dev, TextureHandle& h)
{
	if (h.valid())
	{
		auto& t = dev->resources.textures[h];
		glDeleteTextures(1, &t.native);
		dev->resources.textures.remove(h);
	}
}

void Gfx_DestroyDevice(GfxDevice* dev)
{
#if !EMSCRIPTEN //TODO: @Yuriy
	//GL_CALL(glDeleteVertexArrays(1, &dev->resources.vao));
	GL_CALL(glDeleteFramebuffersEXT(1, &dev->resources.fbo));

	SDL_GL_DeleteContext(dev->gl_context);	
	delete dev;
#endif	
}

GfxDevice* Gfx_CreateDevice(SDL_Window* window, const GfxDeviceConfig& cfg)
{
	GfxDevice* dev = new GfxDevice;

	dev->config = cfg;
	dev->window = window;

	dev->gl_context = SDL_GL_CreateContext(window);
	SDL_assert(dev->gl_context != nullptr);

#ifdef JAM_GLEW
	glewExperimental = GL_TRUE;
	GLenum glew_err = glewInit();
	if (glew_err != GLEW_OK)
	{
		auto errorStr = glewGetErrorString(glew_err);
		SDL_LogError(SDL_LOG_CATEGORY_RENDER, "Failed to initialize GLEW: %s", errorStr);
	}
#endif//JAM_GLEW

	{
		u32 version_major, version_minor;
		const char* const version_string = (const char*)glGetString(GL_VERSION);
		sscanf(version_string, "%u.%u", &version_major, &version_minor);
		dev->gl_version_number = (version_major << 8u) | version_minor;
	}

	dev->context.resources = &dev->resources;
	dev->context.stats = &dev->stats;

#if GL_VERSION_4_3
	if (dev->config.debug && glewIsSupported("GL_KHR_debug"))
	{
		glDebugMessageCallback(gl_debug_callback, dev);
		glDebugMessageControl(GL_DONT_CARE, GL_DONT_CARE, GL_DONT_CARE, 0, nullptr, GL_TRUE);
		glEnable(GL_DEBUG_OUTPUT);
		glEnable(GL_DEBUG_OUTPUT_SYNCHRONOUS);
	}
	else
		if (dev->config.debug && glewIsSupported("GL_AMD_debug_output"))
		{
			glDebugMessageCallbackARB(gl_debug_callback_arb, dev);
			glDebugMessageControlARB(GL_DONT_CARE, GL_DONT_CARE, GL_DONT_CARE, 0, nullptr, GL_TRUE);
		}
		else
			if (dev->config.debug && glewIsSupported("GL_ARB_debug_output"))
			{
				glDebugMessageCallbackAMD(gl_debug_callback_amd, dev);
				glDebugMessageEnableAMD(GL_DONT_CARE, GL_DONT_CARE, 0, nullptr, GL_TRUE);
			}
#endif //GL_VERSION_4_3

	//GL_CALL(glGenVertexArrays(1, &dev->resources.vao));

#if EMSCRIPTEN //TODO: @Yuriy
	GL_CALL(glGenFramebuffers(1, &dev->resources.fbo));
#else
	GL_CALL(glGenFramebuffersEXT(1, &dev->resources.fbo));
#endif
	

	return dev;
}

RasterizerStateHandle Gfx_RasterizerState_Create(GfxDevice* dev, const RasterizerStateDescr& descr)
{
	RasterizerStateGL res;
	res.descr = descr;
	return dev->resources.rasterizer_states.push(res);
}

void Gfx_RasterizerState_Destroy(GfxDevice* dev, RasterizerStateHandle h)
{
	if (h.valid())
	{
		dev->resources.rasterizer_states.remove(h);
	}
}

DepthStencilStateHandle Gfx_DepthStencilState_Create(GfxDevice* dev, const DepthStencilStateDescr& descr)
{
	DepthStencilStateGL res;
	res.descr = descr;
	return dev->resources.depth_stencil_states.push(res);
}

VertexShaderHandle Gfx_VertexShader_Create(GfxDevice* dev, const ShaderCode& code)
{
	VertexShaderGL res;
	res.native = gl_compile_shader(dev, code.data(), ShaderType::Vertex);
	if (res.native)
	{
		return dev->resources.vertex_shaders.push(res);
	}
	else
	{
		return InvalidResourceHandle();
	}
}

PixelShaderHandle Gfx_PixelShader_Create(GfxDevice* dev, const ShaderCode& code)
{
	PixelShaderGL res;
	res.native = gl_compile_shader(dev, code.data(), ShaderType::Pixel);
	if (res.native)
	{
		return dev->resources.pixel_shaders.push(res);
	}
	else
	{
		return InvalidResourceHandle();
	}
}

BufferHandle Gfx_Buffer_Create(GfxDevice* dev, const BufferDescr& descr, const void* data)
{
	SDL_assert((descr.mode == BufferMode::WriteOnce && data == nullptr) == false && "Static buffers must provide valid initial data.");

	BufferGL res;

	res.descr = descr;

	GLint gl_type = descr.type == BufferType::Vertex ? GL_ARRAY_BUFFER : GL_ELEMENT_ARRAY_BUFFER;

	GL_CALL(glGenBuffers(1, &res.native));
	GL_CALL(glBindBuffer(gl_type, res.native));
	GL_CALL(glBufferData(gl_type, descr.count*descr.stride, data, gl_convert(descr.mode)));

	return dev->resources.buffers.push(res);
}

VertexFormatHandle Gfx_VertexFormat_Create(GfxDevice* dev, const VertexFormatDescr& fmt)
{
	VertexFormatGL res;
	res.descr = fmt;
	return dev->resources.vertex_formats.push(res);
}

BlendStateHandle Gfx_BlendState_Create(GfxDevice* dev, const BlendStateDescr& descr)
{
	BlendStateGL res;
	res.descr = descr;
	return dev->resources.blend_states.push(res);
}

SamplerStateHandle Gfx_SamplerState_Create(GfxDevice* dev, const SamplerStateDescr& descr)
{
#if JAM_RENDER_API==JAM_RENDER_API_GLES2 // TODO: set sampler states in webgl / emscripten
	return InvalidResourceHandle();
#else
	SamplerStateGL res;
	res.descr = descr;
	GL_CALL(glGenSamplers(1, &res.native));

	GL_CALL(glSamplerParameteri(res.native, GL_TEXTURE_MAG_FILTER, gl_convert(descr.filter_mag, descr.filter_mip)));
	GL_CALL(glSamplerParameteri(res.native, GL_TEXTURE_MIN_FILTER, gl_convert(descr.filter_min, descr.filter_mip)));

	return dev->resources.sampler_states.push(res);
#endif
}

TechniqueHandle Gfx_Technique_Create(GfxDevice* dev, const TechniqueDescr& descr)
{
	GLint ps = dev->resources.pixel_shaders[descr.ps].native;
	GLint vs = dev->resources.vertex_shaders[descr.vs].native;

	const VertexFormatDescr& vf = dev->resources.vertex_formats[descr.vf].descr;

	GLuint po = glCreateProgram();

	if (po == 0)
		return InvalidResourceHandle();

	GL_CALL(glAttachShader(po, ps));
	GL_CALL(glAttachShader(po, vs));

	// Link the program
	GL_CALL(glLinkProgram(po));

	// Check the link status
	GLint linked = 0;
	GL_CALL(glGetProgramiv(po, GL_LINK_STATUS, &linked));

	if (!linked)
	{
		GLint info_len = 0;

		GL_CALL(glGetProgramiv(po, GL_INFO_LOG_LENGTH, &info_len));

		if (info_len > 1)
		{
			char* infoLog = (char*)malloc(sizeof(char) * info_len);

			GL_CALL(glGetProgramInfoLog(po, info_len, nullptr, infoLog));
			SDL_LogError(SDL_LOG_CATEGORY_RENDER, "Error linking program:\n%s\n", infoLog);

			free(infoLog);
		}

		GL_CALL(glDeleteProgram(po));
		return InvalidResourceHandle();
	}

	TechniqueGL res;

	res.native = po;
	res.vertex_array_object = 0;

	const ShaderConstantBindings* bindings = descr.bindings;
	if (bindings)
	{
		for (u32 i = 0; i < bindings->items.size(); ++i)
		{
			const ShaderConstantBindings::Item& item = bindings->items[i];
			GLint location = GL_CALL(glGetUniformLocation(po, item.name));
			if (location != -1)
			{
				if (item.type == ShaderConstantBindings::BindingType_Vector
					|| item.type == ShaderConstantBindings::BindingType_Matrix
					|| item.type == ShaderConstantBindings::BindingType_Scalar)
				{
					TechniqueGL::UniformBinding b;
					b.index = location;
					b.data = item.data;
					b.count = item.count;
					b.type = item.type;
					res.uniforms.push_back(b);
				}
				else if (item.type == ShaderConstantBindings::BindingType_Sampler)
				{
					TechniqueGL::SamplerBinding b;
					b.index = item.idx;
					b.location = location;
					res.samplers.push_back(b);
				}
			}
		}
	}

	for (u32 i = 0; i < vf.num_elements(); ++i)
	{
		const VertexFormatDescr::Element& elem = vf.element(i);
		char elem_name[256];
		sprintf(elem_name, "%s%d", usage_type_to_string(elem.usage), elem.index);
		GLint location = GL_CALL(glGetAttribLocation(po, elem_name));
		if (location != -1)
		{
			TechniqueGL::AttributeBinding b;
			b.stream = elem.stream;
			b.location = location;
			b.type = vertex_data_type_to_gl(elem.type);
			b.count = vertex_data_type_to_count(elem.type);
			b.normalized = vertex_data_type_to_normalized(elem.type);
			b.stride = vf.stream_stride(elem.stream);
			b.offset = elem.offset;
			res.attributes.push_back(b);
		}
	}

	return dev->resources.techniques.push(res);
}

TextureHandle Gfx_Texture_Load(GfxDevice* dev, const char* filename, TextureType type)
{
	int width, height, components;
	u8 * const pixels = stbi_load(filename, &width, &height, &components, 4);
	SDL_assert(pixels && "Can't load texture");

	std::vector<ColorRGBA8>	flipped(width*height);

	for (int y = 0; y < height; ++y) 
	{
		memcpy(flipped.data() + y * width, pixels + ((height - y - 1) * width) * 4, width * 4);
	}

	TextureDescr descr = TextureDescr::create_2D(width, height, GfxFormat_RGBA8_Unorm);
	TextureHandle tex = Gfx_Texture_Create(g_device, descr, flipped.data(), width * height * 4);

	stbi_image_free(pixels);

	return tex;
}

TextureHandle Gfx_Texture_Create(GfxDevice* dev, const TextureDescr& tex, const void* data, size_t size)
{
	(void)size;

	PixelFormatGL native_fmt = gl_convert(tex.format);

	TextureGL res;

	res.descr = tex;

	GL_CALL(glGenTextures(1, &res.native));
	GL_CALL(glBindTexture(GL_TEXTURE_2D, res.native));
    GL_CALL(glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, tex.repeat ? GL_REPEAT : GL_CLAMP_TO_EDGE));
	GL_CALL(glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, tex.repeat ? GL_REPEAT : GL_CLAMP_TO_EDGE));
	GL_CALL(glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR));
	GL_CALL(glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR));
	GL_CALL(glTexImage2D(GL_TEXTURE_2D, 0, native_fmt.internal_format, tex.width, tex.height, 0, native_fmt.format, native_fmt.type, data));

	return dev->resources.textures.push(res);
}

void Gfx_Texture_Update(GfxDevice* dev, TextureHandle texHandle, const void* data)
{
	TextureGL &res = dev->resources.textures[texHandle];
	TextureDescr tex = res.descr;

	PixelFormatGL native_fmt = gl_convert(tex.format);
	GL_CALL(glBindTexture(GL_TEXTURE_2D, res.native));
	GL_CALL(glTexImage2D(GL_TEXTURE_2D, 0, native_fmt.internal_format, tex.width, tex.height, 0, native_fmt.format, native_fmt.type, data));
}

void Gfx_SetScissorRect(GfxContext* rc, const GfxRect& rect)
{
	// TODO: handle Y flip internally, emulate DX convention
	glScissor(rect.left, rect.top, rect.right - rect.left, rect.bottom - rect.top);
}

void Gfx_SetPresentInterval(GfxDevice* dev, u32 interval)
{
	dev->config.present_interval = interval;
	SDL_GL_SetSwapInterval(interval);
}

#if EMSCRIPTEN
#define glBindFramebufferEXT glBindFramebuffer
#define glFramebufferTexture2DEXT glFramebufferTexture2D
#undef GL_DRAW_FRAMEBUFFER
#define GL_DRAW_FRAMEBUFFER GL_FRAMEBUFFER
#endif

void Gfx_ResetRenderTarget(GfxContext* rc)
{
	GL_CALL(glBindFramebufferEXT(GL_FRAMEBUFFER, 0));
}

void Gfx_SetRenderTarget(GfxContext* rc, const RenderTargetConfig& cfg)
{
	GL_CALL(glBindFramebufferEXT(GL_DRAW_FRAMEBUFFER, rc->resources->fbo));

	auto db = rc->resources->textures[cfg.depth].native;

	if (has_stencil(rc->resources->textures[cfg.depth].descr.format))
	{
		GL_CALL(glFramebufferTexture2DEXT(GL_DRAW_FRAMEBUFFER, GL_DEPTH_STENCIL_ATTACHMENT, GL_TEXTURE_2D, db, 0));
	}
	else
	{
		GL_CALL(glFramebufferTexture2DEXT(GL_DRAW_FRAMEBUFFER, GL_DEPTH_ATTACHMENT, GL_TEXTURE_2D, db, 0));
	}

	GL_CALL(glFramebufferTexture2DEXT(GL_DRAW_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_2D, rc->resources->textures[cfg.color[0]].native, 0));
	GL_CALL(glFramebufferTexture2DEXT(GL_DRAW_FRAMEBUFFER, GL_COLOR_ATTACHMENT1, GL_TEXTURE_2D, rc->resources->textures[cfg.color[1]].native, 0));
	GL_CALL(glFramebufferTexture2DEXT(GL_DRAW_FRAMEBUFFER, GL_COLOR_ATTACHMENT2, GL_TEXTURE_2D, rc->resources->textures[cfg.color[2]].native, 0));
	GL_CALL(glFramebufferTexture2DEXT(GL_DRAW_FRAMEBUFFER, GL_COLOR_ATTACHMENT3, GL_TEXTURE_2D, rc->resources->textures[cfg.color[3]].native, 0));

	#if EMSCRIPTEN //TODO: @Yuriy
	GLenum status = glCheckFramebufferStatus(GL_FRAMEBUFFER);
	#else
	GLenum status = glCheckFramebufferStatusEXT(GL_FRAMEBUFFER);
	#endif

	if (status != GL_FRAMEBUFFER_COMPLETE)
		SDL_LogError(SDL_LOG_CATEGORY_RENDER, "Incomplete frame buffer");

}

void Gfx_SetViewport(GfxContext* rc, const Viewport& _viewport)
{
	GL_CALL(glViewport((int)_viewport.x, (int)_viewport.y, (int)_viewport.w, (int)_viewport.h));
}


