#pragma once

#include "Framework.h"
#include "StaticArray.h"
#include "Color.h"

#include <string>
#include <cstring>

#define JAM_GLES2
#define JAM_GLEW

#define JAM_RENDER_API_NONE			0
#define JAM_RENDER_API_DX11			1
#define JAM_RENDER_API_GLES2		2
#define JAM_RENDER_API_GL			3
#define JAM_RENDER_API_EXTERN		4

// Select API

#ifdef JAM_DX11
#define JAM_RENDER_API JAM_RENDER_API_DX11
#endif //JAM_DX11

#ifdef JAM_GLES2
#define JAM_RENDER_API JAM_RENDER_API_GLES2
#endif //JAM_GLES2

#ifdef JAM_OPENGL
#define JAM_RENDER_API JAM_RENDER_API_GL
#endif //JAM_OPENGL

#ifdef JAM_EXTERN_GFX
#define JAM_RENDER_API JAM_RENDER_API_EXTERN
#endif //JAM_EXTERN_GFX

#ifdef JAM_NO_GFX
#define JAM_RENDER_API JAM_RENDER_API_NONE
#endif //JAM_NO_GFX

// use default api id on some platforms (if not explicitly set)

#ifndef JAM_RENDER_API
#ifdef JAM_PLATFORM_NACL
#define JAM_RENDER_API JAM_RENDER_API_GLES2
#endif //JAM_PLATFORM_NACL

#ifdef JAM_PLATFORM_IOS
#define JAM_RENDER_API JAM_RENDER_API_GLES2
#endif //JAM_PLATFORM_IOS

#ifdef JAM_PLATFORM_MACOS
#define JAM_RENDER_API JAM_RENDER_API_GL
#endif //JAM_PLATFORM_MACOS

#ifdef JAM_PLATFORM_LINUX
#define JAM_RENDER_API JAM_RENDER_API_GL
#endif //JAM_PLATFORM_LINUX
#endif //!JAM_RENDER_API

// If we got here and did not select an API, select some reasonable default

#ifndef JAM_RENDER_API
#define JAM_RENDER_API JAM_RENDER_API_NONE
#endif //JAM_RENDER_API

#include "ResourcePool.h"

class GfxContext;
class GfxDevice;
class PixelShaderDescr;
class ShaderDescr;
class VertexFormatDescr;
class VertexShaderDescr;
struct BlendStateDescr;
struct BufferDescr;
struct DepthStencilStateDescr;
struct RasterizerStateDescr;
struct SamplerStateDescr;
struct ShaderCode;
struct TechniqueDescr;
struct TextureDescr;

typedef ResourceHandle<VertexFormatDescr>		VertexFormatHandle;
typedef ResourceHandle<VertexShaderDescr>		VertexShaderHandle;
typedef ResourceHandle<PixelShaderDescr>		PixelShaderHandle;
typedef ResourceHandle<TextureDescr>			TextureHandle;
typedef ResourceHandle<BufferDescr>				BufferHandle;
typedef ResourceHandle<SamplerStateDescr>		SamplerStateHandle;
typedef ResourceHandle<BlendStateDescr>			BlendStateHandle;
typedef ResourceHandle<DepthStencilStateDescr>	DepthStencilStateHandle;
typedef ResourceHandle<RasterizerStateDescr>	RasterizerStateHandle;
typedef ResourceHandle<TechniqueDescr>			TechniqueHandle;

enum class GfxStage
{
	Vertex,
	Pixel,
	Hull,
	Domain,
	Compute,
	count
};

enum class PrimitiveType
{
	PointList,
	LineList,
	LineStrip,
	LineLoop,
	TriangleList,
	TriangleStrip,
	count
};

enum class ClearFlags : u32
{
	Color = 1 << 0,
	Depth = 1 << 1,
	Stencil = 1 << 2,

	DepthStencil = Depth | Stencil,
	StencilColor = Stencil | Color,
	ColorDepth = Color | Depth,

	ColorDepthStencil = Color | Depth | Stencil
};

struct Viewport
{
	Viewport()
		: x(0), y(0), w(1), h(1)
		, depth_min(0), depth_max(1)
	{
	}

	Viewport(const aabb2& bounds, float _depth_min = 0, float _depth_max = 1)
		: x(bounds.tl().x), y(bounds.tl().y), w(bounds.width()), h(bounds.height())
		, depth_min(_depth_min), depth_max(_depth_max)
	{
	}

	Viewport(float _x, float _y, float _w, float _h, float _depth_min = 0, float _depth_max = 1)
		: x(_x), y(_y), w(_w), h(_h)
		, depth_min(_depth_min), depth_max(_depth_max)
	{
	}

	Viewport(int _x, int _y, int _w, int _h, float _depth_min = 0, float _depth_max = 1)
		: Viewport((float)_x, (float)_y, (float)_w, (float)_h, _depth_min, _depth_max)
	{
	}

	bool operator == (const Viewport& rhs) const
	{
		return
			x == rhs.x && y == rhs.y &&
			w == rhs.w && h == rhs.h &&
			depth_min == rhs.depth_min &&
			depth_max == rhs.depth_max;
	}

	bool operator != (const Viewport& rhs) const
	{
		return !(*this == rhs);
	}

	float x; // top left x
	float y; // top left y
	float w; // width
	float h; // height
	float depth_min;
	float depth_max;
};

struct GfxRect
{
	int left, top, right, bottom;
};

enum GfxFormatType
{
	GfxFormatType_Unknown = 0,
	GfxFormatType_Float = 1,
	GfxFormatType_Float_Typeless = 2,
	GfxFormatType_Float_Uint = 3,
	GfxFormatType_Sint = 4,
	GfxFormatType_Snorm = 5,
	GfxFormatType_sRGB = 6,
	GfxFormatType_Typeless = 7,
	GfxFormatType_Typeless_Uint = 8,
	GfxFormatType_Ufloat = 9,
	GfxFormatType_Uint = 10,
	GfxFormatType_Unorm = 11,
	GfxFormatType_Unorm_Typeless = 12,
	GfxFormatType_Unorm_Uint = 13,
};

enum GfxFormatStorage
{
	GfxFormatStorage_Unknown = 0,
	GfxFormatStorage_BC1 = 1 << 8,
	GfxFormatStorage_BC2 = 2 << 8,
	GfxFormatStorage_BC3 = 3 << 8,
	GfxFormatStorage_BC4 = 4 << 8,
	GfxFormatStorage_BC5 = 5 << 8,
	GfxFormatStorage_BC6H = 6 << 8,
	GfxFormatStorage_BC7 = 7 << 8,
	GfxFormatStorage_R16 = 8 << 8,
	GfxFormatStorage_R24G8 = 9 << 8,
	GfxFormatStorage_R32 = 10 << 8,
	GfxFormatStorage_R32G8X24 = 11 << 8,
	GfxFormatStorage_R8 = 12 << 8,
	GfxFormatStorage_RG16 = 13 << 8,
	GfxFormatStorage_RG32 = 14 << 8,
	GfxFormatStorage_RG8 = 15 << 8,
	GfxFormatStorage_RGB16 = 16 << 8,
	GfxFormatStorage_RGB32 = 17 << 8,
	GfxFormatStorage_RGB8 = 18 << 8,
	GfxFormatStorage_RGBA16 = 19 << 8,
	GfxFormatStorage_RGBA32 = 20 << 8,
	GfxFormatStorage_RGBA8 = 21 << 8,
};

enum GfxFormatComponent
{
	GfxFormatComponent_Unknown = 0,
	GfxFormatComponent_R = 1 << 16,
	GfxFormatComponent_G = 1 << 17,
	GfxFormatComponent_B = 1 << 18,
	GfxFormatComponent_A = 1 << 19,
	GfxFormatComponent_Depth = 1 << 20,
	GfxFormatComponent_Stencil = 1 << 21,

	GfxFormatComponent_RG = GfxFormatComponent_R | GfxFormatComponent_G,
	GfxFormatComponent_RGB = GfxFormatComponent_RG | GfxFormatComponent_B,
	GfxFormatComponent_RGBA = GfxFormatComponent_RGB | GfxFormatComponent_A,
	GfxFormatComponent_DepthStencil = GfxFormatComponent_Depth | GfxFormatComponent_Stencil,
};

enum GfxFormat
{
	GfxFormat_Unknown = GfxFormatComponent_Unknown | GfxFormatStorage_Unknown | GfxFormatType_Unknown,

	GfxFormat_D24_Unorm_S8_Uint = GfxFormatStorage_R24G8 | GfxFormatType_Unorm_Uint | GfxFormatComponent_DepthStencil,
	GfxFormat_D24_Unorm_X8 = GfxFormatStorage_R24G8 | GfxFormatType_Unorm_Typeless | GfxFormatComponent_Depth,
	GfxFormat_D32_Float = GfxFormatStorage_R32 | GfxFormatType_Float | GfxFormatComponent_Depth,
	GfxFormat_D32_Float_S8_Uint = GfxFormatStorage_R32G8X24 | GfxFormatType_Float_Uint | GfxFormatComponent_DepthStencil,
	GfxFormat_R16_Float = GfxFormatStorage_R16 | GfxFormatType_Float | GfxFormatComponent_R,
	GfxFormat_R16_Uint = GfxFormatStorage_R16 | GfxFormatType_Uint | GfxFormatComponent_R,
	GfxFormat_R32_Float = GfxFormatStorage_R32 | GfxFormatType_Float | GfxFormatComponent_R,
	GfxFormat_R32_Uint = GfxFormatStorage_R32 | GfxFormatType_Uint | GfxFormatComponent_R,
	GfxFormat_RG16_Float = GfxFormatStorage_RG16 | GfxFormatType_Float | GfxFormatComponent_RG,
	GfxFormat_RG32_Float = GfxFormatStorage_RG32 | GfxFormatType_Float | GfxFormatComponent_RG,
	GfxFormat_RGB8_Unorm = GfxFormatStorage_RGB8 | GfxFormatType_Unorm | GfxFormatComponent_RGB,
	GfxFormat_RGBA16_Float = GfxFormatStorage_RGBA16 | GfxFormatType_Float | GfxFormatComponent_RGBA,
	GfxFormat_RGBA16_Unorm = GfxFormatStorage_RGB16 | GfxFormatType_Unorm | GfxFormatComponent_RGBA,
	GfxFormat_RGBA32_Float = GfxFormatStorage_RGBA32 | GfxFormatType_Float | GfxFormatComponent_RGBA,
	GfxFormat_RGBA8_Unorm = GfxFormatStorage_RGBA8 | GfxFormatType_Unorm | GfxFormatComponent_RGBA,
};

inline GfxFormatType GfxFormatExtractType(GfxFormat fmt)
{
	return (GfxFormatType)(fmt & 0x000000FF);
}

inline GfxFormatStorage GfxFormatExtractStorage(GfxFormat fmt)
{
	return (GfxFormatStorage)(fmt & 0x0000FF00);
}

inline GfxFormatComponent GfxFormatExtractComponent(GfxFormat fmt)
{
	return (GfxFormatComponent)(fmt & 0x00FF0000);
}

u32 bits_per_pixel(GfxFormat fmt);
u32 bits_per_pixel(GfxFormatStorage fmt);

const char* to_string(GfxFormatStorage storage);
const char* to_string(GfxFormatType type);

enum class TextureType
{
	Tex2D,
	Tex3D,
	TexCUBE
};

struct TextureDescr
{
	inline TextureDescr()
		: width(0)
		, height(0)
		, depth(0)
		, format(GfxFormat_Unknown)
		, type(TextureType::Tex2D)
		, render_target(false)
		, dynamic(false)
        , repeat(false)
	{

	}

	static inline TextureDescr create_2D(u32 _width, u32 _height,
		GfxFormat _format = GfxFormat_RGBA8_Unorm,
		bool _render_target = false, bool _dynamic = false, bool _repeat = false)
	{
		TextureDescr res;
		res.width = _width;
		res.height = _height;
		res.depth = 0;
		res.format = _format;
		res.type = TextureType::Tex2D;
		res.render_target = _render_target;
		res.dynamic = _dynamic;
        res.repeat = _repeat;
		return res;
	}
	static inline TextureDescr create_3D(u32 _width, u32 _height,
		u32 _depth, GfxFormat _format = GfxFormat_RGBA8_Unorm,
		bool _render_target = false, bool _dynamic = false, bool _repeat = false)
	{
		TextureDescr res;
		res.width = _width;
		res.height = _height;
		res.depth = _depth;
		res.format = _format;
		res.type = TextureType::Tex3D;
		res.render_target = _render_target;
		res.dynamic = _dynamic;
        res.repeat = _repeat;
		return res;
	}
	static inline TextureDescr create_CUBE(u32 _size, GfxFormat _format = GfxFormat_RGBA8_Unorm, bool _dynamic = false, bool _repeat = false)
	{
		TextureDescr res;
		res.width = _size;
		res.height = _size;
		res.depth = 0;
		res.format = _format;
		res.type = TextureType::TexCUBE;
		res.render_target = false;
		res.dynamic = _dynamic;
        res.repeat = _repeat;
		return res;
	}

	inline bool operator == (const TextureDescr& rhs)
	{
		return	width == rhs.width &&
			height == rhs.height &&
			depth == rhs.depth &&
			format == rhs.format &&
			type == rhs.type &&
			render_target == rhs.render_target &&
			dynamic == rhs.dynamic &&
            repeat == rhs.repeat;
	}

	inline bool operator != (const TextureDescr& rhs) { return !(*this == rhs); }

	u32				width;
	u32				height;
	u32				depth;
	GfxFormat			format;
	TextureType			type;
	bool				render_target;
	bool				dynamic;
    bool                repeat;
};
enum class BufferType
{
	Vertex,
	Index,

	count
};

enum class BufferMode
{
	WriteOnce,
	WriteRarely,
	WriteOften,

	count
};

struct BufferDescr
{
	BufferDescr(){}
	BufferDescr(BufferType _type, BufferMode _mode, GfxFormat _format, u32 _stride, u32 _count)
		: type(_type)
		, mode(_mode)
		, format(_format)
		, stride(_stride)
		, count(_count)
	{
	}

	BufferType type;
	BufferMode mode;
	GfxFormat format;
	u32 stride;
	u32 count;
};

enum class Blend
{
	Zero,
	One,
	SrcColor,
	InvSrcColor,
	SrcAlpha,
	InvSrcAlpha,
	DestAlpha,
	InvDestAlpha,
	DestColor,
	InvDestColor,

	count
};

enum class BlendOp
{
	Add,
	Subtract,
	RevSubtract,
	Min,
	Max,

	count
};

enum class TextureFilter
{
	Point,
	Linear,
	Anisotropic,

	count
};

enum class TextureWrap
{
	Wrap,
	Mirror,
	Clamp,
	Borderk,

	count
};

enum class CompareFunc
{
	Never,
	Less,
	Equal,
	LessEqual,
	Greater,
	NotEqual,
	GreaterEqual,
	Always,

	count
};

struct BlendStateDescr
{
	Blend		src = Blend::One;
	Blend		dst = Blend::Zero;
	BlendOp		op = BlendOp::Add;

	Blend		alpha_src = Blend::One;
	Blend		alpha_dst = Blend::Zero;
	BlendOp		alpha_op = BlendOp::Add;

	bool		alpha_separate = false;
	bool		enable = false;

	// Helper functions to create common blend states

	static BlendStateDescr make_solid();
	static BlendStateDescr make_alphablend();
	static BlendStateDescr make_additive();
	static BlendStateDescr make_premultiplied_alpha();

};

struct SamplerStateDescr
{
	TextureFilter filter_min = TextureFilter::Linear;
	TextureFilter filter_mag = TextureFilter::Linear;
	TextureFilter filter_mip = TextureFilter::Linear;

	// Helper functions

	static SamplerStateDescr make_linear();
	static SamplerStateDescr make_point();
};

struct DepthStencilStateDescr
{
	CompareFunc cmp_func = CompareFunc::LessEqual;
	bool enable = true;
	bool write_enable = true;

	//TODO: Stencil and others
};

enum class FillMode
{
	Solid,
	Wireframe
};

enum class CullMode
{
	CW,
	CCW,
};

struct RasterizerStateDescr
{
	FillMode fill_mode = FillMode::Solid;
	CullMode cull_mode = CullMode::CCW;
	float depth_bias = 0.0f;
	float depth_bias_slope_scale = 0.0f;
	bool scissor_enable = false;
};

class VertexFormatDescr
{

	static const u32 MaxStreams = 8;
	static const u32 MaxElements = 16;

public:

	enum class DataType : u8
	{
		Unused = 0,
		Float1 = 1,
		Float2 = 2,
		Float3 = 3,
		Float4 = 4,
		Half2 = 5,
		Half4 = 6,
		Short2 = 7,
		Short2N = 8,
		UByte4 = 9,
		Dec3N = 10,
		Color = 11,
	};

	enum class UsageType : u8
	{
		Unused = 0,
		Position = 1,
		Texcoord = 2,
		Color = 3,
		Normal = 4,
		Binormal = 5,
		Tangent = 6,
	};

	struct Element
	{
		Element(){}
		Element(u16 _stream, DataType _type, UsageType _usage, u8 _index);

		u16		stream;
		u16		size;
		u16		offset;
		u8		index;
		DataType	type;
		UsageType	usage;

		bool operator == (const Element& rhs) const
		{
			return stream == rhs.stream
				&& type == rhs.type
				&& usage == rhs.usage
				&& index == rhs.index;
		}
	};

public:

	VertexFormatDescr();

	inline const Element&		element(u32 n)		const	{ return m_elements[n]; }
	inline size_t				num_elements()			const	{ return m_elements.size(); }
	inline u16				stream_stride(u32 n)	const	{ return m_stream_offset[n]; }

	inline bool					has_position() const { return m_has_position; }
	inline bool					has_normal()   const { return m_has_normal; }
	inline bool					has_color()   const { return m_has_color; }

	bool operator == (const VertexFormatDescr& rhs) const
	{
		if (num_elements() == rhs.num_elements())
		{
			for (u32 i = 0; i < num_elements(); ++i)
			{
				if (!(m_elements[i] == rhs.m_elements[i])) return false;
			}
			return true;
		}
		else
		{
			return false;
		}
	}

	inline void add(u16 _stream, DataType _type, UsageType _usage, u8 _index)
	{
		Element elem(_stream, _type, _usage, _index);
		add(elem);
	}

	void add(Element element);

private:

	StaticArray<Element, MaxElements> m_elements;

	u16 m_stream_offset[MaxStreams];

	bool m_has_position;
	bool m_has_normal;
	bool m_has_color;
};

enum class ShaderType
{
	Vertex,
	Pixel,
	Geometry,
	Hull,
	Domain,
	Compute
};

struct ShaderConstantIndex
{
	inline ShaderConstantIndex()
		: index(u32(-1))
		, count(u32(-1))
	{}

	inline ShaderConstantIndex(u32 _index, u32 _count = 1)
		: index(_index)
		, count(_count)
	{}

	bool is_valid() const { return index != u32(-1); }

	u32 index;
	u32 count;
};

struct ShaderConstantDescr
{
	std::string			name;
	u32 				register_index;
	u32 				register_count;

	ShaderConstantDescr()
		: name()
		, register_index(0)
		, register_count(0)
	{
	}

	inline bool operator == (const ShaderConstantDescr& rhs) const
	{
		return	register_index == rhs.register_index &&
			register_count == rhs.register_count &&
			name == rhs.name;
	}
};

struct ShaderConstantTable : public std::vector < ShaderConstantDescr >
{
	ShaderConstantIndex index_from_name(const char* name) const
	{
		ShaderConstantIndex res;
		for (u32 i = 0; i < size(); ++i)
		{
			if ((*this)[i].name == name)
			{
				res.index = (*this)[i].register_index;
				res.count = (*this)[i].register_count;
			}
		}
		return res;
	}
};

enum ShaderCodeType
{
	ShaderCodeType_Binary,
	ShaderCodeType_Text
};

struct ShaderCode : public std::vector < char >
{
	ShaderCode() : type(ShaderCodeType_Binary){}

	ShaderCode(ShaderCodeType _type, const char* _code, size_t _size) : type(_type)
	{
		if (_type == ShaderCodeType_Text && _size == 0)
		{
			_size = (size_t)strlen(_code) + 1;
			this->resize(_size);
			memcpy(this->data(), _code, _size);
		}
		else
		{
			this->resize(_size);
			memcpy(this->data(), _code, _size);
		}
	}

	ShaderCodeType type;
};

struct ShaderCodeText : public ShaderCode
{
	ShaderCodeText(const char* _text, size_t _length = 0)
		: ShaderCode(ShaderCodeType_Text, _text, _length)
	{}
};

struct ShaderCodeBinary : public ShaderCode
{
	ShaderCodeBinary(const char* _bytes, size_t _size)
		: ShaderCode(ShaderCodeType_Binary, _bytes, _size)
	{}
};

struct ShaderConstantBindings
{
	enum BindingType
	{
		BindingType_Scalar,
		BindingType_Vector,
		BindingType_Matrix,
		BindingType_Sampler
	};

	struct Item
	{
		const char* name;
		const void* data;
		u32 count;
		u32 idx;
		BindingType type;
	};

	bool add_constant(const char* name, const float* data, u32 count = 1)
	{
		Item it;
		it.name = name;
		it.data = data;
		it.count = count;
		it.idx = 0;
		it.type = BindingType_Scalar;
		return items.push_back(it);
	}

	bool add_constant(const char* name, const vec4* data, u32 count = 1)
	{
		Item it;
		it.name = name;
		it.data = data;
		it.count = count;
		it.idx = 0;
		it.type = BindingType_Vector;
		return items.push_back(it);
	}

	bool add_constant(const char* name, const mat4x4* data, u32 count = 1)
	{
		Item it;
		it.name = name;
		it.data = data;
		it.count = count;
		it.idx = 0;
		it.type = BindingType_Matrix;
		return items.push_back(it);
	}

	bool add_sampler(const char* name, u32 idx)
	{
		Item it;
		it.name = name;
		it.data = nullptr;
		it.count = 1;
		it.idx = idx;
		it.type = BindingType_Sampler;
		return items.push_back(it);
	}

	StaticArray<Item, 128> items;
};

struct TechniqueDescr
{
	TechniqueDescr(PixelShaderHandle a_ps = InvalidResourceHandle(), VertexShaderHandle a_vs = InvalidResourceHandle(),
		VertexFormatHandle a_vf = InvalidResourceHandle(), const ShaderConstantBindings* a_bindings = nullptr)
		: ps(a_ps)
		, vs(a_vs)
		, vf(a_vf)
		, bindings(a_bindings)
	{}

	PixelShaderHandle				ps;
	VertexShaderHandle				vs;
	VertexFormatHandle				vf;
	const ShaderConstantBindings*	bindings;
};

// This class should be used purely for debugging and error detection purposes.
// It is not supposed to be a classic reference counted pointer scheme.
class ReferenceCount
{

public:

	ReferenceCount() 
		: m_count(0)
	{
	}

	~ReferenceCount()
	{
		SDL_assert(m_count == 0);
	}

	u32 get_reference_count() const
	{
		return m_count;
	}

	void add_reference()
	{
		m_count++;
	}

	void remove_reference()
	{
		SDL_assert(m_count != 0);
		m_count--;
	}

private:

	u32 m_count;

};

struct GfxDeviceStats
{
	u32 draw_calls = 0;
	u32 vertices = 0;
};

struct BufferLock
{
	void* data = nullptr;
	u32 size = 0;
	BufferHandle handle;
};

struct TextureLock
{
	void* data = nullptr;
	u32 size = 0;
	TextureHandle handle;
};

struct RenderTargetConfig
{
	TextureHandle color[8];
	TextureHandle depth;
};

struct GfxDeviceConfig
{
	u32 back_buffer_width = 640;
	u32 back_buffer_height = 480;
	u32 present_interval = 0;
	bool use_full_screen = false;
	bool debug = false;
};

// device
GfxDevice* Gfx_CreateDevice(SDL_Window* window, const GfxDeviceConfig& cfg);
void Gfx_DestroyDevice(GfxDevice* dev);
void Gfx_AddReference(GfxDevice* dev);
void Gfx_RemoveReference(GfxDevice* dev);
GfxContext* Gfx_FrameBegin(GfxDevice* dev);
void Gfx_FrameEnd(GfxDevice* dev);
void Gfx_Present(GfxDevice* dev);
void Gfx_SetPresentInterval(GfxDevice* dev, u32 interval);

const GfxDeviceStats& Gfx_Stats(GfxDevice* dev);
void Gfx_ResetStats(GfxDevice* dev);

// vertex format
VertexFormatHandle Gfx_VertexFormat_Create(GfxDevice* dev, const VertexFormatDescr& fmt);
void Gfx_VertexFormat_Destroy(GfxDevice* dev, VertexFormatHandle h);

// vertex shader
VertexShaderHandle Gfx_VertexShader_Create(GfxDevice* dev, const ShaderCode& code);
void Gfx_VertexShader_Destroy(GfxDevice* dev, VertexShaderHandle& vsh);

// pixel shader
PixelShaderHandle Gfx_PixelShader_Create(GfxDevice* dev, const ShaderCode& code);
void Gfx_PixelShader_Destroy(GfxDevice* dev, PixelShaderHandle& psh);

// technique
TechniqueHandle Gfx_Technique_Create(GfxDevice* dev, const TechniqueDescr& descr);
void Gfx_Technique_Destroy(GfxDevice* dev, TechniqueHandle h);

// texture
TextureHandle Gfx_Texture_Load(GfxDevice* dev, const char* filename, TextureType type = TextureType::Tex2D);
TextureHandle Gfx_Texture_Create(GfxDevice* dev, const TextureDescr& tex, const void* data = nullptr, size_t size = 0);
void Gfx_Texture_Update(GfxDevice* dev, TextureHandle tex, const void* data = nullptr);
const TextureDescr& Gfx_Texture_Descr(GfxDevice* dev, TextureHandle h);
void Gfx_Texture_Destroy(GfxDevice* dev, TextureHandle& th);

// blend state
BlendStateHandle Gfx_BlendState_Create(GfxDevice* dev, const BlendStateDescr& descr);
void Gfx_BlendState_Destroy(GfxDevice* dev, BlendStateHandle h);

// sampler state
SamplerStateHandle Gfx_SamplerState_Create(GfxDevice* dev, const SamplerStateDescr& descr);
void Gfx_SamplerState_Destroy(GfxDevice* dev, SamplerStateHandle h);

// depth stencil state
DepthStencilStateHandle Gfx_DepthStencilState_Create(GfxDevice* dev, const DepthStencilStateDescr& descr);
void Gfx_DepthStencilState_Destroy(GfxDevice* dev, DepthStencilStateHandle h);

// rasterizer state
RasterizerStateHandle Gfx_RasterizerState_Create(GfxDevice* dev, const RasterizerStateDescr& descr);
void Gfx_RasterizerState_Destroy(GfxDevice* dev, RasterizerStateHandle h);

// render target
TextureHandle Gfx_RenderTarget_Create(GfxDevice* dev, const TextureDescr& tex);

// buffers
BufferHandle Gfx_Buffer_Create(GfxDevice* dev, const BufferDescr& descr, const void* data = nullptr);
void Gfx_Buffer_Update(GfxDevice* dev, BufferHandle h, void* data, u32 offset = 0, u32 size = 0, bool discard = false);
void Gfx_Buffer_Destroy(GfxDevice* dev, BufferHandle h);

// context
GfxContext* Gfx_Immediate_Context(GfxDevice* dev);

void Gfx_Clear(GfxContext* rc, ColorRGBA color, ClearFlags clear_flags = ClearFlags::ColorDepthStencil, float depth = 1.0f, u32 stencil = 0);
void Gfx_SetViewport(GfxContext* rc, const Viewport& _viewport);
void Gfx_SetScissorRect(GfxContext* rc, const GfxRect& rect);
void Gfx_SetTechnique(GfxContext* rc, TechniqueHandle h);
void Gfx_SetLineWidth(GfxContext* rc, float width);
void Gfx_SetPrimitive(GfxContext* rc, PrimitiveType type);
void Gfx_SetIndexStream(GfxContext* rc, BufferHandle h);
void Gfx_SetVertexStream(GfxContext* rc, u32 idx, BufferHandle h);
void Gfx_SetPixelTexture(GfxContext* rc, u32 idx, TextureHandle h);
void Gfx_SetPixelSamplerState(GfxContext* rc, u32 idx, SamplerStateHandle next_state);
void Gfx_SetBlendState(GfxContext* rc, BlendStateHandle next_state);
void Gfx_SetDepthStencilState(GfxContext* rc, DepthStencilStateHandle next_state);
void Gfx_SetRasterizerState(GfxContext* rc, RasterizerStateHandle next_state);

void Gfx_ResetRenderTarget(GfxContext* rc);
void Gfx_SetRenderTarget(GfxContext* rc, const RenderTargetConfig& cfg);

void Gfx_Draw(GfxContext* rc, u32 first_vertex, u32 vertex_count);
void Gfx_DrawIndexed(GfxContext* rc, u32 index_count, u32 first_index, u32 base_vertex, u32 num_vertices);

//

inline void Gfx_Destroy(GfxDevice* dev, VertexFormatHandle h) { Gfx_VertexFormat_Destroy(dev, h); }
inline void Gfx_Destroy(GfxDevice* dev, VertexShaderHandle h) { Gfx_VertexShader_Destroy(dev, h); }
inline void Gfx_Destroy(GfxDevice* dev, PixelShaderHandle h) { Gfx_PixelShader_Destroy(dev, h); }
inline void Gfx_Destroy(GfxDevice* dev, TechniqueHandle h) { Gfx_Technique_Destroy(dev, h); }
inline void Gfx_Destroy(GfxDevice* dev, TextureHandle h) { Gfx_Texture_Destroy(dev, h); }
inline void Gfx_Destroy(GfxDevice* dev, BlendStateHandle h) { Gfx_BlendState_Destroy(dev, h); }
inline void Gfx_Destroy(GfxDevice* dev, SamplerStateHandle h) { Gfx_SamplerState_Destroy(dev, h); }
inline void Gfx_Destroy(GfxDevice* dev, DepthStencilStateHandle h) { Gfx_DepthStencilState_Destroy(dev, h); }
inline void Gfx_Destroy(GfxDevice* dev, RasterizerStateHandle h) { Gfx_RasterizerState_Destroy(dev, h); }
inline void Gfx_Destroy(GfxDevice* dev, BufferHandle h) { Gfx_Buffer_Destroy(dev, h); }

