#include <SDL.h>
#include <imgui.h>
#include <stb_image.h>
#include "Framework.h"
#include "GL/glew.h"

#include <Framework/PrimitiveBatch.h>
#include <Framework/Graphics.h>
#include "Gui.h"

#define OFFSETOF(TYPE, ELEMENT) ((size_t)&(((TYPE *)0)->ELEMENT))
static GLuint fontTex;
static bool mousePressed[2] = { false, false };

static GfxContext* s_context = nullptr;
static BlendStateHandle s_blend_state;
static DepthStencilStateHandle s_depth_state;
static RasterizerStateHandle s_raster_state;
static PrimitiveBatch* s_prim = nullptr;
static TextureHandle s_font_texture;

static void imgui_impl_render(ImDrawList** const cmd_lists, int cmd_lists_count)
{
	Gfx_SetBlendState(s_context, s_blend_state);
	Gfx_SetRasterizerState(s_context, s_raster_state);
	Gfx_SetDepthStencilState(s_context, s_depth_state);

	int w, h;
	SDL_GetWindowSize(g_window, &w, &h);
	aabb2 r;
	r.min = vec2(0, 0);
	r.max = vec2(w, h);
	s_prim->begin2d(r, false);
	s_prim->setTexture(s_font_texture);
	s_prim->setSampler(PrimitiveBatch::SamplerState_Point);

	int resolutionScaleX = 1, resolutionScaleY = 1;

	auto convert_rect = [&](ImVec4 v)
	{
		GfxRect rect;
		rect.left = (int)v.x;
		rect.top = h - (int)v.w;
		rect.right = (int)v.z;
		rect.bottom = h - (int)v.y;

		rect.left = (int)(rect.left * resolutionScaleX);
		rect.right = (int)(rect.right * resolutionScaleX);
		rect.top = (int)(rect.top * resolutionScaleY);
		rect.bottom = (int)(rect.bottom * resolutionScaleY);

		return rect;
	};

	for (int n = 0; n < cmd_lists_count; n++)
	{
		const ImDrawList* cmd_list = cmd_lists[n];
		u32 vtx_offset = 0;
		for (const auto& cmd : cmd_list->commands)
		{
			Gfx_SetScissorRect(s_context, convert_rect(cmd.clip_rect));
			PrimitiveBatch::BatchVertex* verts = s_prim->drawVertices(PrimitiveType::TriangleList, cmd.vtx_count);
			for (u32 i = 0; i < cmd.vtx_count; ++i)
			{
				verts[i].pos.x = cmd_list->vtx_buffer[vtx_offset].pos.x;
				verts[i].pos.y = cmd_list->vtx_buffer[vtx_offset].pos.y;
				verts[i].pos.z = 1;

				verts[i].tex.x = cmd_list->vtx_buffer[vtx_offset].uv.x;
				verts[i].tex.y = cmd_list->vtx_buffer[vtx_offset].uv.y;

				verts[i].col0 = ColorRGBA8(cmd_list->vtx_buffer[vtx_offset].col);
				++vtx_offset;
			}
			s_prim->flush();
		}
	}

	s_prim->end2d();
}

#if !EMSCRIPTEN //TODO: @Yuriy
static const char* ImImpl_GetClipboardTextFn()
{
	return SDL_GetClipboardText();
}

static void ImImpl_SetClipboardTextFn(const char* text)
{
	SDL_SetClipboardText(text);
}
#endif

void initGui()
{
	s_context = Gfx_Immediate_Context(g_device);

	BlendStateDescr blend_descr = BlendStateDescr::make_alphablend();
	s_blend_state = Gfx_BlendState_Create(g_device, blend_descr);

	RasterizerStateDescr raster_descr;
	raster_descr.scissor_enable = true;
	//raster_descr.cull_mode = CullMode::None;
	s_raster_state = Gfx_RasterizerState_Create(g_device, raster_descr);

	DepthStencilStateDescr depth_descr;
	depth_descr.enable = false;
	s_depth_state = Gfx_DepthStencilState_Create(g_device, depth_descr);

	ImGuiIO& io = ImGui::GetIO();
	io.DeltaTime = 1.0f / 60.0f;

#if !EMSCRIPTEN //TODO: @Yuriy
	io.IniFilename = SDL_strdup((std::string(SDL_GetPrefPath("Jammy", "ggj")) + "imgui.ini").c_str());
#endif

	// Default font (embedded in code)
	const void* png_data;
	unsigned int png_size;
	ImGui::GetDefaultFontData(NULL, NULL, &png_data, &png_size);
	int tex_x, tex_y, tex_comp;
	void* tex_data = stbi_load_from_memory((const unsigned char*)png_data, (int)png_size, &tex_x, &tex_y, &tex_comp, 0);
	s_font_texture = Gfx_Texture_Create(g_device, TextureDescr::create_2D(tex_x, tex_y, GfxFormat_RGBA8_Unorm), tex_data, tex_x*tex_y * 4);
	stbi_image_free(tex_data);

	io.KeyMap[ImGuiKey_Tab] = SDL_GetScancodeFromKey(SDLK_TAB);
	io.KeyMap[ImGuiKey_LeftArrow] = SDL_GetScancodeFromKey(SDLK_LEFT);
	io.KeyMap[ImGuiKey_RightArrow] = SDL_GetScancodeFromKey(SDLK_RIGHT);
	io.KeyMap[ImGuiKey_UpArrow] = SDL_GetScancodeFromKey(SDLK_UP);
	io.KeyMap[ImGuiKey_DownArrow] = SDL_GetScancodeFromKey(SDLK_DOWN);
	io.KeyMap[ImGuiKey_Home] = SDL_GetScancodeFromKey(SDLK_HOME);
	io.KeyMap[ImGuiKey_End] = SDL_GetScancodeFromKey(SDLK_END);
	io.KeyMap[ImGuiKey_Delete] = SDL_GetScancodeFromKey(SDLK_DELETE);
	io.KeyMap[ImGuiKey_Backspace] = SDL_GetScancodeFromKey(SDLK_BACKSPACE);
	io.KeyMap[ImGuiKey_Enter] = SDL_GetScancodeFromKey(SDLK_RETURN);
	io.KeyMap[ImGuiKey_Escape] = SDL_GetScancodeFromKey(SDLK_ESCAPE);
	io.KeyMap[ImGuiKey_A] = SDLK_a;
	io.KeyMap[ImGuiKey_C] = SDLK_c;
	io.KeyMap[ImGuiKey_V] = SDLK_v;
	io.KeyMap[ImGuiKey_X] = SDLK_x;
	io.KeyMap[ImGuiKey_Y] = SDLK_y;
	io.KeyMap[ImGuiKey_Z] = SDLK_z;

	io.RenderDrawListsFn = imgui_impl_render;

#if !EMSCRIPTEN //TODO: @Yuriy
	io.SetClipboardTextFn = ImImpl_SetClipboardTextFn;
	io.GetClipboardTextFn = ImImpl_GetClipboardTextFn;
#endif

}

void shutdownGui()
{
	delete s_prim;
}

bool updateGuiEvent(SDL_Event& event)
{
	ImGuiIO& io = ImGui::GetIO();
	const Uint32 eventType = event.type;
	switch (eventType)
	{
	case SDL_MOUSEWHEEL:
	{
		io.MouseWheel += event.wheel.y;
	}
	case SDL_MOUSEBUTTONDOWN:
	case SDL_MOUSEBUTTONUP:
	{
		if (event.button.button == 1) {
			io.MouseDown[event.button.button == 1 ? 0 : 1] = event.type == SDL_MOUSEBUTTONDOWN;
		}
		break;
	}
	case SDL_KEYDOWN:
	case SDL_KEYUP:
	{
		SDL_Scancode key = event.key.keysym.scancode;
		if (key >= 0 && key < 512)
		{
			SDL_Keymod modstate = SDL_GetModState();

			if (event.type == SDL_KEYDOWN)
			{
				io.KeysDown[key] = true;
			}
			else
			{
				io.KeysDown[key] = false;
			}
			io.KeyCtrl = (modstate & KMOD_CTRL) != 0;
			io.KeyShift = (modstate & KMOD_SHIFT) != 0;
		}
		break;
	}
	case SDL_TEXTINPUT:
	{
		const SDL_TextInputEvent& textEvent = event.text;
		for (uint i = 0; i < SDL_TEXTINPUTEVENT_TEXT_SIZE && textEvent.text[i]; i++)
		{
			uint32_t keycode = textEvent.text[i];
			if (keycode >= 32 && keycode <= 255)
			{
				io.AddInputCharacter((char)keycode);
			}
		}
		break;
	}

	}

	return io.WantCaptureMouse || io.WantCaptureKeyboard;
}

void updateGuiBegin()
{
	ImGuiIO& io = ImGui::GetIO();

	// Setup resolution (every frame to accommodate for window resizing)
	int w, h;
	SDL_GetWindowSize(g_window, &w, &h);
	io.DisplaySize = ImVec2((float)w, (float) h);                                   // Display size, in pixels. For clamping windows positions.

	// Setup time step
	io.DeltaTime = 1/60.f;

	// Setup inputs
	// (we already got mouse wheel, keyboard keys & characters from glfw callbacks polled in glfwPollEvents())
	int mouse_x, mouse_y;
	Uint32 state = SDL_GetMouseState(&mouse_x, &mouse_y);
	io.MousePos = ImVec2((float)mouse_x, (float)mouse_y);                                          // Mouse position, in pixels (set to -1,-1 if no mouse / on another screen, etc.)

	// Start the frame
	ImGui::NewFrame();
	SDL_StartTextInput();
}

void updateGuiEnd(PrimitiveBatch *prim)
{
	SDL_assert(s_prim == nullptr);
	SDL_assert(prim);

	s_prim = prim;
	SDL_StopTextInput();
	ImGui::Render();
	s_prim = nullptr;
}