#include "Image.h"

#pragma pack(push)
#pragma pack(1)
struct DDSHeader // based on ddraw.h
{
	static const u32 DDPF_ALPHAPIXELS = 0x00000001;
	static const u32 DDPF_FOURCC = 0x00000004;
	static const u32 DDPF_RGB = 0x00000040;
	static const u32 DDPF_LUMINANCE = 0x00020000;

	static const u32 DDPF_RGBA = DDPF_RGB | DDPF_ALPHAPIXELS;

	u32 dwMagic;
	u32 dwSize;
	u32 dwFlags;
	u32 dwHeight;
	u32 dwWidth;
	u32 dwPitchOrLinearSize;
	u32 dwDepth;
	u32 dwMipMapCount;
	u32 dwReserved[11];

	struct
	{
		u32 dwSize;
		u32 dwFlags;
		u32 dwFourCC;
		u32 dwRGBBitCount;
		u32 dwRBitMask;
		u32 dwGBitMask;
		u32 dwBBitMask;
		u32 dwRGBAlphaBitMask;
	} ddpfPixelFormat;

	struct
	{
		u32 dwCaps1;
		u32 dwCaps2;
		u32 Reserved[2];
	} ddsCaps;

	u32 dwReserved2;
};

struct TGAHeader
{
	u8  identsize;          // size of ID field that follows 18 byte header (0 usually)
	u8  colormaptype;      // type of color map 0=none, 1=has palette
	u8  imagetype;          // type of image 0=none,1=indexed,2=rgb,3=grey,+8=rle packed

	s16 colormapstart;     // first color map entry in palette
	s16 colormaplength;    // number of colors in palette
	u8  colormapbits;      // number of bits per palette entry 15,16,24,32

	s16 xstart;             // image x origin
	s16 ystart;             // image y origin
	s16 width;              // image width in pixels
	s16 height;             // image height in pixels
	u8  bits;               // image bits per pixel 8,16,24,32
	u8  descriptor;         // image descriptor bits (vh flip bits)

	// pixel data follows header

};


#pragma pack(pop)

inline u32 make_fourcc(char a, char b, char c, char d)
{
	return (a | (b << 8) | (c << 16) | (d << 24));
}

//////////////////////////////////////////////////////////////////////////

Image::Image()
	: m_format(GfxFormat_Unknown)
	, m_width(0)
	, m_height(0)
	, m_data(nullptr)
	, m_data_size_in_bytes(0)
{
}

Image::~Image()
{
	delete[] m_data;
}

bool Image::create(u32 w, u32 h, GfxFormat format)
{
	delete m_data;

	m_format = format;

	m_width = w;
	m_height = h;

	u32 size_in_bytes = m_width * m_height * bits_per_pixel(m_format) / 8;

	m_data = new u8[size_in_bytes];
	m_data_size_in_bytes = size_in_bytes;

	return true;
}

#if 0 // TODO
bool Image::read_dds(DataStream& stream)
{
	// reset image

	m_format = GfxFormat_Unknown;
	m_width = 0;
	m_height = 0;
	delete m_data;
	m_data = nullptr;

	// load header

	DDSHeader header = { 0 };
	stream.read(&header, sizeof(header));

	if (header.dwMagic != make_fourcc('D', 'D', 'S', ' '))
	{
		Log::error("Image::read_dds() DDS file identifier not found.");
		return false;
	}

	// detect format

	if (header.ddpfPixelFormat.dwFlags == DDSHeader::DDPF_RGBA &&
		header.ddpfPixelFormat.dwRGBBitCount == 32)
	{
		m_format = GfxFormat_RGBA8_Unorm;
	}

	// read pixels

	m_width = header.dwWidth;
	m_height = header.dwHeight;

	u32 size_in_bits = m_width * m_height * header.ddpfPixelFormat.dwRGBBitCount;
	u32 size_in_bytes = size_in_bits / 8;

	m_data = new u8[size_in_bytes];
	m_data_size_in_bytes = size_in_bytes;

	stream.read(m_data, size_in_bytes);

	return true;
}

bool Image::load_dds(const char* filename)
{
	FileIn fi(filename);
	if (fi.valid())
	{
		return read_dds(fi);
	}
	else
	{
		Log::error("File not found: '%s'", filename);
		return false;
	}
}

bool Image::save_tga(const char* filename)
{
	if (m_format != GfxFormat_RGBA8_Unorm)
	{
		SDL_LogError(SDL_LOG_CATEGORY_RENDER, "Saving to TGA only supported for 32bpp RGBA format.");
		return false;
	}

	TGAHeader hdr;

	hdr.identsize = 0;              // size of ID field that follows 18 byte header (0 usually)
	hdr.colormaptype = 0;          // type of color map 0=none, 1=has palette
	hdr.imagetype = 2;              // type of image 0=none,1=indexed,2=rgb,3=grey,+8=rle packed

	hdr.colormapstart = 0;         // first color map entry in palette
	hdr.colormaplength = 0;        // number of colors in palette
	hdr.colormapbits = 0;          // number of bits per palette entry 15,16,24,32

	hdr.xstart = 0;                 // image x origin
	hdr.ystart = 0;                 // image y origin
	hdr.width = (short)m_width;     // image width in pixels
	hdr.height = (short)m_height;   // image height in pixels
	hdr.bits = 32;                  // image bits per pixel 8,16,24,32
	hdr.descriptor = 1 << 5;          // image descriptor bits (vh flip bits)        

	FileOut fo(filename);
	if (fo.valid() == false)
	{
		Log::error("Failed to write file '%s'", filename);
		return false;
	}

	u32 bw = 0;

	bw += fo.write_t(hdr);
	bw += fo.write(m_data, m_data_size_in_bytes);

	assert(bw == sizeof(hdr) + m_data_size_in_bytes);

	return true;
}
#endif