#include "PrimitiveBatch.h"
#include "Graphics.h"

const char* PrimitiveBatch::getSourceVs2d() const
{
	return
		"JAM_IN vec3 a_pos0;\n"
		"JAM_IN vec2 a_tex0;\n"
		"JAM_IN vec4 a_col0;\n"
		"uniform vec4 g_transform;\n"
		"JAM_OUT vec4 v_col0;\n"
		"JAM_OUT vec2 v_tex0;\n"
		"void main()\n"
		"{\n"
		"	gl_Position.xy = a_pos0.xy * g_transform.xy + g_transform.zw;\n"
		"	gl_Position.zw = vec2(0, 1);\n"
		"	v_col0 = a_col0;\n"
		"	v_tex0 = a_tex0;\n"
		"}\n";
}

const char* PrimitiveBatch::getSourceVs3d() const
{
	return
		"JAM_IN vec3 a_pos0;\n"
		"JAM_IN vec2 a_tex0;\n"
		"JAM_IN vec4 a_col0;\n"
		"uniform mat4 g_mat_viewproj;\n"
		"JAM_OUT vec4 v_col0;\n"
		"JAM_OUT vec2 v_tex0;\n"
		"void main()\n"
		"{\n"
		"	gl_Position = vec4(a_pos0, 1) * g_mat_viewproj;\n"
		"	v_col0 = a_col0;\n"
		"	v_tex0 = a_tex0;\n"
		"}\n";
}

const char* PrimitiveBatch::getSourcePsPlain() const
{
	return
		"JAM_IN vec4 v_col0;\n"
		"void main()\n"
		"{\n"
		"	JAM_FRAGCOLOR = v_col0;\n"
		"}\n";
}

const char* PrimitiveBatch::getSourcePsTextured() const
{
	return
		"JAM_IN vec4 v_col0;\n"
		"JAM_IN vec2 v_tex0;\n"
		"uniform sampler2D s_diffuse;\n"
		"void main()\n"
		"{\n"
		"	JAM_FRAGCOLOR = JAM_TEX2D(s_diffuse, v_tex0) * v_col0;\n"
		"}\n";
}

__forceinline line2 Line2(vec2 a, vec2 b)
{
	line2 r = { a, b };
	return r;
};

__forceinline line3 Line3(vec3 a, vec3 b)
{
	line3 r = { a, b };
	return r;
};

class BatchVertexFormat : public VertexFormatDescr
{
public:
	BatchVertexFormat()
	{
		add(0, DataType::Float3, UsageType::Position, 0);
		add(0, DataType::Float2, UsageType::Texcoord, 0);
		add(0, DataType::Color, UsageType::Color, 0);
	}
};

PrimitiveBatch::PrimitiveBatch(GfxDevice* device, u32 max_batch_vertices)
	: m_device(device)
	, m_context(Gfx_Immediate_Context(device))
	, m_max_batch_vertices(max_batch_vertices)
	, m_mode(BatchMode_Invalid)
	, m_curr_prim(PrimitiveType::TriangleList)
	, m_line_width(1.0f)
{
	SDL_assert(max_batch_vertices % 6 == 0);

	Gfx_AddReference(device);

	m_vsh_handle_2d = Gfx_VertexShader_Create(m_device, ShaderCodeText(getSourceVs2d()));
	m_vsh_handle_3d = Gfx_VertexShader_Create(m_device, ShaderCodeText(getSourceVs3d()));
	m_psh_handle_plain = Gfx_PixelShader_Create(m_device, ShaderCodeText(getSourcePsPlain()));
	m_psh_handle_textured = Gfx_PixelShader_Create(m_device, ShaderCodeText(getSourcePsTextured()));

	BatchVertexFormat fmt_desc;

	m_vertex_format_2d = Gfx_VertexFormat_Create(m_device, fmt_desc);
	m_vertex_format_3d = Gfx_VertexFormat_Create(m_device, fmt_desc);

	ShaderConstantBindings bindings;
	bindings.add_constant("g_transform", &m_vsc_transform_2d);
	bindings.add_constant("g_mat_viewproj", &m_vsc_mat_viewproj);
	bindings.add_sampler("s_diffuse", 0);

	m_techniques[TechniqueID_Plain_2D] = Gfx_Technique_Create(m_device, TechniqueDescr(m_psh_handle_plain, m_vsh_handle_2d, m_vertex_format_2d, &bindings));
	m_techniques[TechniqueID_Plain_3D] = Gfx_Technique_Create(m_device, TechniqueDescr(m_psh_handle_plain, m_vsh_handle_3d, m_vertex_format_3d, &bindings));
	m_techniques[TechniqueID_Textured_2D] = Gfx_Technique_Create(m_device, TechniqueDescr(m_psh_handle_textured, m_vsh_handle_2d, m_vertex_format_2d, &bindings));
	m_techniques[TechniqueID_Textured_3D] = Gfx_Technique_Create(m_device, TechniqueDescr(m_psh_handle_textured, m_vsh_handle_3d, m_vertex_format_3d, &bindings));

	BufferDescr vb_descr(BufferType::Vertex, BufferMode::WriteOften, GfxFormat_Unknown, m_max_batch_vertices, sizeof(BatchVertex));

	m_vertex_buffer = Gfx_Buffer_Create(m_device, vb_descr);

	BatchVertex* vertex_data = new BatchVertex[m_max_batch_vertices];
	m_vertices.setup(vertex_data, m_max_batch_vertices);

	Gfx_Buffer_Update(m_device, m_vertex_buffer, vertex_data, 0, sizeof(BatchVertex)*m_max_batch_vertices, true);

	m_sampler_linear = Gfx_SamplerState_Create(m_device, SamplerStateDescr::make_linear());
	m_sampler_point = Gfx_SamplerState_Create(m_device, SamplerStateDescr::make_point());
}

PrimitiveBatch::~PrimitiveBatch()
{
	delete[] m_vertices.data();

	for (u32 i = 0; i < TechniqueID_COUNT; ++i)
	{
		Gfx_Technique_Destroy(m_device, m_techniques[i]);
	}

	Gfx_SamplerState_Destroy(m_device, m_sampler_linear);
	Gfx_SamplerState_Destroy(m_device, m_sampler_point);

	Gfx_Buffer_Destroy(m_device, m_vertex_buffer);

	Gfx_VertexShader_Destroy(m_device, m_vsh_handle_2d);
	Gfx_VertexShader_Destroy(m_device, m_vsh_handle_3d);
	Gfx_PixelShader_Destroy(m_device, m_psh_handle_plain);
	Gfx_PixelShader_Destroy(m_device, m_psh_handle_textured);

	Gfx_RemoveReference(m_device);
}

TechniqueHandle PrimitiveBatch::getNextTechnique()
{
	if (m_curr_texture.valid())
	{
		switch (m_mode)
		{
		case BatchMode_2D:
			return m_techniques[TechniqueID_Textured_2D];
		case BatchMode_3D:
			return m_techniques[TechniqueID_Textured_3D];
		default:
			return InvalidResourceHandle();
		}
	}
	else
	{
		switch (m_mode)
		{
		case BatchMode_2D:
			return m_techniques[TechniqueID_Plain_2D];
		case BatchMode_3D:
			return m_techniques[TechniqueID_Plain_3D];
		default:
			return InvalidResourceHandle();
		}
	}
}

void PrimitiveBatch::flush()
{
	SDL_assert(m_mode != BatchMode_Invalid);

	if (m_vertices.size() == 0) return;

	TechniqueHandle next_technique = getNextTechnique();

	Gfx_Buffer_Update(m_device, m_vertex_buffer, m_vertices.data(), 0, (u32)m_vertices.size_in_bytes(), true);
	Gfx_SetPixelTexture(m_context, 0, m_curr_texture);
	Gfx_SetPixelSamplerState(m_context, 0, m_curr_sampler);
	Gfx_SetTechnique(m_context, next_technique);
	Gfx_SetVertexStream(m_context, 0, m_vertex_buffer);
	
	Gfx_SetPrimitive(m_context, m_curr_prim);
	Gfx_SetLineWidth(m_context, m_line_width);
	Gfx_Draw(m_context, 0, (u32)m_vertices.size());

	m_vertices.clear();
}

void PrimitiveBatch::drawBox(const aabb3& box, const ColorRGBA8& color)
{
	vec3 v[8];

	vec3 center = (box.min + box.max) * 0.5f;
	vec3 dimensions = box.max - box.min;

	v[0] = center + dimensions * vec3(-0.5f, -0.5f, -0.5f);
	v[1] = center + dimensions * vec3(+0.5f, -0.5f, -0.5f);
	v[2] = center + dimensions * vec3(-0.5f, +0.5f, -0.5f);
	v[3] = center + dimensions * vec3(+0.5f, +0.5f, -0.5f);
	v[4] = center + dimensions * vec3(-0.5f, -0.5f, +0.5f);
	v[5] = center + dimensions * vec3(+0.5f, -0.5f, +0.5f);
	v[6] = center + dimensions * vec3(-0.5f, +0.5f, +0.5f);
	v[7] = center + dimensions * vec3(+0.5f, +0.5f, +0.5f);

	drawLine(Line3(v[0], v[1]), color);
	drawLine(Line3(v[2], v[3]), color);
	drawLine(Line3(v[4], v[5]), color);
	drawLine(Line3(v[6], v[7]), color);

	drawLine(Line3(v[0], v[4]), color);
	drawLine(Line3(v[1], v[5]), color);
	drawLine(Line3(v[2], v[6]), color);
	drawLine(Line3(v[3], v[7]), color);

	drawLine(Line3(v[0], v[2]), color);
	drawLine(Line3(v[1], v[3]), color);
	drawLine(Line3(v[4], v[6]), color);
	drawLine(Line3(v[5], v[7]), color);
}

void PrimitiveBatch::drawPolygon(mat2 transform, const glm::vec2* vertices, int vertexCount, const ColorRGBA& color)
{
	std::vector<SDL_Point> lines;

	for (int i = 0; i < vertexCount; i++)
	{
		SDL_Point p;

		glm::vec2 vertex = vertices[i];
		vertex = transform * vertex;

		const float worldScale = 1.0f;
		const glm::vec2 offset(0.0f, 0.0f);

		p.x = (int)((vertex.x * worldScale) + offset.x);
		p.y = (int)((vertex.y * worldScale) + offset.y);

		lines.push_back(p);
	}

	setTexture(InvalidResourceHandle());
	PrimitiveBatch::BatchVertex* v = drawVertices(PrimitiveType::LineLoop, (u32)lines.size());

	ColorRGBA8 gfxColor((Uint8)(color.r * 255), (Uint8)(color.g * 255), (Uint8)(color.b * 255));
	int idx = 0;
	for (SDL_Point p : lines)
	{
		v[idx].pos.x = (float)p.x;
		v[idx].pos.y = (float)p.y;
		v[idx].pos.z = 0;
		v[idx].tex.x = 0;
		v[idx].tex.y = 0;
		v[idx].col0 = gfxColor;

		idx++;
	}
}

void PrimitiveBatch::drawSolidPolygon(mat2 transform, const glm::vec2* vertices, int vertexCount, const ColorRGBA& color)
{
	drawPolygon(transform, vertices, vertexCount, color);
}

void PrimitiveBatch::drawCircle(mat2 transform, const glm::vec2& center, float radius, const ColorRGBA& color)
{
	const int num = 32;
	glm::vec2 a = center + glm::vec2(0.0f, radius);
	for (int i = 1; i <= num; ++i)
	{
		float t = 2.0f* M_PI * float(i) / num;
		glm::vec2 b = center + glm::vec2(radius * sin(t), radius * cos(t));
		drawSegment(transform, a, b, color);
		a = b;
	}
}

void PrimitiveBatch::drawSolidCircle(mat2 transform, const glm::vec2& center, float radius, const glm::vec2& axis, const ColorRGBA& color)
{
	drawCircle(transform, center, radius, color);
	drawSegment(transform, center, center + radius*axis, color);
}

void PrimitiveBatch::drawSegment(mat2 transform, const glm::vec2& p1, const glm::vec2& p2, const ColorRGBA& color)
{
	glm::vec2 v[] = { p1, p2 };
	drawPolygon(transform, v, 2, color);
}

void PrimitiveBatch::drawTriangle(const vec2& a, const vec2& b, const vec2& c, const ColorRGBA8& color)
{
	setTexture(InvalidResourceHandle());

	BatchVertex* v = drawVertices(PrimitiveType::TriangleList, 3);

	v[0].pos.x = a.x;
	v[0].pos.y = a.y;
	v[0].pos.z = 0;
	v[0].tex.x = 0;
	v[0].tex.y = 0;
	v[0].col0 = color;

	v[1].pos.x = b.x;
	v[1].pos.y = b.y;
	v[1].pos.z = 0;
	v[1].tex.x = 0;
	v[1].tex.y = 0;
	v[1].col0 = color;

	v[2].pos.x = c.x;
	v[2].pos.y = c.y;
	v[2].pos.z = 0;
	v[2].tex.x = 0;
	v[2].tex.y = 0;
	v[2].col0 = color;
}

void PrimitiveBatch::drawTriangle(const vec3& a, const vec3& b, const vec3& c, const ColorRGBA8& color)
{
	setTexture(InvalidResourceHandle());

	BatchVertex* v = drawVertices(PrimitiveType::TriangleList, 3);

	v[0].pos.x = a.x;
	v[0].pos.y = a.y;
	v[0].pos.z = a.z;
	v[0].tex.x = 0;
	v[0].tex.y = 0;
	v[0].col0 = color;

	v[1].pos.x = b.x;
	v[1].pos.y = b.y;
	v[1].pos.z = b.z;
	v[1].tex.x = 0;
	v[1].tex.y = 0;
	v[1].col0 = color;

	v[2].pos.x = c.x;
	v[2].pos.y = c.y;
	v[2].pos.z = c.z;
	v[2].tex.x = 0;
	v[2].tex.y = 0;
	v[2].col0 = color;
}

void PrimitiveBatch::drawTriangle(const vec3& pa, const vec3& pb, const vec3& pc, const ColorRGBA8& ca, const ColorRGBA8& cb, const ColorRGBA8& cc)
{
	setTexture(InvalidResourceHandle());

	BatchVertex* v = drawVertices(PrimitiveType::TriangleList, 3);

	v[0].pos.x = pa.x;
	v[0].pos.y = pa.y;
	v[0].pos.z = pa.z;
	v[0].tex.x = 0;
	v[0].tex.y = 0;
	v[0].col0 = ca;

	v[1].pos.x = pb.x;
	v[1].pos.y = pb.y;
	v[1].pos.z = pb.z;
	v[1].tex.x = 0;
	v[1].tex.y = 0;
	v[1].col0 = cb;

	v[2].pos.x = pc.x;
	v[2].pos.y = pc.y;
	v[2].pos.z = pc.z;
	v[2].tex.x = 0;
	v[2].tex.y = 0;
	v[2].col0 = cc;
}

void PrimitiveBatch::begin2d(const aabb2& bounds, bool flipped)
{
	SDL_assert(m_mode == BatchMode_Invalid);

	float width = bounds.width();
	float height = bounds.height();

	vec2 offset = -2.0f * bounds.min / bounds.dimensions();

	if (flipped){
		m_vsc_transform_2d = vec4(2.0f / width, 2.0f / height,  -1.0f+offset.x, -1.0f+offset.y);
	}
	else{
		m_vsc_transform_2d = vec4(2.0f / width, -2.0f / height, -1.0f+offset.x, 1.0f-offset.y);
	}

	m_mode = BatchMode_2D;
}

void PrimitiveBatch::begin2d(u32 width, u32 height)
{
	aabb2 bounds = { vec2(0.0f, 0.0f), vec2(float(width), float(height)) };
	begin2d(bounds);
}

void PrimitiveBatch::begin3d(const mat4x4& view_proj)
{
	SDL_assert(m_mode == BatchMode_Invalid);
	m_vsc_mat_viewproj = glm::transpose(view_proj);
	m_mode = BatchMode_3D;
}

void PrimitiveBatch::end2d()
{
	SDL_assert(m_mode == BatchMode_2D);
	flush();
	m_mode = BatchMode_Invalid;
}

void PrimitiveBatch::end3d()
{
	SDL_assert(m_mode == BatchMode_3D);
	flush();
	m_mode = BatchMode_Invalid;
}

void PrimitiveBatch::drawTexturedQuad(const aabb2& rect, ColorRGBA8 color)
{
	drawTexturedQuad(rect.bl(), rect.br(), rect.tr(), rect.tl(), color);
}

void PrimitiveBatch::drawTexturedQuad(const TexturedQuad2D* q, ColorRGBA8 color)
{
	drawTexturedQuad(q->pos[0], q->pos[1], q->pos[2], q->pos[3], color, q->tex[0], q->tex[1], q->tex[2], q->tex[3]);
}

void PrimitiveBatch::drawTexturedQuad(const TexturedQuad3D* q, ColorRGBA8 color)
{
	drawTexturedQuad(q->pos[0], q->pos[1], q->pos[2], q->pos[3], color, q->tex[0], q->tex[1], q->tex[2], q->tex[3]);
}

void PrimitiveBatch::drawTexturedQuad(
	const vec2& a, const vec2& b,
	const vec2& c, const vec2& d,
	ColorRGBA8 color,
	const vec2& ta, const vec2& tb,
	const vec2& tc, const vec2& td)
{
	BatchVertex* v = drawVertices(PrimitiveType::TriangleList, 6);

	size_t i = 0;

	v[i].pos.x = a.x;
	v[i].pos.y = a.y;
	v[i].pos.z = 0;
	v[i].tex = ta;
	v[i].col0 = color;
	++i;

	v[i].pos.x = b.x;
	v[i].pos.y = b.y;
	v[i].pos.z = 0;
	v[i].tex = tb;
	v[i].col0 = color;
	++i;

	v[i].pos.x = c.x;
	v[i].pos.y = c.y;
	v[i].pos.z = 0;
	v[i].tex = tc;
	v[i].col0 = color;
	++i;

	v[i].pos.x = a.x;
	v[i].pos.y = a.y;
	v[i].pos.z = 0;
	v[i].tex = ta;
	v[i].col0 = color;
	++i;

	v[i].pos.x = c.x;
	v[i].pos.y = c.y;
	v[i].pos.z = 0;
	v[i].tex = tc;
	v[i].col0 = color;
	++i;

	v[i].pos.x = d.x;
	v[i].pos.y = d.y;
	v[i].pos.z = 0;
	v[i].tex = td;
	v[i].col0 = color;
	++i;
}

void PrimitiveBatch::drawTexturedQuad(
	const vec3& pa, const vec3& pb,
	const vec3& pc, const vec3& pd,
	ColorRGBA8 color,
	const vec2& ta, const vec2& tb,
	const vec2& tc, const vec2& td)
{
	BatchVertex* v = drawVertices(PrimitiveType::TriangleList, 6);

	size_t i = 0;

	v[i].pos = pa;
	v[i].tex = ta;
	v[i].col0 = color;
	++i;

	v[i].pos = pb;
	v[i].tex = tb;
	v[i].col0 = color;
	++i;

	v[i].pos = pc;
	v[i].tex = tc;
	v[i].col0 = color;
	++i;

	v[i].pos = pa;
	v[i].tex = ta;
	v[i].col0 = color;
	++i;

	v[i].pos = pc;
	v[i].tex = tc;
	v[i].col0 = color;
	++i;

	v[i].pos = pd;
	v[i].tex = td;
	v[i].col0 = color;
	++i;
}

void PrimitiveBatch::drawRect(const aabb2& rect, const ColorRGBA8& color)
{
	setTexture(InvalidResourceHandle());

	BatchVertex* v = drawVertices(PrimitiveType::TriangleList, 6);

	size_t i = 0;

	v[i].pos.x = rect.tl().x;
	v[i].pos.y = rect.tl().y;
	v[i].pos.z = 0;
	v[i].tex.x = 0;
	v[i].tex.y = 0;
	v[i].col0 = color;
	++i;

	v[i].pos.x = rect.bl().x;
	v[i].pos.y = rect.bl().y;
	v[i].pos.z = 0;
	v[i].tex.x = 0;
	v[i].tex.y = 0;
	v[i].col0 = color;
	++i;

	v[i].pos.x = rect.br().x;
	v[i].pos.y = rect.br().y;
	v[i].pos.z = 0;
	v[i].tex.x = 0;
	v[i].tex.y = 0;
	v[i].col0 = color;
	++i;

	v[i].pos.x = rect.tl().x;
	v[i].pos.y = rect.tl().y;
	v[i].pos.z = 0;
	v[i].tex.x = 0;
	v[i].tex.y = 0;
	v[i].col0 = color;
	++i;

	v[i].pos.x = rect.br().x;
	v[i].pos.y = rect.br().y;
	v[i].pos.z = 0;
	v[i].tex.x = 0;
	v[i].tex.y = 0;
	v[i].col0 = color;
	++i;

	v[i].pos.x = rect.tr().x;
	v[i].pos.y = rect.tr().y;
	v[i].pos.z = 0;
	v[i].tex.x = 0;
	v[i].tex.y = 0;
	v[i].col0 = color;
	++i;
}

void PrimitiveBatch::setLineWidth(float width)
{
	if(m_line_width != width)
		flush();

	m_line_width = width;
}

void PrimitiveBatch::drawLine(const vec2& a, const vec2& b, const ColorRGBA8& color)
{
	drawLine(Line2(a, b), color);
}

void PrimitiveBatch::drawLine(const vec3& a, const vec3& b, const ColorRGBA8& color)
{
	drawLine(Line3(a, b), color);
}

void PrimitiveBatch::drawLine(const line3& line, const ColorRGBA8& color)
{
	setTexture(InvalidResourceHandle());

	BatchVertex* v = drawVertices(PrimitiveType::LineList, 2);

	v[0].pos.x = line.a.x;
	v[0].pos.y = line.a.y;
	v[0].pos.z = line.a.z;
	v[0].tex.x = 0;
	v[0].tex.y = 0;
	v[0].col0 = color;

	v[1].pos.x = line.b.x;
	v[1].pos.y = line.b.y;
	v[1].pos.z = line.b.z;
	v[1].tex.x = 0;
	v[1].tex.y = 0;
	v[1].col0 = color;
}

void PrimitiveBatch::drawLine(const line3& line, const ColorRGBA8& color_start, const ColorRGBA8& color_end)
{
	setTexture(InvalidResourceHandle());

	BatchVertex* v = drawVertices(PrimitiveType::LineList, 2);

	v[0].pos.x = line.a.x;
	v[0].pos.y = line.a.y;
	v[0].pos.z = line.a.z;
	v[0].tex.x = 0;
	v[0].tex.y = 0;
	v[0].col0 = color_start;

	v[1].pos.x = line.b.x;
	v[1].pos.y = line.b.y;
	v[1].pos.z = line.b.z;
	v[1].tex.x = 0;
	v[1].tex.y = 0;
	v[1].col0 = color_end;
}

void PrimitiveBatch::drawLine(const line2& line, const ColorRGBA8& color_start, const ColorRGBA8& color_end)
{
	setTexture(InvalidResourceHandle());

	BatchVertex* v = drawVertices(PrimitiveType::LineList, 2);

	v[0].pos.x = line.a.x;
	v[0].pos.y = line.a.y;
	v[0].pos.z = 0;
	v[0].tex.x = 0;
	v[0].tex.y = 0;
	v[0].col0 = color_start;

	v[1].pos.x = line.b.x;
	v[1].pos.y = line.b.y;
	v[1].pos.z = 0;
	v[1].tex.x = 0;
	v[1].tex.y = 0;
	v[1].col0 = color_end;
}

void PrimitiveBatch::drawLine(const line2& line, const ColorRGBA8& color)
{
	setTexture(InvalidResourceHandle());

	BatchVertex* v = drawVertices(PrimitiveType::LineList, 2);

	v[0].pos.x = line.a.x;
	v[0].pos.y = line.a.y;
	v[0].pos.z = 0;
	v[0].tex.x = 0;
	v[0].tex.y = 0;
	v[0].col0 = color;

	v[1].pos.x = line.b.x;
	v[1].pos.y = line.b.y;
	v[1].pos.z = 0;
	v[1].tex.x = 0;
	v[1].tex.y = 0;
	v[1].col0 = color;
}

void PrimitiveBatch::drawCross(const vec3& pos, float size, const ColorRGBA8& color)
{
	drawLine(Line3(pos + vec3(-0.5f, 0, 0)*size, pos + vec3(+0.5f, 0, 0)*size), color);
	drawLine(Line3(pos + vec3(0, -0.5f, 0)*size, pos + vec3(0, +0.5f, 0)*size), color);
	drawLine(Line3(pos + vec3(0, 0, -0.5f)*size, pos + vec3(0, 0, +0.5f)*size), color);
}

void PrimitiveBatch::drawCross(const vec2& pos, float size, const ColorRGBA8& color)
{
	drawLine(Line2(pos + vec2(-0.5f, 0)*size, pos + vec2(+0.5f, 0)*size), color);
	drawLine(Line2(pos + vec2(0, -0.5f)*size, pos + vec2(0, +0.5f)*size), color);
}

void PrimitiveBatch::drawDiagonalCross(const vec2& pos, float size, const ColorRGBA8& color)
{
	drawLine(Line2(pos + vec2(-0.5f, -0.5f)*size, pos + vec2(+0.5f, +0.5f)*size), color);
	drawLine(Line2(pos + vec2(-0.5f, +0.5f)*size, pos + vec2(+0.5f, -0.5f)*size), color);
}

void PrimitiveBatch::setTexture(TextureHandle tex)
{
	if (m_curr_texture != tex)
	{
		flush();
		m_curr_texture = tex;
	}
}

void PrimitiveBatch::setSampler(SamplerStateHandle smp)
{
	if (m_curr_sampler != smp)
	{
		flush();
		m_curr_sampler = smp;
	}
}

void PrimitiveBatch::setSampler(SamplerState smp)
{
	setSampler(smp == SamplerState_Linear ? m_sampler_linear : m_sampler_point);
}

void PrimitiveBatch::setPrimitive(PrimitiveType prim)
{
	SDL_assert(prim == PrimitiveType::LineLoop || prim == PrimitiveType::LineList || prim == PrimitiveType::TriangleList);

	// need to break batch after line loop too 
	if (m_curr_prim != prim || prim == PrimitiveType::LineLoop)
	{
		flush();
		m_curr_prim = prim;
	}
}

PrimitiveBatch::BatchVertex* PrimitiveBatch::drawVertices(PrimitiveType prim_type, u32 num_vertices)
{
	setPrimitive(prim_type);

	SDL_assert(m_mode != BatchMode_Invalid);
	SDL_assert(num_vertices <= m_max_batch_vertices);

	if (m_vertices.size() + num_vertices >= m_max_batch_vertices)
	{
		flush();
	}

	return m_vertices.alloc_unsafe(num_vertices);
}

