#pragma once

#include "Framework.h"
#include "Color.h"
#include "StackAllocator.h"
#include "Graphics.h"

#include <glm/mat2x2.hpp>
using glm::mat2;

struct TexturedQuad2D
{
	vec2 pos[4];
	vec2 tex[4];
};

struct TexturedQuad3D
{
	vec3 pos[4];
	vec2 tex[4];
};

class PrimitiveBatch
{
public:

	enum SamplerState
	{
		SamplerState_Point,
		SamplerState_Linear,

		SamplerState_COUNT
	};

	enum BatchMode
	{
		BatchMode_Invalid,
		BatchMode_2D,
		BatchMode_3D,

		BatchMode_COUNT
	};

	struct BatchVertex
	{
		vec3		pos;
		vec2		tex;
		ColorRGBA8	col0;
	};

public:

	PrimitiveBatch(GfxDevice* device, u32 max_batch_vertices = 12000);
	~PrimitiveBatch();

	vec4 getTransform2d() const { return m_vsc_transform_2d; }
	const mat4x4& getMatViewProj() const { return m_vsc_mat_viewproj; }

	void begin2d(u32 width, u32 height);
	void begin2d(const aabb2& bounds, bool flipped = true);
	void end2d();

	void begin3d(const mat4x4& view_proj);
	void end3d();

	void setTexture(TextureHandle tex);
	void setSampler(SamplerStateHandle smp);
	void setSampler(SamplerState smp);
	void setPrimitive(PrimitiveType prim);

	void setLineWidth(float width);

	void drawLine(const vec2& a, const vec2& b, const ColorRGBA8& color = ColorRGBA8::White());
	void drawLine(const vec3& a, const vec3& b, const ColorRGBA8& color = ColorRGBA8::White());
	void drawLine(const line2& line, const ColorRGBA8& color = ColorRGBA8::White());
	void drawLine(const line2& line, const ColorRGBA8& color_start, const ColorRGBA8& color_end);
	void drawLine(const line3& line, const ColorRGBA8& color = ColorRGBA8::White());
	void drawLine(const line3& line, const ColorRGBA8& color_start, const ColorRGBA8& color_end);
	void drawRect(const aabb2& rect, const ColorRGBA8& color = ColorRGBA8::White());

	void drawTriangle(const vec2& a, const vec2& b, const vec2& c, const ColorRGBA8& color);
	void drawTriangle(const vec3& a, const vec3& b, const vec3& c, const ColorRGBA8& color);
	void drawTriangle(const vec3& pa, const vec3& pb, const vec3& pc, const ColorRGBA8& ca, const ColorRGBA8& cb, const ColorRGBA8& cc);

	void drawTexturedQuad(const aabb2& rect, ColorRGBA8 color = ColorRGBA8::White());
	void drawTexturedQuad(const TexturedQuad2D* q, ColorRGBA8 color = ColorRGBA8::White());
	void drawTexturedQuad(const TexturedQuad3D* q, ColorRGBA8 color = ColorRGBA8::White());

	void drawTexturedQuad(
		const vec2& a, const vec2& b,
		const vec2& c, const vec2& d,
		ColorRGBA8 color = ColorRGBA8::White(),
		const vec2& ta = vec2(0, 0), const vec2& tb = vec2(1, 0),
		const vec2& tc = vec2(1, 1), const vec2& td = vec2(0, 1));

	void drawTexturedQuad(
		const vec3& pa, const vec3& pb,
		const vec3& pc, const vec3& pd,
		ColorRGBA8 color = ColorRGBA8::White(),
		const vec2& ta = vec2(0, 0), const vec2& tb = vec2(1, 0),
		const vec2& tc = vec2(1, 1), const vec2& td = vec2(0, 1));

	void drawCross(const vec3& pos, float size, const ColorRGBA8& color = ColorRGBA8::White());
	void drawCross(const vec2& pos, float size, const ColorRGBA8& color = ColorRGBA8::White());
	void drawDiagonalCross(const vec2& pos, float size, const ColorRGBA8& color = ColorRGBA8::White());
	void drawBox(const aabb3& box, const ColorRGBA8& color = ColorRGBA8::White());

	void drawPolygon(mat2 transform, const glm::vec2* vertices, int vertexCount, const ColorRGBA& color);
	void drawSolidPolygon(mat2 transform, const glm::vec2* vertices, int vertexCount, const ColorRGBA& color);
	void drawCircle(mat2 transform, const glm::vec2& center, float radius, const ColorRGBA& color);
	void drawSolidCircle(mat2 transform, const glm::vec2& center, float radius, const glm::vec2& axis, const ColorRGBA& color);
	void drawSegment(mat2 transform, const glm::vec2& p1, const glm::vec2& p2, const ColorRGBA& color);

	BatchVertex* drawVertices(PrimitiveType prim_type, u32 num_vertices);

	void flush();

	inline u32 getMaxBatchVertices() const { return m_max_batch_vertices; }

private:

	TechniqueHandle getNextTechnique();

	const char* getSourceVs2d() const;
	const char* getSourceVs3d() const;
	const char* getSourcePsPlain() const;
	const char* getSourcePsTextured() const;

private:

	GfxDevice*					m_device;
	GfxContext*					m_context;

	u32							m_max_batch_vertices;
	BatchMode					m_mode;

	PrimitiveType				m_curr_prim;
	TextureHandle				m_curr_texture;
	SamplerStateHandle			m_curr_sampler;

	SamplerStateHandle			m_sampler_linear;
	SamplerStateHandle			m_sampler_point;

	float						m_line_width;

	BufferHandle				m_vertex_buffer;
	StackAllocator<BatchVertex> m_vertices;

	VertexFormatHandle			m_vertex_format_2d;
	VertexFormatHandle			m_vertex_format_3d;

	vec4						m_vsc_transform_2d;
	mat4x4						m_vsc_mat_viewproj;

	VertexShaderHandle			m_vsh_handle_2d;
	VertexShaderHandle			m_vsh_handle_3d;

	PixelShaderHandle			m_psh_handle_plain;
	PixelShaderHandle			m_psh_handle_textured;

	enum TechniqueID
	{
		TechniqueID_Plain_2D,
		TechniqueID_Plain_3D,
		TechniqueID_Textured_2D,
		TechniqueID_Textured_3D,
		TechniqueID_COUNT
	};

	TechniqueHandle m_techniques[TechniqueID_COUNT];
};
