#pragma once

#include <Framework/Math.h>

#include <cJSON.h>
#include <iostream>
#include <vector>
#include <typeinfo>

// Based on the public domain code by Jon Watte:
// http://www.enchantedage.com/cpp-reflection


struct TypeBase
{
	virtual cJSON* marshal(const void*) const = 0;
	virtual void demarshal(const cJSON*, void*) const = 0;
	virtual char const *name() const = 0;
	virtual size_t size() const = 0;
	virtual void assign(void* const a, const void* const b) const = 0;
	virtual bool equal(const void* const a, const void* const b) const = 0;
};

inline std::ostream& operator<<(std::ostream& str, const vec2& vec) {
	str << "vec2(" << vec.x << ", " << vec.y << ")";
	return str;
}

template <typename T>
inline cJSON* jsonMarshall(const T& src) = delete;
template <typename T>
inline void jsonDemarshall(const cJSON* json, T* dst) = delete;

template<typename T> struct Type : TypeBase
{
	static Type<T> instance;
	// custom marshaling is handled by template specialization
	cJSON* marshal(const void* src) const { return jsonMarshall<T>(*(const T*)src); }
	void demarshal(const cJSON* json, void* dst) const { jsonDemarshall<T>(json, (T*)dst); }
	char const *name() const { return typeid(T).name(); }
	size_t size() const { return sizeof(T); }
	void assign(void* const a, const void* const b) const { *(T*)a = *(T*)b; }
	bool equal(const void* const a, const void* const b) const { return *(T*)a == *(T*)b; }
};

template <typename T> T& instance() {
	static T t;
	return t;
}

template<typename T, typename Q>
TypeBase *get_type(Q T::*mem) {
	return &instance<Type<Q> >();
}

struct AttribBase {};

struct member_t
{
	char const *name;
	TypeBase *type;
	size_t offset;
	std::vector<AttribBase*> attribs;
};

struct ReflectionBase
{
	void ReflectionConstruct();
	virtual size_t size() const = 0;
	virtual char const *name() const = 0;
	virtual size_t memberCount() const = 0;
	virtual const member_t *members() const = 0;
	virtual const member_t* getMember(const char* const name) const = 0;
};

template<typename InfoFields>
struct TypeReflection : InfoFields {
	/* overrides used by ReflectionBase */
	size_t size() const { 
		return sizeof(InfoFields::_type);
	} 

	size_t memberCount() const {
		size_t cnt;
		InfoFields::get_members(cnt);
		return cnt;
	}

	const member_t* members() const {
		size_t cnt;
		return InfoFields::get_members(cnt);
	}

	const member_t* getMember(const char* const name) const {
		size_t cnt;
		const member_t* mems = InfoFields::get_members(cnt);

		for (size_t i = 0; i < cnt; ++i) {
			if (0 == strcmp(mems[i].name, name)) {
				return mems + i;
			}
		}

		return nullptr;
	}

	//_info() { ReflectionConstruct(); }
	static inline TypeReflection &info() {
		return instance<TypeReflection>();
	}
};

#define MEMBER(x, ...) \
   { #x, get_type(&_type::x), (size_t)&((_type*)0)->x, std::vector<AttribBase*>{ __VA_ARGS__ } },

#define RTTI(__type, _mems) \
  struct InfoFields: ReflectionBase { \
	typedef __type _type; \
	inline char const *name() const { return #__type; } \
    static inline member_t const *get_members(size_t &cnt) { \
      static member_t members[] = { _mems {} }; \
      cnt = sizeof(members) / sizeof(members[0]) - 1; \
      return members; \
    } \
  }; \
  typedef TypeReflection<InfoFields> _info; \
  inline static _info &info() { static _info _theInfo; return _theInfo; }

struct CustomAttrib : AttribBase {
	CustomAttrib(int, int) {}
};

struct IReflectionSupport {
	virtual const ReflectionBase& getReflection() const = 0;
};



/*struct Foo {
	int x, y, z;
	std::string str;

	RTTI(Foo,
		MEMBER(x, new CustomAttrib(1, 2))
		MEMBER(y)
		MEMBER(z)
		MEMBER(str)
	);
};*/




cJSON* marshallObjectFields(const void *const obj, const ReflectionBase& info);
cJSON* marshallObject(const void *const obj, const ReflectionBase& info);

cJSON* marshallObjectFieldsDiff(const void *const obj, const void *const baseObj, const ReflectionBase& info);
cJSON* marshallObjectDiff(const void *const obj, const void *const baseObj, const ReflectionBase& info);

template <typename T>
inline cJSON* marshallObject(const T& obj) {
	return marshallObject(&obj, obj.info());
}

template <typename T>
inline cJSON* marshallObjectFields(const T& obj) {
	return marshallObjectFields(&obj, obj.info());
}

void demarshallObjectFields(const cJSON*, void *const obj, const ReflectionBase& info);

template <typename T>
inline void demarshallObjectFields(const cJSON* json, T *const obj) {
	return demarshallObjectFields(json, obj, obj->info());
}
