#pragma once

#include "Framework.h"
#include "StaticArray.h"

#include <vector>

struct InvalidResourceHandle {};

struct UntypedResourceHandle
{
	UntypedResourceHandle(size_t _index) : index(_index) {}
	size_t index;
};

template <typename T>
class ResourceHandle
{
public:

	ResourceHandle() : m_index(0) {}
	ResourceHandle(InvalidResourceHandle) : m_index(0) {}
	explicit ResourceHandle(UntypedResourceHandle h) : m_index(h.index) {}

	bool valid() const { return m_index!=0; }
	size_t index() const { return m_index; }

	bool operator == (const ResourceHandle<T>& rhs) const { return m_index == rhs.m_index; }
	bool operator != (const ResourceHandle<T>& rhs) const { return m_index != rhs.m_index; }

private:

	ResourceHandle(size_t idx) : m_index(idx) {}

	size_t m_index;

};

template <typename T, typename HANDLE_TYPE, size_t CAPACITY>
class ResourcePool
{
public:

	typedef HANDLE_TYPE HandleType;

	ResourcePool()
	{
		T default_value;
		memset(&default_value, 0, sizeof(default_value));

		for(size_t i=CAPACITY; i!=0; --i)
		{
			empty.push_back(i-1);
		}

		push(default_value);
	}

	__forceinline HANDLE_TYPE push(const T& val)
	{
		SDL_assert(empty.size()!=0);
		size_t idx = empty.back();
		empty.pop_back();
		data[idx] = val;
		return HANDLE_TYPE(UntypedResourceHandle(idx));
	}

	__forceinline void remove(HANDLE_TYPE h)
	{
		if( h.valid() ) 
		{
			empty.push_back(h.index());
		}
	}

	__forceinline const T& operator [] (HANDLE_TYPE h) const
	{
		return data[h.index()];
	}

	__forceinline T& operator [] (HANDLE_TYPE h)
	{
		return data[h.index()];
	}

	__forceinline u32 capacity() const { return CAPACITY; }

private:

	T data[CAPACITY];
	StaticArray<size_t, CAPACITY> empty;

};

template <typename T, typename HANDLE_TYPE>
class DynamicResourcePool
{
public:

	typedef HANDLE_TYPE HandleType;

	DynamicResourcePool()
	{
		T default_value;
		memset(&default_value, 0, sizeof(default_value));
		push(default_value);
	}

	__forceinline HANDLE_TYPE push(const T& val)
	{
		size_t idx;
		if (empty.empty())
		{
			idx = data.size();
			data.push_back(val);
		}
		else
		{
			idx = empty.back();
			empty.pop_back();
			data[idx] = val;
		}
		return HANDLE_TYPE(UntypedResourceHandle(idx));
	}

	__forceinline void remove(HANDLE_TYPE h)
	{
		if (h.valid())
		{
			empty.push_back(h.index());
		}
	}

	__forceinline const T& operator [] (HANDLE_TYPE h) const
	{
		return data[h.index()];
	}

	__forceinline T& operator [] (HANDLE_TYPE h)
	{
		return data[h.index()];
	}

private:

	std::vector<T> data;
	std::vector<size_t> empty;

};
