#include "Sprite.h"
#include "PrimitiveBatch.h"

#include <SDL.h>
#include <Framework/FileWatcher.h>
#include <algorithm>
#include <stb_image.h>

uint PixelSize = 4;

void drawSprite(SDLSprite *s, SDL_Renderer *renderer, int idx, SDL_Rect *dstrect)
{
	int horizontal = s->tw / s->sw;
	int vertical = s->th / s->sh;
	int count = (vertical * horizontal) + 1;

	idx %= count; // loop animation

	int x = idx % count;
	int y = idx / count;

	SDL_Rect r = { x * s->sw, y * s->sh, s->sw, s->sh };
	SDL_RenderCopy(renderer, s->texture, &r, dstrect);
}

void print(SDLSprite *s, SDL_Renderer *renderer, const char *str, int x, int y, int k)
{
	const char *glyphs = "0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ";

	SDL_Rect r = { x, y, s->sw, s->sh };
	while (*str)
	{
		char *c = SDL_strchr(glyphs, *str);
		int idx = int(c - glyphs);

		if (c)
			drawSprite(s, renderer, idx, &r);

		r.x += s->sw + k;
		str++;
	}
}


std::vector<Sprite*> g_loadedSprites;

void loadSpriteData(Sprite *const sprite, const char *textureFile, const u32 colorKey, bool clip, bool repeat)
{
	int width, height, components;

	u8 *const srcPixels = stbi_load(textureFile, &width, &height, &components, 4);
	SDL_assert(srcPixels && "Can't load texture");

	if (srcPixels && width > 0 && height > 0)
	{
		if (colorKey)
		{
			const u32 numPixels = width * height;
			for (u32 i = 0; i < numPixels; ++i)
			{
				u8& r = srcPixels[i * 4 + 0];
				u8& g = srcPixels[i * 4 + 1];
				u8& b = srcPixels[i * 4 + 2];
				u8& a = srcPixels[i * 4 + 3];

				if (((u32(r) << 16u) + (u32(g) << 8u) + u32(b)) == colorKey) {
					a = 0u;
				}
			}
		}

		int minX = width;
		int maxX = 0;
		int minY = height;
		int maxY = 0;

		for (int y = 0; y < height; ++y) {
			for (int x = 0; x < width; ++x) {
				// NOTE: Flipped Y so that the minY will refer to coords after flipping the image.
				const u8 a = srcPixels[((height - y - 1) * width + x) * 4 + 3];
				if (a != 0) {
					minX = std::min(minX, x);
					maxX = std::max(maxX, x);
					minY = std::min(minY, y);
					maxY = std::max(maxY, y);
				}
			}
		}

		if (minX == height) {
			minX = 0;
			maxX = 0;
			minY = 0;
			maxY = 0;
		}

		if (!clip) {
			minX = 0;
			maxX = width - 1;
			minY = 0;
			maxY = height - 1;
		}

		sprite->pixels.resize((maxX - minX + 1) * (maxY - minY + 1));
		sprite->offset = glm::ivec2(minX, minY);

		sprite->w = (maxX - minX + 1);
		sprite->h = (maxY - minY + 1);
		sprite->origSize = glm::ivec2(width, height);

		// Flip the image and copy the final pixels
		for (int y = 0; y < sprite->h; ++y) {
			memcpy(
				sprite->pixels.data() + y * sprite->w,
				srcPixels + ((height - y - 1 - sprite->offset.y) * width + sprite->offset.x) * 4,
				sprite->w * 4
			);
		}

        TextureDescr texDescr = TextureDescr::create_2D(sprite->w, sprite->h, GfxFormat_RGBA8_Unorm, false, false, repeat);
		TextureHandle tex = Gfx_Texture_Create(g_device, texDescr, sprite->pixels.data(), sprite->w * sprite->h * 4);

		sprite->texture = tex;
	}

	stbi_image_free(srcPixels);
}

Sprite* loadSprite(const char *filename, const u32 colorKey, bool clip, bool repeat, const std::function<void(Sprite* const)>& onReload)
{
	const char *textureFile = jamTranslatePath(filename);

	Sprite *const sprite = new Sprite;

	loadSpriteData(sprite, textureFile, colorKey, clip, repeat);
	onReload(sprite);

	g_loadedSprites.push_back(sprite);

	FileWatcher::watchFile(textureFile, [sprite, textureFile, colorKey, clip, repeat, onReload] {
		loadSpriteData(sprite, textureFile, colorKey, clip, repeat);
		onReload(sprite);
	});

	return sprite;
}

Sprite* loadSprite(const char *filename, const u32 colorKey, bool clip, bool repeat)
{
	return loadSprite(filename, colorKey, clip, repeat, [](Sprite *const) {});
}

Sprite* cloneSprite(Sprite *const s)
{
	Sprite* newSprite = new Sprite;
	newSprite->h = s->h;
	newSprite->w = s->w;
	newSprite->offset = s->offset;
	newSprite->origSize = s->origSize;
	newSprite->pixels = s->pixels;

	TextureDescr texDescr = TextureDescr::create_2D(newSprite->w, newSprite->h, GfxFormat_RGBA8_Unorm, false, false, false /* lets hope we don't need this*/);
	newSprite->texture = Gfx_Texture_Create(g_device, texDescr, newSprite->pixels.data(), newSprite->w * newSprite->h * 4);

	return newSprite;
}

u8 linearRgbToSrgbByte(float x) {
	x = std::min(std::max(0.0f, x), 1.0f);
	x = sqrt(x);
	x *= 255;
	return u8(x);
}

void drawSprite(PrimitiveBatch *const prim, const Sprite& sprite, const vec2& pos, float angle) {
	
	vec2 dstMin;
	aabb2 dstRect = { dstMin, dstMin + vec2(sprite.w * PixelSize, sprite.h * PixelSize) };
	
	auto rotate = [angle](vec2 k) {
		float sn = sin(angle);
		float cs = cos(angle);
		return vec2(
			k.x * cs - k.y * sn,
			k.x * sn + k.y * cs
		);
	};

	prim->setTexture(sprite.texture);
	prim->drawTexturedQuad(
		(vec2(pos.x, pos.y) + vec2(sprite.offset) * float(PixelSize)) + rotate(dstRect.bl()), 
		(vec2(pos.x, pos.y) + vec2(sprite.offset) * float(PixelSize)) + rotate(dstRect.br()), 
		(vec2(pos.x, pos.y) + vec2(sprite.offset) * float(PixelSize)) + rotate(dstRect.tr()), 
		(vec2(pos.x, pos.y) + vec2(sprite.offset) * float(PixelSize)) + rotate(dstRect.tl()));
}

