#pragma once

#include "Framework.h"

class Stream
{

public:

	virtual ~Stream() {}

	virtual u32		write(const void* buf, u32 size) = 0;
	template <typename T>
	inline u32		write(const T& val) { return write(&val, sizeof(val)); }

	virtual u32		read(void* buf, u32 size) = 0;
	template <typename T>
	inline u32		read(T& val) { return read(&val, sizeof(val)); }

	virtual u32		tell() = 0;
	virtual void	seek(u32 pos) = 0;
	virtual void	skip(s32 distance) = 0;
	virtual void	rewind() = 0;

	virtual bool	valid() = 0;

	virtual u32		length() = 0;

};

class MemStream : public Stream
{
public:

	MemStream(const void* data, u32 size);
	MemStream(void* data, u32 size);

	virtual u32		write(const void* buf, u32 size);
	virtual u32		read(void* buf, u32 size);

	virtual u32		tell();
	virtual void	seek(u32 pos);
	virtual void	skip(s32 distance);
	virtual void	rewind();

	virtual bool	valid();

	virtual u32		length() { return m_size; }

private:

	u8*			m_data_rw;
	const u8*	m_data_ro;

	u32  m_pos;
	u32  m_size;

};

class VoidStream : public Stream
{
public:

	VoidStream()
		: m_pos(0)
		, m_size(0)
	{}

	virtual u32		write(const void*, u32 size) { m_pos += size; return m_pos; }
	virtual u32		read(void*, u32 size) { m_pos += size; return m_pos; }

	virtual u32		tell() { return m_pos; }
	virtual void	seek(u32 pos) { m_pos = pos; }
	virtual void	skip(s32 distance) { m_pos += distance; }
	virtual void	rewind() { m_pos = 0; }

	virtual bool	valid() { return true; }

	virtual u32		length() { return m_size; }

private:

	u32  m_pos;
	u32  m_size;

};


class FileBase : public Stream
{

public:

	FileBase() : m_file(nullptr) {};
	virtual ~FileBase(){};

	virtual u32		tell();
	virtual void	seek(u32 pos);
	virtual void	skip(s32 distance);
	virtual void	rewind();

	virtual bool	valid();

protected:

	FILE* m_file;

private:

};

class FileIn : public FileBase
{

public:

	FileIn(const char* filename);
	virtual ~FileIn();

	virtual u32	read(void* buf, u32 size);
	virtual u32	write(const void*, u32) { return 0; }
	virtual u32	length();

protected:

private:

};

class FileOut : public FileBase
{

public:

	FileOut(const char* filename, u32 buffer_size = (1 << 20));
	virtual ~FileOut();

	virtual u32	read(void*, u32) { return 0; }
	virtual u32	write(const void* buf, u32 size);
	virtual u32 length() { return 0; }
	void		close();

protected:

	void flush();

private:

	FileOut& operator=(const FileOut&);

	char*		m_buffer;
	const u32	m_buffer_size;
	u32			m_buffer_pos;

};
